#----------------------------------------------------------------
# Copyright (c) 2002 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------

=head2 Browser.pm

Browser.pm is the main module containing the GUI (so named because
I originally envisioned it having an interface sort of like a web
browser.)

The following is the hierarchy of who creates and destroys whom:

 Browser::main_loop()
   BrowserData
   BrowserWindow
     Stage
       Roster
       Grades

=cut

use strict;

use Tk;
use Tk ':variables';
use ExtraGUI;
use GradeBook;
use Crunch;
use Report;
use Words;
use MyWords;
use UtilOG;
use DateOG;
use Input;
use NetOG;
use Digest::SHA1;
use Version;

package Browser;


# See the Words and MyWords modules for info on the following.
our $words;                        # initialized in main_loop()
our $words_prefix;                # initialized in each routine
sub w {
    my $key = shift;
    return get_w($words_prefix,$key);
}
sub get_w {
    my $prefix = shift;
    my $key = shift;
    my $newline = "";
    if ($key =~ m/\n$/) {$newline="\n"}
    chomp $key;
    return $words->get($prefix.".".$key).$newline;
}

our $mw; # main window

sub main_loop {
  my $file_name = shift;

  $words = Words->new(FORMAT=>"terminal",LANGUAGE=>"en");
  local $words_prefix = "b.main_loop";

  $mw = MainWindow->new;
  $mw->geometry("+90+30");
  $mw->maxsize(($mw->screenwidth)-50,($mw->screenheight)-50);
  $mw->minsize(480,600);
    # ... Minsize fixes a bug that makes the window too short when
    #     you run it from the command line with a file as an argument.
  ExtraGUI::fonts_init($mw);
  my $browser_data = BrowserData->new($file_name);
  my $browser_window = BrowserWindow->new($browser_data,$mw);
  $browser_window->set_footer_text(w("startup_info"));
  $mw->repeat(60000,\&periodic_actions); # every 60000 milliseconds, i.e., once a minute
  Tk::MainLoop();
}

sub periodic_actions {
  foreach my $gb(@ogr::open_files) {
    $gb->auto_save_if_they_want_it();
  }
}


=head3 empty_toplevel_window()

Takes one argument, which is the title. Returns
a Toplevel, which it has set to transient.

=cut

sub empty_toplevel_window {
  my $title = shift;
  my $it = $mw->Toplevel();
  $it->transient($mw);
  $it->title($title);
  return $it;
}

=head3 main_window()

I'd prefer not to use this, since empty_toplevel_window()
keeps Browser's data safer, but Tk::FileDialog requires this.

=cut

sub main_window {
  return $mw;
}

#---------------------------------------------------
# Prefs class
#---------------------------------------------------

=head3 Prefs

Preferences for how the GUI is displayed.

=cut

package Prefs;

sub get_focus_highlight_thickness {return 2}
sub get_focus_highlight_color     {return 'blue'}
sub get_roster_hover_color        {return '#e8f0ff'}
sub get_roster_bg_color_1         {return '#e0ffe0'}
sub get_roster_bg_color_2         {return '#fcfffc'}

#---------------------------------------------------
# BrowserData class
#---------------------------------------------------

=head3 BrowserData

This contains all the in-memory information about the state of
an open gradebook file. Almost all of this information is in
a GradeBook file, but there are a few more pieces of data maintained
here, such as a flag that tells whether it has a file open.
the name of the file, etc.

This module does not contain any GUI stuff at all (oops, now it does,
unfortunately...), so for instance
if the GUI was going to be rewritten using a different library,
this parts wouldn't change. To keep GUI stuff out of here, we make
sure that BrowserData code never has to call any methods of any of
the GUI objects. The GUI objects know about the BrowserData, and
they call its methods as needed.

=cut

package BrowserData;

sub w{return Browser::w(@_)}

sub new {
  my $class = shift;
  my $file_name = shift;
  my $self = {};
  local $words_prefix = "b.browser_data.new";
  bless($self,$class);
  $self->{FILE_OPEN} = 0;
  $self->{GB} = "";

  if ($file_name ne "") {
    my $yep_go_ahead = 0;
    my $sub = sub {
      my $password = shift;
      $self->open($file_name,$password);
      $yep_go_ahead = 1;
    };
    ExtraGUI::ask(PROMPT=>w('password'),CALLBACK=>$sub,PASSWORD=>1);
    # The three lines that are commented out below improve the program cosmetically, but make it significantly
    # slower to open a file.
    #Browser::main_window->withdraw(); # Otherwise it's there when we give a filename on the cmd line.
    my $mw = Browser::main_window;
    $mw->waitVariable(\$yep_go_ahead);
    #Browser::main_window->waitVariable(\$yep_go_ahead);
    #Browser::main_window->deiconify();
    #Browser::main_window->raise();
  }
  return $self;
}

=head4 file_is_open

Tells whether file is open (read-only).

=cut

sub file_is_open {
  my $self = shift;
  return $self->{FILE_OPEN};
}

sub file_name {
  my $self = shift;
  return $self->{FILE_NAME};
}

=head4 open()

  open($file_name,$password);

Calls GradeBook->read() and stores the newly created
GradeBook object in $self->{GB}.

=cut

sub open {
  my $self = shift;
  my $file_name = shift;
  my $password = shift;
  local $words_prefix = "b.authentication";

  my $gb = GradeBook->read($file_name,$password);
  if (!ref $gb) {
    ExtraGUI::error_message("$gb");
    return 0;
  }

  my $auth = $gb->{AUTHENTICITY};
  if ($auth ne "" && $auth ne "bad header") {
    ExtraGUI::confirm(w("inauthentic"),
          sub {
            if (!shift) {
              my $mac = LineByLine->new(KEY=>$password,INPUT=>$file_name);
              ExtraGUI::show_text(TEXT=>$mac->report_tampering(),WIDTH=>81,PATH=>Preferences->new()->get('recent_directory'));
            }
          }
          ,w('ok'),w('view_report'));
    return 0;
  }
  if ($auth eq "bad header") {
    ExtraGUI::error_message("This file appears not to have authentication codes. Codes will be added.");
  }

  my $confirm_result = 2;
  if ($gb->auto_save_file_exists()) {
		ExtraGUI::confirm((sprintf w('autosave_check'),$gb->autosave_filename()),sub{$confirm_result=shift});
    $mw->waitVariable(\$confirm_result); # When they make their decision, the var changes to 0 or 1.
  }
  if ($confirm_result==0) {return 0}

  ogr::add_to_list_of_open_files($gb);
  $self->{GB} = $gb;
  $self->{FILE_NAME} = $file_name;
  $self->{FILE_OPEN} = 1;
  return 1;
}

=head4 close()

Saves and closes the file that's currently open. Removes it from the list of files
that are currently open (which are the ones we're supposed to save in case we get
a signal killing us).

=cut

sub close {
  my $self = shift;
  my $file_name = shift;
  local $words_prefix = "b.data.close";
  my $gb = $self->{GB};
  my $result = "";
  if ($self->file_is_open()) { 
    $self->save();
    $self->{FILE_OPEN} = 0;
    ogr::remove_from_list_of_open_files($gb);
    $self->{GB}->close();
    $self->{GB} = "";
  }
  return $result;
}

sub save {
  my $self = shift;
  my $file_name = shift;
  local $words_prefix = "b.data.close";
  my $gb = $self->{GB};
  my $result = "";
  if ($self->file_is_open()) { 
    if (ref $gb) {
      $result = $gb->write(); 
      if ($result ne "") {return $result}
    }
  }
  return $result;
}

=head4 key_to_name()

              KEY=>"",
              ORDER=>"lastfirst", # can be firstlast

=cut

sub key_to_name {
  my $self = shift;
  my %args = (
              KEY=>"",
              ORDER=>"lastfirst", # can be firstlast
              @_,
             );
  if (!$self->file_is_open()) {return}
  my $gb = $self->{GB};
  my $key = $args{KEY};
  my ($first,$last) = $gb->name($key);
  if ($args{ORDER} eq "lastfirst") {
    return "$last, $first";
  }
  else {
    return "$first $last";
  }
}

=head4 get_a_grade()

Arguments are category, assignment, and student key.

=cut

sub get_a_grade {
  my $self = shift;
  my $a_key = shift;
  my $s_key = shift;
  if (!$self->file_is_open()) {return}
  my $gb = $self->{GB};
  my ($cat,$ass) = (GradeBook::first_part_of_label($a_key),
                              GradeBook::second_part_of_label($a_key));
  #print "get_a_grade: $s_key, $cat, $ass, ref=".ref($gb)."=\n";
  return $gb->get_current_grade($s_key,$cat,$ass);
}

=head4 set_a_grade()

Arguments are category, assignment, student key, and grade.

=cut

sub set_a_grade {
  my $self = shift;
  my $a_key = shift;
  my $s_key = shift;
  my $grade = shift;
  if (!$self->file_is_open()) {return}
  my $gb = $self->{GB};
  my ($cat,$ass) = (GradeBook::first_part_of_label($a_key),
                              GradeBook::second_part_of_label($a_key));
  #print "in set_a_grade, $a_key,$cat,$ass\n";
  my %grades = ($s_key=>$grade);
  $gb->set_grades_on_assignment(CATEGORY=>$cat,ASS=>$ass,GRADES=>\%grades);
}
#---------------------------------------------------
# BrowserWindow class
#---------------------------------------------------

=head3 BrowserWindow

Encapsulates the whole GUI window.

=cut

package BrowserWindow;

sub w{return Browser::w(@_)}

#------------ Begin public methods --------------

=head4 new()

 BrowserWindow->new($browser_data,$parent_widget);

=cut

sub new {
  my $class = shift;
  my $data = shift;
  my $parent = shift;
  my $self = {};
  bless($self,$class);
  $self->is_modified(0);

  # Create widgets.
  $self->{PARENT} = $parent;
  $self->{DATA} = $data;
  $self->{MENU_BAR} = $self->menu_bar($parent);
  $self->{FOOTER_TEXT} = "";
  $self->{HEADER_TEXT} = "";
  my $footer = $parent->Label(-textvariable=>\($self->{FOOTER_TEXT}),-font=>ExtraGUI::font('plain'));
  my $header = $parent->Label(-textvariable=>\($self->{HEADER_TEXT}),-font=>ExtraGUI::font('plain'));
  $self->{STAGE} = Stage->new($self,$parent,$data);

  # Pack widgets.
  $self->{MENU_BAR}->pack(-side=>'top',-anchor=>'w',-fill=>'x');
  $header->pack(-side=>'top',-anchor=>'w');
  $self->{STAGE}->{FRAME}->pack(-side=>'top',-anchor=>'w');
  $footer->pack(-side=>'bottom',-anchor=>'w');
  $self->refresh_all();
  $parent->protocol("WM_DELETE_WINDOW",[sub{$_[0]->grades_queue(); $_[0]->quit()},$self]);
  return $self;
}

=head4 set_footer_text

Set the text at the bottom of the browser window.

=cut

sub set_footer_text {
  my $self = shift;
  my $message = shift;
  my $old_message = $self->{FOOTER_TEXT};
  # Problem: sometimes we get two messages, one immediately after the other.
  my $new = "$message........$old_message";
  $new = substr($new,0,72);
  $self->{FOOTER_TEXT} = $new; # This variable is tied to the widget.
}

sub set_header_text {
  my $self = shift;
  $self->{HEADER_TEXT} = shift;
}

=head4 is_modified()

Get or set a flag that says whether the file has been modified relative to
the version on disk. Side-effects: (1) Set the text at the footer of the
window. (2) Set the modification time of the GradeBook object.

=cut

sub is_modified {
  local $words_prefix = "b.is_modified";
  my $self = shift;
  if (@_) {
    $self->{IS_MODIFIED} = shift;
    if (exists($self->{DATA}) && $self->{DATA}->file_is_open()) {
      my $gb = $self->{DATA}->{GB};
      if ($self->{IS_MODIFIED}) {
        # $self->set_footer_text(w('modified')); # not really necessary, because we autosave anyway; it also tends to obscure other, more relevant messages
        $gb->mark_modified_now();
      }
      else {
        # $self->set_footer_text(w('not_modified')); # not really necessary, because we autosave anyway; it also tends to obscure other, more relevant messages
      }
    }
    else {
      $self->set_footer_text('');
    }
  }
  return $self->{IS_MODIFIED};
}

=head4 refresh_all()

Calls refrest_title(), refresh_header(), refresh_roster(), and refresh_assignments().

=cut


sub refresh_all {
  my $self = shift;
  $self->refresh_title();
  $self->refresh_header();
  $self->refresh_roster();
  $self->refresh_assignments();
}

=head4 refresh_header()

Sets header text depending on the file that's open.

=cut

sub refresh_header {
  my $self = shift;
  local $words_prefix = "b.refresh_header";
  my $browser_data = $self->{DATA};
  if ($browser_data->file_is_open()) {
    $self->set_header_text($browser_data->{GB}->title());
  }
  else {
    $self->set_header_text(w("no_file"));
  }
}


=head4 refresh_title()

Put's program name in header.

=cut

sub refresh_title {
  my $self = shift;
  local $words_prefix = "b.refresh_title";
  my $browser_data = $self->{DATA};
  if ($browser_data->file_is_open()) {
    $mw->title(w("program_name").": ".UtilOG::filename_with_path_stripped($browser_data->file_name()));
  }
  else {
    $mw->title(w("program_name"));
  }
}


=head4 refresh_roster()

Flushes grades queue and refreshes $self->{ROSTER}.

=cut

sub refresh_roster {
  my $self = shift;
  my $data = $self->{DATA};
  $self->grades_queue();
  $self->{ROSTER}->refresh();
}

=head4 refresh_assignments()

Flushes grades queue and refreshes $self->{ASSIGNMENTS}.

=cut
sub refresh_assignments {
  my $self = shift;
  my $data = $self->{DATA};
  $self->grades_queue();
  $self->{ASSIGNMENTS}->refresh();
}

=head4 grades_queue()

When the user types in a new grade, it's in a fragile state. They could click somewhere
and select a different student or assignment, and then we need to make sure that the
student's precious grade data isn't lost. To make this totally safe, we implement
a special queue, where we save the grade that's being edited, keystroke by keystroke.
Internally, we maintain this as a hash, so when you make your second
keystroke, the score recorded from the first keystroke gets replaced.
It's vital to make sure this info is saved when we save the file, but that's relatively
easy (see save_file).
The other thing is that we want to make sure it gets recorded ASAP so that
the change is visible throughout the application. For that, we just
try to make sure to flush every single time the user does anything! That means
flushing the queue in refresh_roster(), refresh_assignments(),
add_or_drop()

=cut

my %queue; # moved lexical decl out of BEGIN block as a workaround for a bug in Perl 5.8.4
BEGIN {
  %queue = ();
  sub grades_queue {
    my $self = shift;
    my %args = (
      ACTION=>'flush',                # 'put', or 'flush'
      KEY=>'',                        # e.g. einstein_al.hw.12
      SCORE=>'',
      ENTRY=>'',
      LINKED_VAR=>'',
      @_,
    );  
    local $words_prefix = "b.grades_queue";
    my $action = $args{ACTION};
    my $key = $args{KEY};
    my $score = $args{SCORE};
    my $linked_var = $args{LINKED_VAR};
    my $entry = $args{ENTRY};
    my $data = $self->{DATA};

    # We flush this queue on any excuse, and we may often do it when no file is
    # open. When that happens, well, we sure hope the queue was empty! In any case,
    # we just get dangling references unless we bail out here.
    # Note that a 'put' causes the modification flag to be set, but a 'flush' doesn't.
    # That's because we do flushes at the drop of a hat, unless there's been a previous
    # put, the flush isn't doing anything.

    if (!$data->file_is_open()) {return}

    my $gb = $data->{GB};

    if ($action eq 'put') {
      $queue{$key} = $score;
      $self->is_modified(1);
    }
    if ($action eq 'flush') {
      foreach my $key(keys %queue) {
        $key =~ m/([^\.]+)\.([^\.]+)\.([^\.]+)/;
        my ($who,$cat,$ass) = ($1,$2,$3);
        my $grade = $queue{$key};
        my $gv = $data->get_a_grade($cat.".".$ass,$who);
        my $change_entry = ($grade=~m/[x\+]$/i);
        my $trailing_x = ($grade =~ m/x\+?$/i);
        $grade =~ s/x//i;
        $grade =~ s/\+/\.5/g;
        $grade =~ s/^0(\d+)/$1/; # strip leading zeroes
        if ($change_entry) {
          if (ref $linked_var) {
            $$linked_var = $grade;
          }
          if (ref $entry) {
            $entry->icursor('end'); # otherwise the cursor gets out of whack because we edited the entry
          }
        }
        my $bogus = '';
        # First, check if they're numeric. There are several cases: numeric (possibly with decimal point), null,
        # only whitespace, erroneous input with bogus characters in it. The following regex checks whether it contains
        # only digits and dots, contains no more than one dot, and contains at least one digit. Can also have the
        # optional x on the end for extra credit, but that's already been stripped off. A minus sign on the front is also OK.
        #                        at least 1 digit       sign, then only digits and dots       not more than one dot
          my $old_was_numeric = ($gv    =~ m/[0-9]/) && ($gv    =~ m/^\-?[0-9\.]+$/)       && !($gv    =~ m/\..*\./);
          my $new_is_numeric  = ($grade =~ m/[0-9]/) && ($grade =~ m/^\-?[0-9\.]+$/)       && !($grade =~ m/\..*\./);
          # Nonnumeric input can be ok, but only in the special case of a blank grade:
          if ((!$new_is_numeric) && $grade ne '') {
            $bogus = 'nonnumeric';
          }
        # If they're numeric, we only make the change if they're numerically different. It might seem kind of bogus to
        # check equality on numbers that might be floating point, but how would it make any difference if there was
        # a rounding error in the 8th decimal place of the student's grade? The point here is to avoid thinking the file
        # has been changed, and needs to be saved, when in fact it hasn't. One commonly occurring case is that the student
        # has a blank and you want to change it to a zero, or vice-versa; to me, this is a significant change, since it
        # indicates the student is at least going through the motions of participating. Another possibility is that you
        # erroneously enter somthing like "20foo", then realize you need to edit out the bogus characters and make it "20".
        if ($grade != $gv   ||  (!$old_was_numeric)  || (!$new_is_numeric)) {
                              # old code had this:  $grade eq '' && $gv eq '0'   or   $grade eq '0' && $gv eq '' ) {
          my $do_it = sub {
            if (!@_) {return} # Happens if they're asked to fix a too-high score, and they hit cancel.
            $grade = shift;
            #print "flushing $who,$cat,$ass,new=$grade,old=$gv\n";
            $self->{STAGE}->grades_has_set_grade($who,$cat,$ass,$grade);
            if (ref $linked_var) {
              $$linked_var = $grade;
            }
            if (ref $entry) {
              $entry->icursor('end'); # otherwise the cursor gets out of whack because we edited the entry
            }
          };
          my $max = $gb->assignment_property($cat.".".$ass,"max");
          if ($bogus eq '' && $grade>$max && !$trailing_x && $max>0) {
            $bogus = 'too_high';
          }
          if ($bogus ne '') {
            my $ass_name = $gb->assignment_name($cat,$ass);
            my $name = $data->key_to_name(KEY=>$who,ORDER=>'firstlast');
            ExtraGUI::ask(
                 PROMPT=>(sprintf w($bogus),$grade,$name,$ass_name,$max),
                 TITLE=>w("${bogus}_title"),
                 WIDTH=>40,
                 CALLBACK=>$do_it,
                 DEFAULT=>$grade,
                 BEEP=>1,
            );
          } # end if too high
          else {
            &$do_it($grade);
          } # end if not too high
        } # end if grade has actually changed
      } # end loop over queue
      %queue = ();
    } # end if flushing
    if ($action eq 'initialize') {
      %queue = ();
    }
  }
}

=head4 delete_assignment()

Delete an assignment.

=cut

sub delete_assignment {
  my $self = shift;
  local $words_prefix = "b.delete_assignment";
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $cat = "";
  my $ass = "";
  my $key = "";
  if ($data->file_is_open() && exists $self->{STAGE} && exists $self->{STAGE}->{ASSIGNMENTS}) {
    $cat=$self->{STAGE}->{ASSIGNMENTS}->get_active_category();
    $key=$self->{STAGE}->{ASSIGNMENTS}->selected();
    $ass = $key;
    $ass =~ s/\.(.*)//;
    $ass = $1;
  };
  if ($cat eq '' || $ass eq '') {return}

  ExtraGUI::confirm(sprintf(w('confirm'),($gb->assignment_name($cat,$ass))),
    sub {
      if (shift) {
        $gb->delete_assignment($key);
        $self->{STAGE}->{ASSIGNMENTS}->selected($cat);
        $self->is_modified(1);
        $self->{STAGE}->{ROSTER}->refresh();
        $self->{STAGE}->{ASSIGNMENTS}->refresh();
      }
    }
  );
}

=head4 edit_assignment()

Give the user a form for editing information about a particular assignment.

=cut

sub edit_assignment {
  my $self = shift;
  local $words_prefix = "b.edit_assignment";
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $roster = $self->{STAGE}->{ROSTER};
  my $cat = "";
  my $ass = "";
  my $key = "";
  if ($data->file_is_open() && exists $self->{STAGE} && exists $self->{STAGE}->{ASSIGNMENTS}) {
    $cat=$self->{STAGE}->{ASSIGNMENTS}->get_active_category();
    $key=$self->{STAGE}->{ASSIGNMENTS}->selected();
    $ass = $key;
    $ass =~ s/\.(.*)//;
    $ass = $1;
  };
  if ($cat eq '' || $ass eq '') {return}

  my %old = (
    'name'=>$gb->assignment_name($cat,$ass),
    'due'=>$gb->assignment_property($key,'due'),
    'max'=>$gb->assignment_property($key,'max'),
    'ignore'=>$gb->assignment_property($key,'ignore'),
    'key'=>$ass,
  );
  if ($old{'ignore'} eq '') {$old{'ignore'} = 'false'}

  my $callback = sub {
    my $new = shift;
    my $did_something = 0;
    foreach my $what(keys %old) {
      if (!(exists $old{$what} && $old{$what} eq $new->{$what})) {
        $did_something = 1;
      }
    }
    if ($did_something) {
      if ($new->{'due'} eq '') {delete $new->{'due'}}
      if ($new->{'ignore'} eq 'false') {delete $new->{'ignore'}}

      if ($new->{'key'} ne $old{'key'}) {
        my $result = $gb->rekey_assignment($cat,$old{'key'},$new->{'key'});
        if ($result ne "") {
                ExtraGUI::error_message($result); # bug: always in English
        }
        $ass = $new->{'key'};
        $key = "$cat.$ass";
      }
      delete $new->{'key'};


      # The basic idea is that we don't want to store "name" parameters when the name of the assignment
      # can be correctly inferred from its category and key. The reason we need to set the assignment
      # properties twice is that if we don't do it the first time, the test against the default
      # assignment name won't work. Note that this has to work if they change the key, the name,
      # both, or neither.
      my $new_name = $new->{'name'};
      $new_name =~ s/\"/\'/g; # Otherwise we can't embed it in quotes in the gradebook file.
      delete $new->{'name'};
      $gb->assignment_properties($key,$new);
      if ($new_name ne $gb->default_assignment_name($cat,$ass)) {
        $new->{'name'} = $new_name;
      }
      $gb->assignment_properties($key,$new);

      $self->is_modified(1);
      $roster->refresh();
      $self->{STAGE}->{ASSIGNMENTS}->refresh(); # necesary if assignment was renamed
    };
  };

  ExtraGUI::fill_in_form(
    TITLE=>(sprintf w("title"),ucfirst($gb->assignment_name($cat,$ass))),
    CALLBACK=>$callback,
    COLUMNS=>2,
    INPUTS=>[
          Input->new(KEY=>"name",PROMPT=>w("name"),TYPE=>'string',DEFAULT=>$old{'name'}),
          Input->new(KEY=>"key",PROMPT=>w("key"),TYPE=>'string',DEFAULT=>$old{'key'}),
          Input->new(KEY=>"max",PROMPT=>w("max"),TYPE=>'numeric',DEFAULT=>$old{'max'}),
          Input->new(KEY=>"due",PROMPT=>w("due"),TYPE=>'string',DEFAULT=>$old{'due'}),
          Input->new(KEY=>"ignore",PROMPT=>w("ignore"),TYPE=>'string',DEFAULT=>$old{'ignore'},
                           WIDGET_TYPE=>'radio_buttons',
                           RADIO_BUTTONS=>{'false'=>'no','true'=>'yes'},
                           RADIO_BUTTON_ORDER=>['true','false']),
    ]
  );

}

=head4 new_assignment()

Creates a new assignment. Calls refresh_assignments(), clicked_on_category(),
and clicked_on_assignment().

=cut

sub new_assignment {
  my $self = shift;
  local $words_prefix = "b.new_assignment";

  my $data = $self->{DATA};
  $self->is_modified(1);
  my $gb = $data->{GB};
  my $cat = $self->{ASSIGNMENTS}->get_active_category();
  my $insert_before = ""; # This should be set by the menu item they select.
  my $assignment_exists_message = w('assignment_exists');

  my $callback = sub {
    my $results = shift; # a hash ref
    my $name;
    if ($gb->category_property_boolean($cat,'single')) {
      $name = '1';
    }
    else {
      $name = $results->{"name"};
    }
    my $max = $results->{"max"};
    my $due = $results->{"due"};
    $name =~ s/\"/\'/g; # can't quote it if it contains quotes
    my $raw_name = $name;
    $name =~ s/[^\w]/_/g; # make a valid database key
    if ($due ne "") {$due = DateOG::disambiguate_year($due,$gb->term())}
    my $props = "\"max:$max\"";
    if ($due ne "") {$props = $props .",\"due:$due\""}
    if ($name ne $raw_name) {$props = $props .",\"name:$raw_name\""}
    if ($gb->assignment_exists("$cat.$name")) {
      ExtraGUI::error_message(sprintf $assignment_exists_message,$name,$cat);
    }
    else {
      my $result = $gb->add_assignment(CATEGORY=>$cat,ASS=>$name,COMES_BEFORE=>$insert_before,PROPERTIES=>$props);
      if ($result) {ExtraGUI::error_message($result); return 1}
      $gb->use_defaults_for_assignments();
      $self->{ASSIGNMENTS}->refresh_assignments();
      $self->{ASSIGNMENTS}->clicked_on_category($cat);
      $self->{ASSIGNMENTS}->clicked_on_assignment($name);
    }
  };

  if ($cat eq "") {print "no category\n";return} # Shouldn't happen, since menu item should be dimmed.
  my $cat_props = $gb->category_properties_comma_delimited($cat);

  my $whoops = '';
  my @inputs = ();
  if ($gb->category_property_boolean($cat,'single')) {
    if ($gb->category_contains_assignments($cat)) {$whoops = w('single_assignment_cat_already_has_one')}
  }
  else {
    push @inputs,Input->new(KEY=>"name",PROMPT=>w("name"),TYPE=>'string',BLANK_ALLOWED=>0);
  }
  push @inputs,Input->new(KEY=>"max",PROMPT=>w("max_score"),TYPE=>'numeric',MIN=>0,
                          DEFAULT=>GradeBook::get_property($cat_props,"max"));
  push @inputs,(Input->new(KEY=>"due",PROMPT=>w("due"),TYPE=>'string',BLANK_ALLOWED=>1));

  if ($whoops eq '') {
    ExtraGUI::fill_in_form(
      TITLE=>(sprintf w("new_assignment_title"),$gb->category_name_singular($cat)),
      CALLBACK=>$callback,
      COLUMNS=>1,
      INPUTS=>\@inputs,
    );
  }
  else {
    ExtraGUI::error_message($whoops);
  }

}

sub new_file {
  my $self = shift;
  local $words_prefix = "new_file"; # shared with TermUI.pm

  $self->close_file();

  my $file_name = "";

  my $results = {}; # inputs from the form

  # Due to Perl's lack of a real way of nesting subroutines, the
  # callback routines have to be in reverse order.

  my $callback2 = sub {
    local $words_prefix = "new_file"; # shared with TermUI.pm
    $results = shift;
    my $title = $results->{"title"};
    my $staff = $results->{"staff"};
    my $days_of_week = $results->{"days_of_week"};
    my $year = $results->{"year"};
    my $month = $results->{"month"};
    my $term = $year."-".$month;
    my $password = $results->{"password"};
    $title =~ s/\"/\\\"/g;
    my $gb = GradeBook->new(TITLE=>$title,STAFF=>$staff,DAYS=>$days_of_week,
                            TERM=>$term,FILE_NAME=>$file_name,PASSWORD=>$password);
    my $result = $gb->write();
    if ($result ne '') {
      ExtraGUI::error_message($result)
    }
    else {
      $self->open_file($file_name,$password);
      ExtraGUI::message(w('file_created'));
    }
  };

  my $callback1 = sub {
  $file_name = shift;
  if (!$file_name) {return}
  if (-e $file_name) {ExtraGUI::error_message(sprintf w('file_exists'),$file_name); return}
  ExtraGUI::fill_in_form(
    TITLE=>$file_name,
    CALLBACK=>$callback2,
    COLUMNS=>1,
    INPUTS=>[
      Input->new(KEY=>"title",PROMPT=>w("title"),TYPE=>'string',BLANK_ALLOWED=>0),
      Input->new(KEY=>"staff",PROMPT=>w("staff_gui"),
                              DEFAULT=>UtilOG::guess_username(),TYPE=>'string',BLANK_ALLOWED=>0),
      Input->new(KEY=>"days_of_week",PROMPT=>w("days_of_week"),
                              DEFAULT=>"MTWRF",TYPE=>'string',BLANK_ALLOWED=>1),
      Input->new(KEY=>"year",PROMPT=>w("year"),
                              DEFAULT=>DateOG::current_date("year"),TYPE=>'numeric',MIN=>1900),
      Input->new(KEY=>"month",PROMPT=>w("month"),DEFAULT=>DateOG::current_date("month"),
                  TYPE=>'numeric',MIN=>1,MAX=>12),
      Input->new(KEY=>"password",PROMPT=>w("password"),DEFAULT=>"",TYPE=>'string'),
    ]
  );
  };


  ExtraGUI::choose_file(
    CALLBACK=>$callback1,
    TITLE=>w('file_dlog_title'),
    CREATE=>1,
    PATH=>Preferences->new()->get('recent_directory'),
    WHAT=>'output',
  );
  
}

sub prepare_list_of_recent_files_for_menu {
  my $self = shift;
  my $prefs = Preferences->new(); # Can't use the one associated with the gradebook, because may not have one open.
  my $recent_files = $prefs->get('recent_files');
  my @list = ();
  if ($recent_files ne '') {
    my $n=0;
    my @recent_files = sort (split /,/ , $recent_files);
    foreach my $recent_file(@recent_files) {
      if (-e Portable::do_glob_easy($recent_file)) {
        my $tail_of_filename = UtilOG::filename_with_path_stripped($recent_file);
        if ($tail_of_filename ne '') {
          $n++;
          my $label = "$n. $tail_of_filename";
          push @list,["recent_file$n",$label,$recent_file,sub{my $safe=$self->{SAFE}; &$safe(); $self->open_file($recent_file)}];
        }
      }
    }
  }
  return @list;
}

=head4 open_file()

Normally, we call this routine with no arguments, and it offers the
user a dialog box and opens a file they select. You can also call it
with the filename and password as an argument, in which case there is no interaction
with the user.

=cut

sub open_file {
  my $self = shift;
  my $file_name = "";
  if (@_) {$file_name = shift}
  my $password = "---no password---";
  if (@_) {$password = shift}

  local $words_prefix = "b.open_file";
  my $data = $self->{DATA};
  if ($data->file_is_open()) {$self->close_file()}
  my $sub2 = sub{
      if (!@_) {return} # user hit cancel instead of entering password
      $password = shift;
      if ($file_name eq "") {return}
      my $result = $data->open($file_name,$password);
      if ($result) {
        $self->is_modified(0);
      }
      $self->refresh_all();
      $self->enable_and_disable_menu_items();
      my $gb = $data->{GB};
      my $prefs = $gb->preferences();
      my $recent_files = $prefs->get('recent_files');
      my @recent_files = split /,/ , $recent_files;
      my $full_path = UtilOG::absolute_pathname($file_name);
      #print "full_path=$full_path\n";
      my @rebuilt = ($full_path);
      foreach my $recent_file(@recent_files) {
        my $already_there = 0;
        foreach my $already(@rebuilt) {
          if ($already eq $recent_file) {$already_there = 1}
        }
        if (@rebuilt<5 && !$already_there) {
          push @rebuilt,$recent_file;
        }
      }
      $prefs->set('recent_files',(join ',',@rebuilt));
      my $this_directory = UtilOG::directory_containing_filename($full_path);
      #print "--$full_path,$this_directory\n";
      $prefs->set('recent_directory',$this_directory);

      # Rebuild the list of recent files in the file menu:
      $self->clear_recent_files_from_menu();
      my $menu = $self->{FILE_MENUB}->menu();
      my $l = $self->{RECENT_FILE_LABELS};
      my @prepared_list = $self->prepare_list_of_recent_files_for_menu();
      foreach my $stuff(@prepared_list) {
        my ($foo,$label,$recent_file,$sub) = @$stuff; # $foo is something like "recent_file3" for the third file -- do I ever even use this?
        $menu->insert($menu->index($self->{LABEL_OF_CLEAR_RECENT}),'command',-label=>$label,-command=>$sub);
        push @$l,$label;
      }
      $self->enable_or_disable_specific_menu_item('FILE','clear_recent'); # did all the menus above, but now need to recheck this item on this menu

  };
  my $sub1 = sub {
      $file_name = shift;
      if ($file_name eq "") {return}
      if ($password eq "---no password---") {
        ExtraGUI::ask(PROMPT=>w('password'),CALLBACK=>$sub2,PASSWORD=>1);
      }
      else {
        &$sub2($password);
      }
  };
  if ($file_name eq "") {
    my @args = (CALLBACK=>$sub1,TITLE=>w('dlog_title'));
    my $prefs = Preferences->new(); # Can't use the one associated with the gradebook, because may not have one open.
    my $recent_directory = $prefs->get('recent_directory');
    #print "recent directory = $recent_directory\n";
    if ($recent_directory ne '') {
      push @args,'PATH';
      push @args,$recent_directory;
    }
    ExtraGUI::choose_file(@args);
  }
  else {
    &$sub1($file_name);
  }
  $self->enable_and_disable_menu_items();
}

sub rekey_file {
  my $self = shift;
  local $words_prefix = "b.rekey_file";
  my $file_name;
  my $sub2 = sub {
      my $password = shift;
      my $mac = LineByLine->new(KEY=>$password,ITERATIONS=>GradeBook::iterations_for_line_by_line());
      my $err = $mac->rekey(INFILE=>$file_name,OUTFILE=>$file_name);
      if ($err ne '') {ExtraGUI::error_message($err)}
  };
  my $sub1 = sub {
      $file_name = shift;
      if ($file_name eq "") {return}
      ExtraGUI::ask(PROMPT=>w('password'),CALLBACK=>$sub2,PASSWORD=>1);
  };
  my @args = (CALLBACK=>$sub1,TITLE=>w('dlog_title'));
  my $prefs = Preferences->new(); # Can't use the one associated with the gradebook, because they don't have one open.
  my $recent_directory = $prefs->get('recent_directory');
  if ($recent_directory ne '') {
    push @args,'PATH';
    push @args,$recent_directory;
  }
  ExtraGUI::choose_file(@args);
}

sub reconcile {
  my $self = shift;
  local $words_prefix = "b.reconcile";
  my ($file_a,$file_b,$recent_directory,$password,$a);
  my $sub4 = sub {
    local $words_prefix = "b.reconcile";
    my $ass = shift;
    my $who = shift;
    my $a_grade = shift;
    my $b_grade = shift;
    my $message_form = w('pick');
    my $message = sprintf $message_form,$who,$ass,$a_grade,$b_grade;
    my $decision;
    ExtraGUI::confirm($message,sub {$decision = shift},w('yes'),w('no'));
    Browser::main_window->waitVariable(\$decision);
    return $decision;
  };
  my $sub3 = sub {
      local $words_prefix = "b.reconcile";
		  my $file_b = shift;
      my $b = GradeBook->read($file_b,$password); # assume same password
      if (!ref $b) {ExtraGUI::error_message($b); return}
      my $log = $a->union($b,$sub4);
      if ($log eq '') {$log="No changes.\n"}
      $log = "Changing file $file_a,\n"."folding in file $file_b\n".$log;
      my $result = $a->write_to_named_file($file_a);
      if ($result) {ExtraGUI::error_message($result); return}
			ExtraGUI::show_text(TITLE=>w('summary'),TEXT=>$log);
  };
  my $sub2 = sub {
      local $words_prefix = "b.reconcile";
      $password = shift;
      $a = GradeBook->read($file_a,$password);
      if (!ref $a) {ExtraGUI::error_message($a); return}
      my @args2 = (CALLBACK=>$sub3,TITLE=>w('file_to_fold_in'));
      if ($recent_directory ne '') {
        push @args2,'PATH';
        push @args2,$recent_directory;
		  }
      ExtraGUI::choose_file(@args2);
		};
  my $sub1 = sub {
      local $words_prefix = "b.reconcile";
      $file_a = shift;
      if ($file_a eq "") {return}
      ExtraGUI::ask(PROMPT=>w('password'),CALLBACK=>$sub2,PASSWORD=>1);
  };
  my @args = (CALLBACK=>$sub1,TITLE=>w('file_to_change'));
  my $prefs = Preferences->new(); # Can't use the one associated with the gradebook, because they don't have one open.
  $recent_directory = $prefs->get('recent_directory');
  if ($recent_directory ne '') {
    push @args,'PATH';
    push @args,$recent_directory;
  }
  ExtraGUI::choose_file(@args);
}

sub strip_watermark {
  my $self = shift;
  local $words_prefix = "b.strip_watermark";
  my $file_name;
  my $sub = sub {
      $file_name = shift;
      my $mac = LineByLine->new();
      my $err = $mac->strip(INFILE=>$file_name,OUTFILE=>$file_name);
      if ($err ne '') {ExtraGUI::error_message($err)}
  };
  my @args = (CALLBACK=>$sub,TITLE=>w('dlog_title'));
  my $prefs = Preferences->new(); # Can't use the one associated with the gradebook, because they don't have one open.
  my $recent_directory = $prefs->get('recent_directory');
  if ($recent_directory ne '') {
    push @args,'PATH';
    push @args,$recent_directory;
  }
  ExtraGUI::choose_file(@args);
}

=head4 close_file()

Calls close() on the BrowserData object. Rebuilds the menu above the roster, and disables items in all menus
that you can't do when there's no file open.

=cut

sub close_file {
  my $self = shift;
  my $file_name = shift;
  my $data = $self->{DATA};
  if (!$data->file_is_open()) {return}
  my $result = $data->close();
  $self->enable_and_disable_menu_items();
  $self->{STAGE}->{ASSIGNMENTS}->refresh_categories();
  $self->{STAGE}->{ASSIGNMENTS}->{CAT_INTENTIONALLY_SELECTED} = '';
  $self->{STAGE}->{ROSTER}->make_options_menu();
  $self->refresh_all();
  if ($result) {
    ExtraGUI::error_message($result)
  }
  else {
    $self->is_modified(0);
  }
}

=head4 save_file()

Flush the grades queue, save the file, enable and disable menu items,
and refresh all.

=cut

sub save_file {
  my $self = shift;
  my $file_name = shift;
  my $data = $self->{DATA};
  $self->grades_queue(); # flush the queue
  if (!$data->file_is_open()) {return}
  my $result = $data->save();
  $self->enable_and_disable_menu_items();
  if ($result) {ExtraGUI::error_message($result)}
  $self->is_modified(0);
}

=head4 clone()

Clones the file. Clears the list of assignments and the roster, but leaves categories and grading standards intact.

=cut

sub clone {
  my $self = shift;
  local $words_prefix = "b.clone";
  my $gb = $self->{DATA}->{GB};
  $gb->clear_assignment_list();
  $gb->clear_roster();
  $gb->clear_grades();
  my $staff = $gb->staff();
  my $standards = GradeBook::hash_to_comma_delimited($gb->standards());
  my $callback1 = sub {
    my $file_name = shift;
    $gb->file_name($file_name);
    $gb->write_to_named_file($file_name);
  };
  my $callback2 = sub {
    my $inputs = shift;
    my $term = $inputs->{'year'}.'-'.$inputs->{'month'};
    $gb->set_class_data({'title'=>$inputs->{'title'},'staff'=>$staff,'days'=>$inputs->{'days_of_week'},'term'=>$term,'dir'=>$inputs->{'dir'},'standards'=>$standards});
    $gb->password($inputs->{'password'});
    ExtraGUI::choose_file(TITLE=>w('save_as'),WHAT=>'output',CREATE=>1,CALLBACK=>$callback1);
  };
  ExtraGUI::fill_in_form(
      TITLE=>w('what'),
      CALLBACK=>$callback2,
      COLUMNS=>1,
      INPUTS=>[
        Input->new(KEY=>"title",PROMPT=>w("title"),TYPE=>'string',BLANK_ALLOWED=>0,DEFAULT=>$gb->title()),
        Input->new(KEY=>"days_of_week",PROMPT=>w("days_of_week"),
                              DEFAULT=>"MTWRF",TYPE=>'string',BLANK_ALLOWED=>1),
        Input->new(KEY=>"year",PROMPT=>w("year"),
                              DEFAULT=>DateOG::current_date("year"),TYPE=>'numeric',MIN=>1900),
        Input->new(KEY=>"month",PROMPT=>w("month"),DEFAULT=>DateOG::current_date("month"),
                  TYPE=>'numeric',MIN=>1,MAX=>12),
        Input->new(KEY=>"password",PROMPT=>w("password"),DEFAULT=>"",TYPE=>'string'),
        Input->new(KEY=>"dir",PROMPT=>w("dir"),TYPE=>'string',DEFAULT=>$gb->dir()),
      ]
    );
}

=head4 server()

Make requests to the server. The argument tells what kind of request.

=cut
sub server {
  my $self = shift;
  my $what = shift;
  local $words_prefix = "b.server";
  my $gb = $self->{DATA}->{GB};
  my $prefs = $gb->preferences();
  my $recent_dir = $prefs->get('recent_directory');
  my $server_domain = $prefs->get('server_domain');
  my $server_user = $prefs->get('server_user');
  my $server_account = $prefs->get('server_account');
  my $server_key = $prefs->get('server_key');
  my $server_class = $gb->dir();
  my $err_info =  "domain=$server_domain user=$server_user account=$server_account key=$server_key class=$server_class\n";
  #print "$err_info\n";
  my $request = NetOG->new();
  if ($what eq 'sent_messages') {
    my $who = shift;
  	ExtraGUI:show_text(TEXT=>$who);
  }
	if ($what eq 'post_message' || $what eq 'post_message_one') {
    my $post_it = sub {
      local $words_prefix = "b.server";
      my $args = shift;
      my $to = $args->{'to'};
      my $do_email = (1==($args->{'do_email'}));
      my $body = $args->{'body'};
      my $subject = $args->{'subject'};
      my $other_headers = '';
      if (exists $args->{'other_headers'}) {$other_headers = $args->{'other_headers'}}
      my $date = DateOG::current_date_for_message_key();
      if ($to eq '*') {$to = join "," , ($gb->student_keys());} # This lists only active students.
      # To make sure the message's key is unique, we append the 1st 4 chars of an SHA1 hash to the date.
      my $hash = Digest::SHA1::sha1_base64($to.$body.$date);
      $hash =~ m/^(....)/; # get 1st 4 chars
      $hash = $1;
      $hash =~ s@/@_@g; # Unix filenames shouldn't have slashes in them.
      $hash =~ s@\+@_@g; # Plus signs also seem to cause problems.
      my $err = $request->be_client(GB=>$gb,
          HOST=>$server_domain,SERVER_KEY=>$server_key,
          PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
                 'what'=>'post_message','body'=>$body,'to'=>"to=$to",'subject'=>"subject=$subject",
                 'other_headers'=>$other_headers,'date'=>$date,'hash'=>$hash,'do_email'=>$do_email,
                }
          );
      if ($err eq '') {$err = $request->{RESPONSE_DATA}}
      if ($err ne '') {
        if ($err eq 'set_server_key' || $err eq 'error_connecting') {$err = w($err)}
        ExtraGUI::error_message("Error: $err")
      }
    };
    my $class_description = $gb->title(); # in case the following fails because they aren't connected yet
    my $foofoo = $request->be_client(GB=>$gb,
        HOST=>$server_domain,SERVER_KEY=>$server_key,
        PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
               'what'=>'get_class_data',
               'get_what'=>'description'}
        );
    if ($foofoo eq '') {$class_description = $request->{RESPONSE_DATA}}
    my $default_recipient = '*';
    if ($what eq 'post_message_one') {
      $default_recipient = shift;
    }
    my @inputs = ();
    push @inputs,Input->new(KEY=>'to',PROMPT=>w('to'),TYPE=>'string',BLANK_ALLOWED=>0,DEFAULT=>$default_recipient);
    push @inputs,Input->new(KEY=>'do_email',PROMPT=>w('do_email'),
           TYPE=>'string',WIDGET_TYPE=>'radio_buttons',BLANK_ALLOWED=>0,DEFAULT=>0);
    push @inputs,Input->new(KEY=>'subject',PROMPT=>w('subject'),TYPE=>'string',BLANK_ALLOWED=>0,DEFAULT=>$class_description);
    push @inputs,Input->new(KEY=>'body',PROMPT=>w('body'),TYPE=>'string',BLANK_ALLOWED=>0,WIDGET_TYPE=>'text');
    my $do_form = sub {
      ExtraGUI::fill_in_form(INPUTS=>\@inputs,CALLBACK=>$post_it,COLUMNS=>1,WIDTH=>120);
  	};
    if ($default_recipient ne '*') {
      &$do_form();
    }
    else {
			ExtraGUI::confirm(w('confirm_whole_class'),$do_form);
    }
  }
  if ($what eq 'class_description') {
    my $err = $request->be_client(GB=>$gb,
        HOST=>$server_domain,SERVER_KEY=>$server_key,
        PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
               'what'=>'get_class_data',
               'get_what'=>'description'}
        );
    if ($err ne '') {ExtraGUI::error_message("Error, $err, $err_info")}
    else {ExtraGUI::show_text(TITLE=>w('class_description'),TEXT=>$request->{RESPONSE_DATA},PATH=>$recent_dir);}
  }
  if ($what eq 'roster') {
    my $err = $request->be_client(GB=>$gb,
        HOST=>$server_domain,SERVER_KEY=>$server_key,
        PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
               'what'=>'get_class_data',
               'get_what'=>'roster'}
        );
    if ($err ne '') {ExtraGUI::error_message("Error, $err, $err_info")}
    else {ExtraGUI::show_text(TITLE=>w('roster'),TEXT=>$request->{RESPONSE_DATA},PATH=>$recent_dir);}
  }
  if ($what eq 'email' || $what eq 'emailone') {
    my $list = '';
    if ($what eq 'email') {
      my @roster = $gb->student_keys();
      foreach my $who(@roster) {$list = $list.",".$who}
      $list =~ s/^,//;
    }
    if ($what eq 'emailone') {
      $list = shift;
    }
    my $err = $request->be_client(GB=>$gb,
        HOST=>$server_domain,SERVER_KEY=>$server_key,
        PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
               'what'=>'get_class_data',
               'get_what'=>'email','who'=>$list}
        );
    if ($err ne '') {ExtraGUI::error_message(w('error').", $err, $err_info")}
    else {ExtraGUI::show_text(TITLE=>'',TEXT=>$request->{RESPONSE_DATA},PATH=>$recent_dir);}
  }
  if ($what eq 'list_work') {
    my $err = $request->be_client(GB=>$gb,
        HOST=>$server_domain,SERVER_KEY=>$server_key,
        PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
               'what'=>'get_class_data',
               'get_what'=>'list_work'}
        );
    if ($err ne '') {ExtraGUI::error_message("Error, $err")}
    else {
    my $list = $request->{RESPONSE_DATA};
    my @stuff = ();
    my $box = Browser::empty_toplevel_window('');
    my $f1 = $box->Frame()->pack();
    my $due = DateOG::current_date_sortable_with_hm();
    my $due_date_entry = $f1->Label(-text=>'due date')->pack(-side=>'left');
    my $due_date_entry = $f1->Entry(-width=>16,-textvariable=>\$due)->pack(-side=>'left');
    my $s = $box->Scrolled("Canvas",
        -scrollbars=>'e',
        -height=>(($mw->screenheight)-150),
    )->pack();
    my $c = $s->Subwidget("canvas");
    my $n = 0;
    my @checked = ();
    while ($list=~m/^([^\n]+)$/mg) {
      push @stuff,$1;
      my $b = $s->Checkbutton(-text=>$1,-variable=>\$checked[$n]);
      $checked[$n] = 0;
      $s->createWindow(0,25*($n+1),-window=>$b,-anchor=>'w');
      $n++;
    }
    my %scores;
    my @roster;
    my $f = $box->Frame();
    my $get_em = sub {
                 my $do_what = shift;
                 $box->destroy();
                 my $which = '';
                 for (my $i=0; $i<$n; $i++) {
                   if ($checked[$i]) {$which = $which . $stuff[$i]."\n"}
                 }
                 $request->be_client(GB=>$gb,
                      HOST=>$server_domain,SERVER_KEY=>$server_key,
                      PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
                             'what'=>'get_class_data','get_what'=>'get_scores','due'=>$due,'which'=>$which,
                             'client_time_zone'=>my_time_zone()
                             }
                 );
                 my $r = $request->{RESPONSE_DATA};
                 @roster = $gb->student_keys();
                 my $t = '';
                 $r =~ m/\n=key,([^\n]*)\n/; # Extract the key that tells us what problems these are the scores on.
                 my @key = split /,/ , $1;
                 $t = $t . sprintf "%25s ",'';
                 my $count = 0;
                 foreach my $key(@key) {
                   my $char = chr(ord('a')+$count);
                   $count++;
                   $t = "$t$char";
                 }
                 $t = "$t\n";
                 foreach my $who(@roster) {
                   my ($first,$last) = $gb->name($who);
                   $t = $t . sprintf "%25s","$first $last";
                   if ($r =~ m/$who\=([^\n]*)/m) {
                     my $scores = $1; # a list of binary bits
                     $t = $t . " $1";
                     my $total = $scores;
                     $total =~ s/0//g;
                     $scores{$who} = length $total;
                   }
                   $t = $t . "\n";
                 }
                 $t = $t . "\nkey:\n";
                 $count = 0;
                 foreach my $key(@key) {
                   my $char = chr(ord('a')+$count);
                   $count++;
                   $t = $t."  $char $key\n";
                 }
                 if ($do_what eq 'show') {
                   ExtraGUI::show_text(TITLE=>'',TEXT=>$t,PATH=>$recent_dir);
                 }
                 if ($do_what eq 'save') {
                    my $a_key = shift;
                    foreach my $who(@roster) {
                      if (exists $scores{$who}) {
                        $self->grades_queue(ACTION=>'put',KEY=>"$who.$a_key",SCORE=>$scores{$who});
                      }
                    }
                    $self->grades_queue();
                    $self->refresh_assignments();
                 }
    };
    $f->Button(-text=>'ok',
               -command=>sub {&$get_em('show')}
    )->pack(-side=>'left');
    $f->Button(-text=>'cancel',-command=>sub{$box->destroy()})->pack(-side=>'left');
    $f->Button(
             -text=>"view log",
             -command=>sub
               {
                 $box->destroy();
                 my $which = '';
                 for (my $i=0; $i<$n; $i++) {
                   if ($checked[$i]) {$which = $which . $stuff[$i]."\n"}
                 }
                 $request->be_client(GB=>$gb,
                      HOST=>$server_domain,SERVER_KEY=>$server_key,
                      PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
                             'what'=>'get_class_data','get_what'=>'answers','which'=>$which,
                             }
                 );
                 my $r = $request->{RESPONSE_DATA};
                 ExtraGUI::show_text(TITLE=>$which,TEXT=>$r,WIDTH=>100,PATH=>$recent_dir);
               }
    )->pack(-side=>'left');
    if ($self->{ASSIGNMENTS}->specific_assignment_selected()) {
      my $ass = $self->{ASSIGNMENTS}->selected();
      $ass =~ m/([^\.]*)\.([^\.]*)/;
      my ($c,$a) = ($1,$2);
      my $ass_name = $gb->assignment_name($c,$ass);
      $f->Button(
             -text=>"save in $ass_name",
             -command=>sub{&$get_em('save',$ass)}
      )->pack(-side=>'left');
    }
    $f->pack();
    $box->geometry(ExtraGUI::preferred_location());
    }#end if no error
  }
  if ($what eq "upload") {

    my $n = 40;
    my ($bar,$w,$text);


    $w = Browser::empty_toplevel_window('uploading');
    $w->geometry(ExtraGUI::preferred_location());
    $text = "-" x $n;
    $w->withdraw();
    $w->Label(-text=>'uploading...')->pack(-side=>'top');
    $bar = $w->Label(-textvariable=>\$text,-font=>ExtraGUI::font('fixed_width'))->pack(-side=>'top');

    $w->deiconify();
    $w->raise();
    my $error = Report::upload_grades(GB=>$gb,
         PROGRESS_BAR_CALLBACK=>sub{
            my $x = shift;
            my $m = $x*$n;
                  Browser::main_window()->update;
            $text = ("=" x $m . "-" x ($n-$m));
                  Browser::main_window()->update;
          },
                FINAL_CALLBACK=>sub{
            $w->destroy()
          }
    );
    if ($error ne '') {ExtraGUI::error_message($error)}
  }
}

# returns the local time zone in units of hours; result may be a non-integer;
# west of Greenwich is negative
sub my_time_zone {
  my $t = time();
  my @local = localtime($t);
  my @gmt = gmtime($t);
  my $secs = ($local[0]-$gmt[0])+($local[1]-$gmt[1])*60+($local[2]-$gmt[2])*3600;
  if ($local[3]!=$gmt[3] || $local[4]!=$gmt[4] || $local[5]!=$gmt[5]) {
    if ($local[5]>$gmt[5] || ($local[5]==$gmt[5] && $local[4]>$gmt[4]) || ($local[5]==$gmt[5] && $local[4]==$gmt[4] && $local[3]>$gmt[3])) {
      $secs = $secs + 3600*24;
    }
    else {
      $secs = $secs - 3600*24;
    }
  }
  return $secs/3600.;
}


=head4 options()

Set options. The argument tells what kind: 'server', etc.

=cut
sub options {
  my $self = shift;
  my $what = shift;
  local $words_prefix = "b.options";
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $prefs;
  if ($gb) {$prefs = $gb->preferences()} else {$prefs = Preferences->new()}
  if ($what eq 'server') {
    my $username = UtilOG::guess_username();
    my $web_callback = sub {
      my $results = shift;
      my $prefs = $gb->preferences();
      $prefs->set('server_domain',$results->{'domain'});
      $prefs->set('server_user',$results->{'user'});
      $prefs->set('server_account',$results->{'account'});
      $prefs->set('server_key',$results->{'key'});
      $gb->dir($results->{'class'});
      $self->is_modified(1);
    };
    my $default_domain = $prefs->get('server_domain');
    #print "default_domain=$default_domain\n";
    if ($default_domain eq '') {$default_domain='lightandmatter.com'}
    my $default_user = $prefs->get('server_user');
    if ($default_user eq '') {$default_user=$username}
    my $default_account = $prefs->get('server_account');
    if ($default_account eq '') {$default_account=$username}
    my $default_key = $prefs->get('server_key');
    ExtraGUI::fill_in_form(
      TITLE=>w('web.title'),
      CALLBACK=>$web_callback,
      COLUMNS=>1,
      INPUTS=>[
        Input->new(KEY=>"domain",PROMPT=>w("web.server"),TYPE=>'string',DEFAULT=>$default_domain),
        Input->new(KEY=>"user",PROMPT=>w("web.server_username"),TYPE=>'string',DEFAULT=>$default_user),
        Input->new(KEY=>"account",PROMPT=>w("web.server_account"),TYPE=>'string',DEFAULT=>$default_account),
        Input->new(KEY=>"class",PROMPT=>w("web.server_class"),TYPE=>'string',DEFAULT=>$gb->dir()),
        Input->new(KEY=>"key",PROMPT=>w("web.server_key"),TYPE=>'string',DEFAULT=>$default_key),
      ]
    );
  }
  if ($what eq 'beep') {
    my $to_beep_or_not_to_beep = $prefs->get('beep');
    my @inputs = (
       Input->new(KEY=>'beep',
                  PROMPT=>w('preferences.beeping'),
                  DEFAULT=>$to_beep_or_not_to_beep,
                  TYPE=>'string',
                  WIDGET_TYPE=>'radio_buttons'
       ),
    );
    ExtraGUI::fill_in_form(
      INPUTS => \@inputs,
      COLUMNS=>1,
      CALLBACK=>sub{
        my $results = shift;
        $prefs->set('beep', $results->{'beep'});
      }
    );
  }
  if ($what eq 'justify') {
    my $justify = $prefs->get('justify');
    if ($justify eq '' || ($justify!=0 && $justify!=1)) {$justify=0} # default to left because right can cause problem with Perl/Tk bug
    my @inputs = (
       Input->new(KEY=>'justify',
                   PROMPT=>w('preferences.justify'),
                   DEFAULT=>$justify,
                   TYPE=>'string',
                               WIDGET_TYPE=>'radio_buttons',
                                     RADIO_BUTTONS=>{0=>'left',1=>'right'},
                   RADIO_BUTTON_ORDER=>[0,1],
                         ),
    );
    ExtraGUI::fill_in_form(
      INPUTS => \@inputs,
      COLUMNS=>1,
      CALLBACK=>sub{
        my $results = shift;
        $prefs->set('justify', $results->{'justify'});
        $self->refresh_roster();
      }
    );
  }
  if ($what eq 'standards' && $gb) {
    my ($box,$f);
    my $grade_symbols = 'A B C D F';
    my @symbols = ();
    my %standards = ();

    # The following have to be in reverse order because perl doesn't really have nested subroutines.

    my $standards_callback2 = sub {
      my $results = shift;
      $standards{$symbols[$#symbols]} = 0;
      for (my $i=0; $i<=$#symbols-1; $i++) {
        my $symbol = $symbols[$i];
        $standards{$symbol} = $results->{$symbol};
        if ($i>=1 && $standards{$symbol}>=$standards{$symbols[$i-1]}) {
          ExtraGUI::error_message(w('standards.not_in_order'));
        }
      }
      my $standards_comma_delimited = "";
      for (my $i=0; $i<=$#symbols; $i++) {
        $standards_comma_delimited = $standards_comma_delimited 
            . ',"'.$symbols[$i].':'.$standards{$symbols[$i]}.'"';
      }
      $standards_comma_delimited =~ s/^\,//; # remove leading comma
      $gb->set_standards($standards_comma_delimited);
      $self->is_modified(1);
    };

    my $standards_callback1 = sub {
      local $words_prefix = "b.options";
      my $preexisting_standards = $gb->standards();
      @symbols = split /\s+/,$grade_symbols;
      if (@symbols<2) {ExtraGUI::error_message(w('standards.at_least_two')); return}
      if (@symbols>100) {ExtraGUI::error_message(w('standards.too_many')); return}
      # Check for duplicates:
      for (my $i=0; $i<=$#symbols; $i++) {
        for (my $j=0; $j<=$#symbols; $j++) {
          if ($i!=$j && $symbols[$i] eq $symbols[$j]) {ExtraGUI::error_message(w('standards.not_unique')); return}
        }
      }
      my @inputs = ();
      # The last symbol has to have a minimum of zero, so it doesn't even have a space in the form.
      for (my $i=0; $i<=$#symbols-1; $i++) {
        my $symbol = $symbols[$i];
        my $default = "";
        if (exists $preexisting_standards->{$symbol}) {$default=$preexisting_standards->{$symbol}}
        push @inputs, Input->new(KEY=>$symbol,PROMPT=>$symbol,TYPE=>'numeric',MIN=>0,MAX=>100,DEFAULT=>$default);
      }
      ExtraGUI::fill_in_form(
        TITLE=>w('standards.pct_title'),
        CALLBACK=>$standards_callback2,
        COLUMNS=>2,
        INPUTS=>\@inputs
      );
    };

    my $previous_standards = $gb->standards(); # hash ref like {"A"=>90,...}
    my @previous_symbols = keys(%$previous_standards);
    @previous_symbols = sort {$previous_standards->{$b} <=> $previous_standards->{$a}} @previous_symbols;
    $box = Browser::empty_toplevel_window(w('standards.symbols_title'));
    $f = $box->Frame->pack(-side => 'bottom');
    $f->Button(-text=>w("ok"),-command=>sub{$box->destroy(); &$standards_callback1(); }
    )->pack(-side=>'left');
    $f->Button(-text=>w("cancel"),-command=> sub{$box->destroy(); }
    )->pack(-side=>'left');
    my $e = $box->Entry(-textvariable=>\$grade_symbols,-width=>60)->pack(-side=>'top');
    my $type;
    if (@previous_symbols) {
      $type = 3;
      $grade_symbols = '';
      foreach my $symbol(@previous_symbols) {$grade_symbols = $grade_symbols . ' '.$symbol}
    }
    else {
      $type = 1;
      $grade_symbols = 'A B C D F';
    }
    $grade_symbols =~ s/^ //; # strip leading blank
    $box->Radiobutton(-text=>'ABCDF',-justify=>'left',-value=>1,-variable=>\$type,
               -command=>sub{$grade_symbols='A B C D F'})
         ->pack(-side=>'top',-anchor=>'w');
    $box->Radiobutton(-text=>'ABCDF+-',-justify=>'left',-value=>2,-variable=>\$type,
               -command=>sub{$grade_symbols='A+ A A- B+ B B- C+ C C- D+ D D- F'})
         ->pack(-side=>'top',-anchor=>'w');
    $box->Radiobutton(-text=>w('standards.custom'),-justify=>'left',-value=>3,-variable=>\$type,
              -command=>sub{$grade_symbols=''; $e->focus()})
         ->pack(-side=>'top',-anchor=>'w');
  }
}

=head4 report()

Display a report. The argument tells what kind: 'class_totals', etc.

=cut

sub report {
  my $self = shift;
  my $what = shift;
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $prefs = $gb->preferences();
  my $recent_dir = $prefs->get('recent_directory');
  my $filename = $gb->file_name();
  local $words_prefix = "b.report";
  if ($what eq "stats_ass") {
    my $format = "plain";
    my $title = w("$what.title");
    my $cat=$self->{ASSIGNMENTS}->get_active_category();
    my $ass=$self->{ASSIGNMENTS}->selected();
    $ass =~ s/^[^.]*\.//;
    ExtraGUI::show_text(TITLE=>$title,TEXT=>Report::statistics_ass($gb,$format,$cat,$ass)->text(),PATH=>$recent_dir);
  }
  if ($what eq "sort_by_overall" || $what eq "sort_by_category" || $what eq "sort_by_assignment") {
    # These are all the ones that require the user to decide what to sort by.
    my $format = "plain";
    my $title = w("$what.title");
    my $sort_type = -1;
    ExtraGUI::confirm($title,sub{$sort_type = shift},w('sort_by_score'),w('sort_by_name'));
    my $mw = Browser::main_window;
    $mw->waitVariable(\$sort_type);
    if ($sort_type == 0) {$format=$format.'*'}
    if ($what eq "sort_by_overall") {
      ExtraGUI::show_text(TITLE=>$title,TEXT=>Report::sort($gb,$format)->text(),PATH=>$recent_dir);
    }
    if ($what eq "sort_by_category") {
      my $cat=$self->{ASSIGNMENTS}->get_active_category();
      ExtraGUI::show_text(TITLE=>$title,TEXT=>Report::sort($gb,$format,$cat)->text(),PATH=>$recent_dir);
    }
    if ($what eq "sort_by_assignment") {
      my $cat=$self->{ASSIGNMENTS}->get_active_category();
      my $ass=$self->{ASSIGNMENTS}->selected();
      $ass =~ s/^[^.]*\.//;
      ExtraGUI::show_text(TITLE=>$title,TEXT=>Report::sort($gb,$format,$cat,$ass)->text(),PATH=>$recent_dir);
    }
  }
  if ($what eq "student") {
    my $r = $self->{ROSTER};
    my $key = $r->selected();
    my $name = $data->key_to_name(KEY=>$key,ORDER=>"firstlast");
    my $t= Report::student(GB=>$gb,STUDENT=>$key,FORMAT=>"plain");
    ExtraGUI::show_text(TITLE=>$name,TEXT=>$t->text(),PATH=>$recent_dir);
  }
  if ($what eq "table") {
    my $r = $self->{ROSTER};
    my ($t,$width)= Report::table(GB=>$gb,FORMAT=>"plain");
    ExtraGUI::show_text(TITLE=>'',TEXT=>$t->text(),WIDTH=>$width,PATH=>$recent_dir);
  }
  if ($what eq "roster") {
    my @roster = $gb->student_keys();
    my $t = '';
    foreach my $who(@roster) {
      my ($first,$last) = $gb->name($who);
      $t = $t . "$first $last\n";
    }
    my $roster_filename = $filename;
    $roster_filename =~ s/\.gb/\.roster/;
    ExtraGUI::show_text(TITLE=>'',TEXT=>$t,PATH=>$recent_dir,FILENAME=>$roster_filename);
  }
  if ($what eq "stats") {
    ExtraGUI::show_text(TITLE=>w('stats_title'),TEXT=>Report::stats(GB=>$gb,FORMAT=>"plain")->text(),PATH=>$recent_dir);
  }
}

=head4 quit()

Takes no arguments. Closes the file this Browser object has open, and any other files; exits.

=cut

sub quit {
  my $self = shift;
  $self->close_file();
  ogr::clean_up_before_exiting();
  if (ref $mw) {$mw->destroy}
}

=head4 edit_student()

Takes no arguments.

=cut

sub edit_student {
  my $self = shift;
  local $words_prefix = "b.edit_student";
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $roster = $self->{STAGE}->{ROSTER};
  my $student=$roster->selected();
  my $name = $data->key_to_name(KEY=>$student,ORDER=>'firstlast');


  if ($student eq "" || !($name =~ m/\w+/)) {return}
         # ...See comments on the similar test in add_or_drop().

  my ($first,$last) = $gb->name($student);
  my %old = (
    'last'=>$last,
    'first'=>$first,
    'id'=>$gb->get_student_property($student,'id'),
  );

  my $callback = sub {
    my $new = shift;
    my $did_something = 0;
    foreach my $what(keys %old) {
      if (!(exists $old{$what} && $old{$what} eq $new->{$what})) {
        $gb->set_student_property($student,$what,$new->{$what});
        $did_something = 1;
      }
    }
    if ($did_something) {
      $self->is_modified(1);
      $roster->refresh();
    };
  };

  ExtraGUI::fill_in_form(
    TITLE=>(sprintf w("title"),$name),
    CALLBACK=>$callback,
    COLUMNS=>2,
    INPUTS=>[
          Input->new(KEY=>"last",PROMPT=>w("last"),TYPE=>'string',DEFAULT=>$old{'last'}),
          Input->new(KEY=>"first",PROMPT=>w("first"),TYPE=>'string',DEFAULT=>$old{'first'}),
          Input->new(KEY=>"id",PROMPT=>w("id"),TYPE=>'string',DEFAULT=>$old{'id'}),
    ]
  );
}

=head4 add_or_drop()

Takes one argument, 'add', 'drop', or 'reinstate'.

=cut

sub add_or_drop {
  my $self = shift;
  my $action = shift;
  local $words_prefix = "b.add_or_drop";
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $roster = $self->{STAGE}->{ROSTER};
  my $student=$roster->selected();
  my $name = $data->key_to_name(KEY=>$student,ORDER=>'firstlast');

  my $did_something = sub {
      $self->is_modified(1);
      $roster->refresh();
  };

  if ($action eq 'drop') {
    if ($student eq "" || !($name =~ m/\w+/)) {return}
         # ...The $name clause shouldn't be necessary, but one time
         # I somehow got this dialog box to come up with null string for the
         # name. Wish I could figure out how to reproduce that, but anyway
         # this check keeps anything bogus from happening.
    ExtraGUI::confirm(
                     sprintf(w('confirm'),$name),
                     sub {
                       if (shift) {
                         $gb->drop_student($student);
                         $roster->selected("");
                         &$did_something();
                       }
                     }
    );
  }
  if ($action eq 'add') {
    my $continue_looping = 1;
    while ($continue_looping) {
      $continue_looping = 0;
      my $did_form = 0;
      ExtraGUI::fill_in_form(
        INPUTS => [
          Input->new(KEY=>'last',PROMPT=>w('last'),TYPE=>'string',BLANK_ALLOWED=>0),
          Input->new(KEY=>'first',PROMPT=>w('first'),TYPE=>'string',BLANK_ALLOWED=>0),
          Input->new(KEY=>'id',PROMPT=>w('id'),TYPE=>'string'),
        ],
        TITLE=>w('add_title'),
        INFO=>w('add_students_info'),
        CANCEL_CALLBACK=>sub{$did_form=1},
        CALLBACK=>sub {
                       my $results = shift; # hash ref
                       my ($key,$err) = $gb->add_student(
                         LAST=>$results->{'last'},
                         FIRST=>$results->{'first'},
                         ID=>$results->{'id'});
                       if ($err ne '') {ExtraGUI::error_message(sprintf w($err),($results->{'first'}.' '.$results->{'last'}))}
                       &$did_something();
                       $continue_looping = 1;
                       $did_form = 1;
                     },
        XOFFSET=>300,
        YOFFSET=>30,
      );
      Browser::main_window->waitVariable(\$did_form);
    }
  }
  if ($action eq 'reinstate') {

    # Currently, there's no way to select a dropped student from the regular menu, but
    # I may add that in the future. So right here, I should check if the selected student
    # is a dropped student, and if she is, select her automatically in the dialog box I make.

    my $who = '';

    my $box = Browser::empty_toplevel_window(w('reinstate_title'));
    my $lb = $box->Scrolled("Listbox",-scrollbars=>"e")
                       ->pack(-side=>'top',-expand=>1,-fill=>'y');

    my @k = $gb->student_keys("dropped");
    my @names = ();
    foreach my $key(@k) {
      push @names,$data->key_to_name(KEY=>$key);
    }
    $lb->insert('end',@names);
    $lb->bind('<Button-1>',
      sub {
        $who = $k[$lb->curselection()];
      }
    );
    $box->Button(-text=>w("ok"),-command=>sub{
      $box->destroy();
      $gb->reinstate_student($who);
      &$did_something();
    })->pack();
    $box->Button(-text=>w("cancel"),-command=>sub{$box->destroy()})->pack();

  }

}

sub edit_category_weights {
  my $self = shift;
  local $words_prefix = "b.edit_category_weights";
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $assignments = $self->{STAGE}->{ASSIGNMENTS};

  if (!$gb->has_categories()) {
    ExtraGUI::error_message(w('no_categories'));
    return;
  }

  if (!$gb->weights_enabled()) {
    ExtraGUI::error_message(w('not_weighted'));
    return;
  }

  my $did_something = sub {
      $self->is_modified(1);
      $assignments->refresh();
  };

  my $categories_ref = $gb->category_array();
  my @categories = @$categories_ref;

  my @inputs = ();
  foreach my $cat(@categories) {
    push @inputs,Input->new(KEY=>$cat,PROMPT=>(sprintf w('weight_for'),$gb->category_name_plural($cat)),
                            TYPE=>'numeric',MIN=>0,DEFAULT=>$gb->category_property($cat,'weight'));
  }

    ExtraGUI::fill_in_form(
      INPUTS => \@inputs,
      COLUMNS=>2,
      TITLE=>w('title'),
      CALLBACK=>sub {
                       local $words_prefix = "edit_category_weights";
                       my $results = shift; # hash ref

                       foreach my $cat(keys %$results) {
                         $gb->category_property($cat,'weight',$results->{$cat});
                       }
                       &$did_something();
                     }
    );

}

sub new_category {
  my $self = shift;
  local $words_prefix = "edit_categories"; # shared with TermUI
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $assignments = $self->{STAGE}->{ASSIGNMENTS};

  my $did_something = sub {
      $self->is_modified(1);
      $assignments->refresh();
  };

  my @inputs = (
        Input->new(KEY=>'key',PROMPT=>w('enter_short_name'),TYPE=>'string',BLANK_ALLOWED=>0),
        Input->new(KEY=>'sing',PROMPT=>w('enter_singular_noun'),TYPE=>'string',BLANK_ALLOWED=>0),
        Input->new(KEY=>'pl',PROMPT=>w('enter_plural_noun'),TYPE=>'string'),
        Input->new(KEY=>'n_drop',PROMPT=>w('enter_number_to_drop'),DEFAULT=>0,TYPE=>'numeric',MIN=>0),
        Input->new(KEY=>'count',PROMPT=>w('b.will_it_count'),DEFAULT=>1,TYPE=>'string',
                   WIDGET_TYPE=>'radio_buttons'),
        Input->new(KEY=>'single',PROMPT=>w('b.is_it_single'),DEFAULT=>0,TYPE=>'string',
                   WIDGET_TYPE=>'radio_buttons'),
        Input->new(KEY=>'max',PROMPT=>w('enter_max'),TYPE=>'numeric',MIN=>0),
      );
    if ($gb->weights_enabled()!=0) {
      push @inputs,Input->new(KEY=>'weight',PROMPT=>w('gimme_weight'),TYPE=>'numeric',MIN=>0);
    }

    ExtraGUI::fill_in_form(
      INPUTS => \@inputs,
      COLUMNS=>1,
      TITLE=>w('add_title'),
      CALLBACK=>sub {
                       local $words_prefix = "edit_categories"; # shared with TermUI
                       my $results = shift; # hash ref
                       my $key = $results->{'key'};
                       # The following only affects the key, not the name the user sees:
                         $key = lc($key);
                         $key =~ s/[^\w]//g;
                       my $sing = $results->{'sing'};
                       my $pl  = $results->{'pl'};
                       my $drop =   $results->{'n_drop'};
                       my $ignored =  !($results->{'count'});
                       my $single = $results->{'single'};
                       my $max =  $results->{'max'};
                       my $w =  $results->{'weight'};
                       my $stuff = "\"catname:$sing,$pl\"";
                       if ($max) {$stuff = $stuff . ",\"max:$max\""}
                       if ($drop>=1) {$stuff = $stuff . ",\"drop:$drop\""}
                       if ($ignored) {$stuff = $stuff . ",\"ignore:true\""}
                       if ($single) {$stuff = $stuff . ",\"single:true\""}
                       if ($w ne "") {$stuff = $stuff . ",\"weight:$w\""}
                       if ($gb->category_exists($key)) {ExtraGUI::error_message(w('category_exists')); return}
                       $gb->add_category($key,$stuff);
                       &$did_something();
                     }
    );

}


sub negate_binary {
  my $x = shift;
  if ($x) {return 0} else {return 1}
}

=head4 edit_category()

This lets you edit anything about the category except for the weights,
which is what edit_category_weights() is for. The "ignore" and "max"
parameters of a category are the defaults for its assignments. If the
user changes these, she'll be prompted as to whether to propagate the
new value down to all the assignments in the category.

=cut

sub edit_category {
  my $self = shift;
  local $words_prefix = "edit_categories"; # shared with TermUI
  $self->grades_queue();
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $assignments = $self->{STAGE}->{ASSIGNMENTS};
  my $roster = $self->{STAGE}->{ROSTER};
  my $cat = "";
  if ($data->file_is_open() && exists $self->{STAGE} && exists $self->{STAGE}->{ASSIGNMENTS}) {
    $cat=$self->{STAGE}->{ASSIGNMENTS}->get_active_category();
  };
  if ($cat eq '') {return}


  my %old = (
    'sing'=>$gb->category_name_singular($cat),
    'pl'=>$gb->category_name_plural($cat),
    'drop'=>$gb->category_property($cat,'drop'),
    'max'=>$gb->category_property($cat,'max'),
    'ignore'=>$gb->category_property_boolean($cat,'ignore'),
    'single'=>$gb->category_property_boolean($cat,'single'),
  );
  if ($old{'drop'} eq '') {$old{'drop'} = 0}
  if ($old{'ignore'} eq '') {$old{'ignore'} = 0}
  if ($old{'single'} eq '') {$old{'single'} = 0}

  my $callback = sub {
    local $words_prefix = "edit_categories"; # shared with TermUI
    my $new = shift;
    my $did_something = 0;
    my %changed;
    my $changed_heritable = 0;
    foreach my $what(keys %old) {
      if (!(exists $old{$what} && $old{$what} eq $new->{$what})) {
        $did_something = 1;
        $changed{$what} = 1;
        if ($what eq 'ignore' || $what eq 'max') {$changed_heritable=1}
      }
    }
    if ($did_something) {
      my $sing = $new->{'sing'};
      my $pl  = $new->{'pl'};
      my $drop =   $new->{'drop'};
      my $ignore =   $new->{'ignore'};
      my $max =  $new->{'max'};
      my $single = $old{'single'};
      if (exists $new->{'single'}) {$single = $new->{'single'}}

      $sing =~ s/\"/\'/g; # Otherwise we can't embed it in quotes in the gradebook file.
      $pl =~ s/\"/\'/g;   # ditto.

      my $stuff = "\"catname:$sing,$pl\"";
      if ($max ne '') {$stuff = $stuff . ",\"max:$max\""}
      if ($drop>=1) {$stuff = $stuff . ",\"drop:$drop\""}
      if ($ignore) {$stuff = $stuff . ",\"ignore:true\""}
      if ($single) {$stuff = $stuff . ",\"single:true\""}
      my $w = $gb->category_property($cat,'weight');
      if ($w ne "") {$stuff = $stuff . ",\"weight:$w\""}
      $gb->set_all_category_properties($cat,$stuff);
      $self->is_modified(1);
      $roster->refresh();
      $self->{STAGE}->{ASSIGNMENTS}->refresh(); # necesary if this caused names of assmts to change

      if ($changed_heritable && $gb->number_of_assignments_in_category($cat)>0) {
        my @inputs = ();
        if ($changed{'ignore'}) {
          my $ignore_prompt = w('propagate_ignore');
          if (!$ignore) {$ignore_prompt=w('propagate_not_ignore')}
          push @inputs,Input->new(KEY=>'ignore',PROMPT=>$ignore_prompt,TYPE=>'string',DEFAULT=>1,
                   WIDGET_TYPE=>'radio_buttons');
        }
        if ($changed{'max'}) {
          push @inputs,Input->new(KEY=>'max',PROMPT=>w('propagate_max'),TYPE=>'string',DEFAULT=>1,
                   WIDGET_TYPE=>'radio_buttons');
        }
        ExtraGUI::fill_in_form(
          INPUTS => \@inputs,
          COLUMNS=>1,
          CALLBACK=>sub {
                      local $words_prefix = "edit_categories"; # shared with TermUI
                      my $results = shift;
                      my $assignments = $gb->array_of_assignments_in_category($cat);
                      my @things = ('ignore','max');
                      foreach my $ass(@$assignments) {
                        foreach my $thing(@things) {
                          if ($results->{$thing}) {
                            my $props = $gb->assignment_properties("$cat.$ass");
                            my $change_to = $new->{$thing};
                            if ($thing eq 'ignore') {$change_to = ('false','true')[$change_to]}
                            $props->{$thing} = $change_to;
                            $gb->assignment_properties("$cat.$ass",$props);
                          }
                        }
                      }
                    }
        );
      }
    };
  };


  my @inputs = (
        Input->new(KEY=>'sing',PROMPT=>w('enter_singular_noun'),TYPE=>'string',BLANK_ALLOWED=>0,
                        DEFAULT=>$gb->category_name_singular($cat)),
        Input->new(KEY=>'pl',PROMPT=>w('enter_plural_noun'),TYPE=>'string',
                        DEFAULT=>$gb->category_name_plural($cat)),
        Input->new(KEY=>'drop',PROMPT=>w('enter_number_to_drop'),DEFAULT=>$old{'drop'},
                                           TYPE=>'numeric',MIN=>0),
        Input->new(KEY=>'max',PROMPT=>w('enter_max'),DEFAULT=>$old{'max'},
                                           TYPE=>'numeric',MIN=>0,BLANK_ALLOWED=>1),
        Input->new(KEY=>'ignore',PROMPT=>w('b.is_it_ignored'),DEFAULT=>$old{'ignore'},
                   TYPE=>'string',WIDGET_TYPE=>'radio_buttons'),
      );

    if ($gb->number_of_assignments_in_category($cat)<2) {
      push @inputs,
        Input->new(KEY=>'single',PROMPT=>w('b.is_it_single'),DEFAULT=>$old{'single'},TYPE=>'string',
                   WIDGET_TYPE=>'radio_buttons'),
    }

    ExtraGUI::fill_in_form(
      INPUTS => \@inputs,
      COLUMNS=>1,
      TITLE=>w('edit_category_dialog_title'),
      CALLBACK=>$callback
    );

}





#------------ End public methods --------------

=head4 menu_bar()

This private method adds the menu bar to the parent window.

=cut

sub menu_bar {
  my $self = shift;
  my $parent = shift;
  local $words_prefix = "b.menus";

  $self->{MENU_ITEM_DIRECTORY} = {};

  my $add_item = sub {
    my $menu = shift;
    my $items = shift;
    my $key = shift;
    my $text = shift;
    my $cmd = shift;
    if ($key eq '-') {
      my @stuff = (Separator=>'');
      push @$items,\@stuff;
    }
    else {
      if ($text eq "") {$text = w($key)}
      my @stuff = ('command'=>$text,-command=>$cmd); #see Mastering Perl/Tk p. 262; command tells what type (command, not radiobutton,...)
      if (@_) {push @stuff,(-accelerator=>shift)}
      push @$items,\@stuff;
      $self->{MENU_ITEM_DIRECTORY}->{$menu."*".$key} = $#$items;
    }
  };

  my $bind_accel = sub {
    $parent->bind('<Control-Key-'.lc(accel($_[0])).'>'=>$_[1]);
    $parent->bind('<Control-Key-'.uc(accel($_[0])).'>'=>$_[1]);
  };

  # Flush the grades queue whenever the user selects a menu item:
  my $safe = sub {$self->grades_queue()};
  $self->{SAFE} = $safe;

  my $bar = $parent->Frame(-relief=>'raised',-borderwidth=>'2');

  my @items = ();


  &$add_item("FILE_MENUB",\@items,'new',  '',sub{&$safe(); $self->new_file()},  'Ctrl+'.accel('fn'));
  &$add_item("FILE_MENUB",\@items,'open', '',sub{&$safe(); $self->open_file()}, 'Ctrl+'.accel('fo'));
  &$add_item("FILE_MENUB",\@items,'save', '',sub{&$safe(); $self->save_file()}, 'Ctrl+'.accel('fs'));
  &$add_item("FILE_MENUB",\@items,'close','',sub{&$safe(); $self->close_file()},'Ctrl+'.accel('fw'));
  &$add_item("FILE_MENUB",\@items,'quit', '',sub{&$safe(); $self->quit()},      'Ctrl+'.accel('fq'));
  &$add_item("FILE_MENUB",\@items,'-');
  &$add_item("FILE_MENUB",\@items,'about','',sub{$self->about()});
  &$add_item("FILE_MENUB",\@items,'-');
  &$add_item("FILE_MENUB",\@items,'rekey','',sub{&$safe(); $self->rekey_file()});
  &$add_item("FILE_MENUB",\@items,'strip_watermark','',sub{&$safe(); $self->strip_watermark()});
  &$add_item("FILE_MENUB",\@items,'reconcile','',sub{&$safe(); $self->reconcile()});
  &$add_item("FILE_MENUB",\@items,'standards','',sub{&$safe(); $self->options('standards')});
  &$add_item("FILE_MENUB",\@items,'clone','',sub{&$safe(); $self->clone()});
  &$add_item("FILE_MENUB",\@items,'-');
  my $prefs = Preferences->new(); # Can't use the one associated with the gradebook, because may not have one open.
  my $recent_files = $prefs->get('recent_files');
  $self->{RECENT_FILE_LABELS} = [];
  my @prepared_list = $self->prepare_list_of_recent_files_for_menu();
  my $l = $self->{RECENT_FILE_LABELS};
  foreach my $stuff(@prepared_list) {
    my ($foo,$label,$recent_file,$sub) = @$stuff; # $foo is something like "recent_file3" for the third file -- do I ever even use this?
    &$add_item("FILE_MENUB",\@items,$foo,$label,$sub);
    push @$l,$label;
  }
  $self->{LABEL_OF_CLEAR_RECENT} = w('clear_recent');
  &$add_item("FILE_MENUB",\@items,'clear_recent','',
             sub {
               my $prefs = Preferences->new();
               $prefs->set('recent_files','');
               $self->clear_recent_files_from_menu();
               $self->enable_and_disable_menu_items();
             }
  );
  $self->{FILE_MENUB}
   = $bar->Menubutton(-text=>w("file"),-menuitems=>\@items,-tearoff=>0)->pack(-side=>'left');

  &$bind_accel('fn',sub{&$safe; $self->new_file()});
  &$bind_accel('fo',sub{&$safe; $self->open_file()});
  &$bind_accel('fs',sub{&$safe; $self->save_file()});
  &$bind_accel('fw',sub{&$safe; $self->close_file()});
  &$bind_accel('fq',sub{&$safe; $self->quit()});

  @items = ();
  &$add_item("PREFERENCES_MENUB",\@items,'beep','',sub{$self->options('beep')});
  &$add_item("PREFERENCES_MENUB",\@items,'justify','',sub{$self->options('justify')});
  $self->{PREFERENCES_MENUB}
   = $bar->Menubutton(-text=>w("preferences"),-menuitems=>\@items,-tearoff=>0)->pack(-side=>'left');


  @items = ();
  &$add_item("STUDENTS_MENUB",\@items,'add','',sub{&$safe(); $self->add_or_drop('add')});
  &$add_item("STUDENTS_MENUB",\@items,'reinstate','',sub{&$safe(); $self->add_or_drop('reinstate')});
  &$add_item("STUDENTS_MENUB",\@items,'edit',w('edit_disabled'),sub{&$safe(); $self->edit_student()});
  &$add_item("STUDENTS_MENUB",\@items,'drop',(sprintf w('drop'),''),sub{&$safe(); $self->add_or_drop('drop')});
  $self->{STUDENTS_MENUB}
   = $bar->Menubutton(-text=>w("students"),-menuitems=>\@items,-tearoff=>0)->pack(-side=>'left');

  @items = ();
  &$add_item("ASSIGNMENTS_MENUB",\@items,'new_category','',sub{&$safe(); $self->new_category()});
  &$add_item("ASSIGNMENTS_MENUB",\@items,
             'edit_category',w('edit_category_blank'),sub{&$safe(); $self->edit_category()});
  &$add_item("ASSIGNMENTS_MENUB",\@items,'category_weights','',sub{&$safe(); $self->edit_category_weights()});
  &$add_item("ASSIGNMENTS_MENUB",\@items,'-');
  &$add_item("ASSIGNMENTS_MENUB",\@items,
             'new_assignment',w('new_assignment_blank'),sub{&$safe(); $self->new_assignment()});
  &$add_item("ASSIGNMENTS_MENUB",\@items,
             'edit_assignment',w('edit_assignment_blank'),sub{&$safe(); $self->edit_assignment()});
  &$add_item("ASSIGNMENTS_MENUB",\@items,
             'delete_assignment',w('delete_assignment_blank'),sub{&$safe(); $self->delete_assignment()});
  $self->{ASSIGNMENTS_MENUB}
   = $bar->Menubutton(-text=>w("assignments"),-menuitems=>\@items,-tearoff=>0)->pack(-side=>'left');

  @items = ();
  &$add_item("REPORT_MENUB",\@items,'statistics','',sub{&$safe(); $self->report('stats')});
  &$add_item("REPORT_MENUB",\@items,'statistics_ass',w('statistics_ass_blank'),sub{&$safe(); $self->report('stats_ass')});
  &$add_item("REPORT_MENUB",\@items,'student','',sub{&$safe(); $self->report('student')});
  &$add_item("REPORT_MENUB",\@items,'sort_by_overall','',sub{&$safe(); $self->report('sort_by_overall')});
  &$add_item("REPORT_MENUB",\@items,'sort_by_category',w('sort_by_category_blank'),sub{&$safe(); $self->report('sort_by_category')});
  &$add_item("REPORT_MENUB",\@items,'sort_by_assignment',w('sort_by_assignment_blank'),sub{&$safe(); $self->report('sort_by_assignment')});
  &$add_item("REPORT_MENUB",\@items,'table',w('table'),sub{&$safe(); $self->report('table')});
  &$add_item("REPORT_MENUB",\@items,'roster',w('roster'),sub{&$safe(); $self->report('roster')});
  $self->{REPORT_MENUB}
   = $bar->Menubutton(-text=>w("report"),-menuitems=>\@items,-tearoff=>0)->pack(-side=>'left');

  @items = ();
  &$add_item("SERVER_MENUB",\@items,'settings','',sub{&$safe(); $self->options('server')});
  &$add_item("SERVER_MENUB",\@items,'class_description','',sub{&$safe(); $self->server('class_description')});
  &$add_item("SERVER_MENUB",\@items,'roster','',sub{&$safe(); $self->server('roster')});
  &$add_item("SERVER_MENUB",\@items,'list_work','',sub{&$safe(); $self->server('list_work')});
  &$add_item("SERVER_MENUB",\@items,'upload','',sub{&$safe(); $self->server('upload')});
  &$add_item("SERVER_MENUB",\@items,'-');
  &$add_item("SERVER_MENUB",\@items,'email','',sub{&$safe(); $self->server('email')});
  &$add_item("SERVER_MENUB",\@items,'emailone','',sub{&$safe(); $self->server('emailone',$self->{ROSTER}->selected())});
  &$add_item("SERVER_MENUB",\@items,'post_message','',sub{&$safe(); $self->server('post_message')});
  &$add_item("SERVER_MENUB",\@items,'post_message_one','',sub{&$safe(); $self->server('post_message_one',$self->{ROSTER}->selected())});
  $self->{SERVER_MENUB}
   = $bar->Menubutton(-text=>w("server"),-menuitems=>\@items,-tearoff=>0)->pack(-side=>'left');

  $self->enable_and_disable_menu_items();
  return $bar;
}

sub clear_recent_files_from_menu {
  my $self = shift;
  my $menu = $self->{FILE_MENUB}->menu();
  my $l = $self->{RECENT_FILE_LABELS};
  foreach my $label(@$l) {
    $menu->delete($label);
	}
  @$l = ();
}

sub about {
  local $words_prefix = "b.about";
  my $text = w('about_og');
  my $version = Version::version();
  $text = sprintf $text,$version;
        ExtraGUI::show_text(TITLE=>w('about_og_title'),TEXT=>$text,WIDTH=>100);
}

# Set the accelerator buttons for the menus:
sub accel {
  my $what = shift; # fs for file menu/save, etc.
  return {
    'fs'=>'S',    'fo'=>'O',    'fn'=>'N',    'fw'=>'W',    'fq'=>'Q'
  }
  ->{$what};
}

=head4 enable_and_disable_menu_items()

Looks at whether a file is open, whether a student is selected, etc., and enables and
disables the menu items appropriately. This doesn't have any effect on things like
control-O for open, but that's ok, because, e.g., open_file() automatically closes
the file anyway before opening a new one.

=cut

sub enable_and_disable_menu_items {
  my $self = shift;
  my $data = $self->{DATA};
  local $words_prefix = "b.menus";
  my $dir = $self->{MENU_ITEM_DIRECTORY};


  my $assignments_menub = $self->{ASSIGNMENTS_MENUB};
  my $cat = "";
  my $ass = "";
  if ($data->file_is_open() && exists $self->{STAGE} && exists $self->{STAGE}->{ASSIGNMENTS}) {
    $cat=$self->{STAGE}->{ASSIGNMENTS}->get_active_category();
    $ass=$self->{STAGE}->{ASSIGNMENTS}->selected();
    $ass =~ s/\.(.*)//;
    $ass = $1;
  };


  # File menu.
  my $file_menub = $self->{FILE_MENUB};
  if ($data->file_is_open()) {
    $file_menub->entryconfigure($dir->{'FILE_MENUB*new'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*open'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*close'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*save'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*rekey'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*strip_watermark'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*reconcile'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*standards'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*clone'},-state=>'normal');
  }
  else { # no file open
    $file_menub->entryconfigure($dir->{'FILE_MENUB*new'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*open'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*close'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*save'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*rekey'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*strip_watermark'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*reconcile'},-state=>'normal');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*standards'},-state=>'disabled');
    $file_menub->entryconfigure($dir->{'FILE_MENUB*clone'},-state=>'disabled');
  }
  $self->enable_or_disable_specific_menu_item('FILE','clear_recent');

  # Students menu:
  my $students_menub = $self->{STUDENTS_MENUB};
  my $student = "";
  if ($data->file_is_open() && exists $self->{STAGE} && exists $self->{STAGE}->{ROSTER}) {
    $student=$self->{STAGE}->{ROSTER}->selected();
  };
  if ($student ne "" && $data->file_is_open()) {
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*drop'},
                       -state=>'normal',
                       -label=>(sprintf w('drop'),$data->key_to_name(KEY=>$student,ORDER=>'firstlast')));
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*edit'},
                       -state=>'normal',
                       -label=>(sprintf w('edit'),$data->key_to_name(KEY=>$student,ORDER=>'firstlast')));
  }  
  else { # no student selected
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*drop'},-state=>'disabled');
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*edit'},-state=>'disabled',-label=>w('edit_disabled'));
  }  
  if ($data->file_is_open()) {
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*add'},-state=>'normal');
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*reinstate'},-state=>'normal');
  }
  else {
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*add'},-state=>'disabled');
    $students_menub->entryconfigure($dir->{'STUDENTS_MENUB*reinstate'},-state=>'disabled');
  }

  # Assignments menu:
  my $cat_selected = $cat ne "" && $data->file_is_open(); # 2nd clause is for when we're in the process of quitting
  my $ass_selected = $cat_selected && $ass ne "";
  if ($cat_selected) {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*new_assignment'},
                       -state=>'normal',
                       -label=>(sprintf w('new_assignment'),
                                ucfirst($data->{GB}->category_name_singular($cat))));
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*edit_category'},
                       -state=>'normal',
                       -label=>(sprintf w('edit_category'),
                                ucfirst($data->{GB}->category_name_plural($cat))));
  }
  else { # no category selected
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*new_assignment'},-state=>'disabled');
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*edit_category'},
                       -state=>'disabled',-label=>w('edit_category_blank'));
  }
  if ($ass_selected) {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*edit_assignment'},
                       -state=>'normal',
                       -label=>(sprintf w('edit_assignment'),
                                ucfirst($data->{GB}->assignment_name($cat,$ass))));
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*delete_assignment'},
                       -state=>'normal',
                       -label=>(sprintf w('delete_assignment'),
                                ucfirst($data->{GB}->assignment_name($cat,$ass))));
  }
  else { # no assignment selected
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*edit_assignment'},-state=>'disabled',-label=>w('edit_assignment_blank'));
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*delete_assignment'},-state=>'disabled',-label=>w('delete_assignment_blank'));
  }
  if ($data->file_is_open() && $data->{GB}->weights_enabled() && $data->{GB}->has_categories()) {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*category_weights'},-state=>'normal')
  }
  else {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*category_weights'},-state=>'disabled')
  }
  if ($data->file_is_open()) {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*new_category'},-state=>'normal')
  }
  else {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*new_category'},-state=>'disabled')
  }

  # Reports menu.
  my $report_menub = $self->{REPORT_MENUB};
  if ($cat ne "" && $data->file_is_open()) { # 2nd clause is for when we're in the process of quitting
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*sort_by_category'},
                       -state=>'normal',
                       -label=>(sprintf w('sort_by_category'),
                                ucfirst($data->{GB}->category_name_plural($cat))));
  }  
  else { # no category selected
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*sort_by_category'},
                                            -label=>w('sort_by_category_blank'),
                                             -state=>'disabled');
  }  
  if ($cat ne "" && $ass ne "" && $data->file_is_open()) { # 2nd clause is for when we're in the process of quitting
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*sort_by_assignment'},
                       -state=>'normal',
                       -label=>(sprintf w('sort_by_assignment'),
                                ucfirst($data->{GB}->assignment_name($cat,$ass))));
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*statistics_ass'},
                       -state=>'normal',
                       -label=>(sprintf w('statistics_ass'),
                                ucfirst($data->{GB}->assignment_name($cat,$ass))));
  }  
  else { # no assignment selected
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*sort_by_assignment'},
                     -label=>w('sort_by_assignment_blank'),
                     -state=>'disabled');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*statistics_ass'},
                     -label=>w('statistics_ass_blank'),
                     -state=>'disabled');
  }  
  if ($data->file_is_open()) {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*new_category'},-state=>'normal')
  }
  else {
    $assignments_menub->entryconfigure($dir->{'ASSIGNMENTS_MENUB*new_category'},-state=>'disabled')
  }

  if ($student ne "") {
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*student'},-state=>'normal');
  }  
  else { # no student selected
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*student'},-state=>'disabled');
  }  
  if ($data->file_is_open()) {
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*statistics'},-state=>'normal');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*sort_by_overall'},-state=>'normal');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*table'},-state=>'normal');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*roster'},-state=>'normal');
  }
  else {
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*statistics'},-state=>'disabled');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*sort_by_overall'},-state=>'disabled');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*table'},-state=>'disabled');
    $report_menub->entryconfigure($dir->{'REPORT_MENUB*roster'},-state=>'disabled');
  }

  # Server menu.
  my $server_menub = $self->{SERVER_MENUB};
  if ($data->file_is_open() && $student ne "") {
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*emailone'},-state=>'normal',
                       -label=>(sprintf w('emailonename'),$data->key_to_name(KEY=>$student,ORDER=>'firstlast')));
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*post_message_one'},-state=>'normal',
                       -label=>(sprintf w('post_message_name'),$data->key_to_name(KEY=>$student,ORDER=>'firstlast')));
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*sent_messages'},-state=>'normal',
                       -label=>(sprintf w('sent_messages_name'),$data->key_to_name(KEY=>$student,ORDER=>'firstlast')));
  }
  else { # no student selected
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*emailone'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*post_message_one'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*sent_messages'},-state=>'disabled');
  }
  if ($data->file_is_open()) {
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*settings'},-state=>'normal');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*class_description'},-state=>'normal');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*roster'},-state=>'normal');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*list_work'},-state=>'normal');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*upload'},-state=>'normal');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*email'},-state=>'normal');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*post_message'},-state=>'normal');
  }
  else {
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*settings'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*class_description'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*roster'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*list_work'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*upload'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*email'},-state=>'disabled');
    $server_menub->entryconfigure($dir->{'SERVER_MENUB*post_message'},-state=>'disabled');
  }
}

=head4 enable_or_disable_specific_menu_item()

Call with name of menu (e.g., FILE) and item (e.g. clear_recent). This is not implemented for every
combination of inputs. The idea is that sometimes we just want to tweak one item, and for efficiency,
we don't want to call enable_and_disable_menu_items(), which does every single menu. In cases like that,
enable_and_disable_menu_items() is coded so that it actually calls this routine.

=cut

sub enable_or_disable_specific_menu_item {
  my $self = shift;
  my $menu = shift;
  my $item = shift;
  my $dir = $self->{MENU_ITEM_DIRECTORY};
  if ($menu eq 'FILE') {
    my $file_menub = $self->{FILE_MENUB};
    if ($item eq 'clear_recent') {
      my $index = $self->{FILE_MENUB}->menu()->index($self->{LABEL_OF_CLEAR_RECENT});
      if (split(',',Preferences->new()->get('recent_files'))) { # We have a list of recent files.
        $file_menub->entryconfigure($index,-state=>'normal');
      }
      else {
        $file_menub->entryconfigure($index,-state=>'disabled');
      }
	  }
  }
}


#---------------------------------------------------
# Stage class
#---------------------------------------------------

=head3 Stage class

This is the ``center-stage'' area of the window, where the user
selects students and edits their grades on assignments.
The object's basic job is just to pass messages
among its parts.
Its children, Roster, Grades, and Assignments,
let it know when they change the state of the GUI. For instance,
when the Roster object sees that the user has selected a
certain student, it calls Stage::roster_has_set_student(),
which then gives the relevant information to Grades.
We avoid endless loops, because Stage knows that when it
gets a call to this method, the call came from Roster, and
therefore it doesn't need to call Roster to let it know about
the change.

=cut

package Stage;

sub new {
  my $class = shift;
  my $browser_window = shift;
  my $parent = shift;
  my $data = shift;
  my $self = {};
  bless($self,$class);
  $self->{BROWSER_WINDOW} = $browser_window;
  $self->{PARENT} = $parent;
  $self->{DATA} = $data;
  $self->{FRAME} = $parent->Frame();
  $browser_window->{ASSIGNMENTS} = Assignments->new($self,$self->{FRAME},$data);
  $browser_window->{ROSTER} = Roster->new($self,$self->{FRAME},$data);
  $self->{ROSTER} = $browser_window->{ROSTER};
  $self->{ASSIGNMENTS} = $browser_window->{ASSIGNMENTS};
  $self->{ASSIGNMENT} = "";
  $self->{STUDENT} = "";
  return $self;
}


sub roster_has_set_student {
  my $self = shift;
  my $k = shift; # key of student
  if (!$self->{DATA}->file_is_open()) {return}
  $self->{BROWSER_WINDOW}->grades_queue(); # flush
  $self->{STUDENT} = $k;
  $self->{BROWSER_WINDOW}->enable_and_disable_menu_items();
}

=head4 assignments_has_set_assignment()

Enables and disables menu items, then calls set_assignment().

=cut

sub assignments_has_set_assignment {
  my $self = shift;
  my $k = shift; # key of assignment
  if (!$self->{DATA}->file_is_open()) {return}
  $self->{BROWSER_WINDOW}->grades_queue(); # flush
  $self->{BROWSER_WINDOW}->enable_and_disable_menu_items();
  $self->{ROSTER}->set_assignment($k);
  $self->{ASSIGNMENT} = $k;
}

=head4 grades_has_set_grade()

Simply passes the message to the BrowserData object via its
set_a_grade() method.

=cut

sub grades_has_set_grade {
  my $self = shift;
  my $s_key = shift;
  my $c_key = shift;
  my $a_key = shift;
  my $grade = shift;
  #print "setting $a_key,$s_key,$grade in grades_has_set_grade\n";
  $self->{DATA}->set_a_grade($c_key.".".$a_key,$s_key,$grade);
}

#---------------------------------------------------
# Assignments class
#---------------------------------------------------

=head3 Assignments class

This class encapsulates all the data about the assignments
displayed on the screen.

=cut

package Assignments;

sub new {
  my $class = shift;
  my $stage = shift;
  my $parent = shift;
  my $data = shift;
  my $self = {};
  bless($self,$class);
  $self->{STAGE} = $stage;
  $self->{PARENT} = $parent;
  $self->{DATA} = $data;
  $self->{FRAME} = $parent->Frame()->pack(-side=>'left',-expand=>1,-padx=>10);
  $self->{CATEGORIES_TEXTVARIABLE} = '';
  $self->{CATEGORIES_MENU} = $self->{FRAME}->Optionmenu(
         -takefocus=>0,
         -font=>ExtraGUI::font('plain'),
         -textvariable=>\($self->{CATEGORIES_TEXTVARIABLE}),
  )->pack(-side=>'top',-expand=>1);
  $self->{ASSIGNMENTS_LISTBOX} = $self->{FRAME}->Scrolled(
         "Listbox",
         -scrollbars=>"e",
         -takefocus=>0,
         -font=>ExtraGUI::font('plain'),
         -height=>25
  )->pack(-side=>'top',-expand=>1);
  $self->{ASSIGNMENTS_LISTBOX}->Subwidget('yscrollbar')->configure(-takefocus=>0);
  $self->{KEYS} = [];                # ref to an array, which is stored in the order shown on the screen
  $self->{NAMES} = [];                # in the same order as KEYS, but shows the assmts in human-readable form
  $self->{CAT_INTENTIONALLY_SELECTED} = ''; # as opposed to {SELECTED}, which may be set to a default because the user hasn't selected a category yet
  $self->{LEAVE_CAT_SELECTION_ALONE} = 0; # if set to 1, prevents selection of wrong cat when menu is rebuilt
  $self->{INFINITE_LOOP_PROTECTION} = 0; # see clicked_on_category()
  return $self;
}

sub refresh {
  my $self = shift;
  $self->clear();
  my $data = $self->{DATA};
  # If a file is open, fill it back up:
  if ($data->file_is_open()) {
    $self->refresh_categories();
    $self->refresh_assignments();
  }
}

=head4 refresh_categories()

Rebuild the popup menu of categories from scratch. If nothing is selected, we
select the first category. Sets up a callback from the menu to decode_cats_option_menu().

=cut

sub refresh_categories {
    my $self = shift;
    my $data = $self->{DATA};
    my $gb = $data->{GB};
    my $menu = $self->{CATEGORIES_MENU}; # a Tk::Optionmenu object
    my @c = ();
    my @names = ();
    if ($gb) {
      my $aa = $gb->category_array();
      @c = @$aa;
      $self->{CAT_KEYS} = \@c;
      foreach my $cat(@c) {
        push @names,$gb->category_name_plural($cat);
      }
      my $save_intent = $self->{CAT_INTENTIONALLY_SELECTED};
      $self->{LEAVE_CAT_SELECTION_ALONE} = ($save_intent ne '');
      $menu->configure(-options=>\@names,-command=>sub{decode_cats_option_menu($self,\@_,\@names,\@c)});
         # The arrays @names and @c are closure-ized at this point.
      $self->{CAT_INTENTIONALLY_SELECTED} = $save_intent;
      $self->{LEAVE_CAT_SELECTION_ALONE} = 0;
      if (@c) {
        my $active = $self->get_active_category();
        if ($active ne '') {
          $self->clicked_on_category($active);
        }
        else {
          my $cat = $self->{CAT_INTENTIONALLY_SELECTED};
          if ($cat eq '') {
            # An Optionmenu always has something selected, so we have to say the first item is selected:
            $cat = $c[0];
          }
          $self->clicked_on_category($cat);
        }
        $menu->configure(-state=>'normal');
      }
	  }
    else {
      #print "no gradebook open in refresh_categories\n";
      $menu->configure(-state=>'disabled'); # doesn't seem to work
      $menu->configure(-options=>[]); # doesn't seem to work
    }
}

=head4 decode_cats_option_menu()

This is the callback that gets called when the user selects a category from the categories
menu. Since all we get back from Tk is the actual text of the menu item, it takes some work
to figure out the corresponding database key. Calls clicked_on_category() with the result.

=cut

sub decode_cats_option_menu {
  my $self = shift;
  my ($args_ref,$names_ref,$keys_ref) = @_;
  my @args = @$args_ref;
  my @names = @$names_ref;
  my @keys = @$keys_ref;
  my $name = pop @args;
  if (!($self->{LEAVE_CAT_SELECTION_ALONE})) {
    for (my $j=0; $j<=$#names; $j++) {
      if ($names[$j] eq $name) {
        $self->clicked_on_category($keys[$j]);
      }
    }
  }
  else {
    my $cat = $self->{CAT_INTENTIONALLY_SELECTED};
    $self->clicked_on_category($cat);
  }
}

# This is the same as the name of another routine in another package.
sub refresh_assignments {
  my $self = shift;
  my $data = $self->{DATA};
  my $gb = $data->{GB};
  my $ass_lb = $self->{ASSIGNMENTS_LISTBOX};
  my @names = ();
  my $active_cat = $self->get_active_category();
  #print "in refresh_assignments, =$active_cat=\n";
  $self->clear_assignments();
  if ($active_cat ne "") {
      my $aa = $gb->array_of_assignments_in_category($active_cat);
      my @assignments_in_cat = @$aa;
      $self->{ASS_KEYS} = \@assignments_in_cat;
      foreach my $ass(@assignments_in_cat) {
        push @names,$gb->assignment_name($active_cat,$ass);
      }
      $ass_lb->insert('end',@names);
      $ass_lb->bind(
          '<Button-1>',
          sub{$self->clicked_on_assignment($self->{ASS_KEYS}->[$ass_lb->curselection()])}
      );
  }
}


=head4 clear()

This gets called by refresh(). Don't call it directly and expect the GUI
to get redrawn.

=cut

sub clear {
  my $self = shift;
  $self->{KEYS} = [];
  $self->{NAMES} = [];

  # Clear categories.
  $self->{CAT_KEYS} = [];
  $self->{NAMES} = [];
  $self->selected("");
  $self->{CATEGORIES_MENU}->configure(-options=>[''],-state=>'disabled');

  # Clear assignments.
  my $ass_lb = $self->{ASSIGNMENTS_LISTBOX};
  $ass_lb->delete(0,'end'); # Delete all of them.
}


sub clear_assignments {
  my $self = shift;
  $self->{ASS_KEYS} = [];
  $self->{NAMES} = [];
  $self->selected($self->get_active_category());
  my $ass_lb = $self->{ASSIGNMENTS_LISTBOX};
  $ass_lb->delete(0,'end'); # Delete all of them.
}



=head4 get_active_category()

Get the key of the category that is currently selected. Calls
selected() and returns the category portion of the key.

=cut

sub get_active_category {
  my $self = shift;
  my $key = $self->selected();
  #print "in get_active_category, =$key=\n";
  if ($key eq "") {return ""}
  if ($key =~ m/^[^\.]+$/) {return $key}
  return GradeBook::first_part_of_label($key);
}

=head4 selected()

Get or set the key of the category or assignment that is currently selected.

=cut

sub selected {
  my $self = shift;
  if (@_) {
    my $what = shift;
    $self->{SELECTED}=$what;
  	if ($what eq '') {
      $what = $self->{CAT_INTENTIONALLY_SELECTED};
    }
    $self->{STAGE}->assignments_has_set_assignment($self->{SELECTED});
  }
  return $self->{SELECTED};
}

=head4 specific_assignment_selected()

Has a specific assignment been selected?

=cut

sub specific_assignment_selected {
  my $self = shift;
  my $key = $self->selected();
  if ($key eq "") {return ""}
  return GradeBook::second_part_of_label($key);
}

=head4 clicked_on_assignment()

Takes one argument, which is the assignment part of the key. Calls selected().

=cut

sub clicked_on_assignment {
  my $self = shift;
  my $key = shift;
  #print "clicked_on_assignment, =$key=\n";
  $self->selected($self->get_active_category().".".$key);
}

=head4 clicked_on_category()

Call selected() and refresh_assignments().

=cut 

sub clicked_on_category {
  my $self = shift;
  my $key = shift;
  $self->selected($key);
  $self->refresh_assignments();
  $self->{STAGE}->{BROWSER_WINDOW}->enable_and_disable_menu_items();
  $self->{CAT_INTENTIONALLY_SELECTED} = $key;

  my $cats_menu = $self->{CATEGORIES_MENU};
  my $gb = $self->{DATA}->{GB};
  my $cat_plural = $gb->category_name_plural($key);
  if ($self->{CATEGORIES_TEXTVARIABLE} ne $cat_plural) {
    # We get into this state when rebuilding the categories menu. Tk selects the first category by default, but that's no what we want.
    if (!($self->{INFINITE_LOOP_PROTECTION})) {
      $self->{INFINITE_LOOP_PROTECTION} = 1; # Avoid infinite recursion
      $cats_menu->setOption($key); # This could cause infinite recursion
    }
  }
  $self->{INFINITE_LOOP_PROTECTION} = 0;
}


#---------------------------------------------------
# Roster class
#---------------------------------------------------

=head3 Roster class

This class encapsulates all the data about the roster
displayed on the screen.

=cut

package Roster;

sub w{return Browser::w(@_)}

sub new {
  my $class = shift;
  my $stage = shift;
  my $parent = shift;
  my $data = shift;
  my $self = {};
  bless($self,$class);

  $self->{KEYS} = [];                # ref to an array, which is stored in the order shown on the screen
  $self->{NAMES} = [];                # in the same order as KEYS, but shows the names in human-readable form
  $self->{KBSEL} = '';
  $self->{KBCTRL} = 0;                # is the control key currently being held down?
  $self->{SELECTED_STUDENT} = '';
  $self->{ASSIGNMENT} = '';
  $self->{OPTION_VAR} = '1-overall';

  $self->{STAGE} = $stage;
  $self->{PARENT} = $parent;
  $self->{DATA} = $data;
  my $f = $parent->Frame()->pack(-side=>'left',-expand=>1,-fill=>'y',-padx=>10);
  my $f2 = $f->Frame()->pack(-side=>'top',-anchor=>'w');
  my $f3 = $f2->Frame()->grid(-row=>0,-column=>0);
  $f3->Label(-text=>'keyboard',-font=>ExtraGUI::font('plain'))->pack(-side=>'left');
  $f3->Label(-textvariable=>\$self->{KBSEL},-font=>ExtraGUI::font('plain'),-relief=>'groove',-width=>8,-anchor=>'w')->pack(-side=>'left');
  $self->{OPTIONS_FRAME} = $f2->Frame()->grid(-row=>0,-column=>1);
  $self->make_options_menu('1-overall','','');
  my $height = $f->screenheight - 200;
  $self->{SPREADSHEET} = $f2->Scrolled(
      "Canvas",
      -scrollbars=>"e",
      -width=>220,
      -height=>$height,
  )->grid(-row=>1,-column=>0,-columnspan=>2);
  $self->{SPREADSHEET}->Subwidget('yscrollbar')->configure(-takefocus=>0);
  $self->{SPREADSHEET}->Subwidget('canvas')->configure(-takefocus=>0);
  return $self;
}

sub make_options_menu {
  my $self = shift;
  my $what = shift;
  my $cat = shift;
  my $ass = shift;
  local $words_prefix = "b.roster_options_menu";
  if (exists $self->{OPTIONS_MENU}) {$self->{OPTIONS_MENU}->destroy()}
  my $gb = $self->{DATA}->{GB};
  my $file_is_open;
  if (ref $gb) {$file_is_open=1} else {$file_is_open=0}
  my @options = ();
  $self->{OPTION_VAR} = '';
  if ($file_is_open) {
    $options[0] = w('overall');
    $self->{OPTION_VAR} = $options[0];
  }
  if ($cat ne '') {
    $options[1] = $gb->category_name_plural($cat);
    if ($ass ne '') {$self->{OPTION_VAR} = $options[1]}
  }
  if ($ass ne '') {
    my $assignment_menu_item = $gb->assignment_name($cat,$ass);
    if ($gb->category_property_boolean($cat,'single')) {
      $assignment_menu_item = $assignment_menu_item." ".w('grade'); # Has to be different from text of menu item for category.
    }
    $options[2] = $assignment_menu_item;
    $self->{OPTION_VAR} = $options[2];
  }

  #print "in make_options_menu, cat=$cat= ass=$ass=\n";
  #print "  ".$self->{OPTION_VAR}."\n";

  @options = reverse @options;
  # Reverse order as a workaround for bug in some Perl/Tk implementations of Optionmenu. In some implementations of Perl/Tk,
  # setting the variable linked to the Optionmenu doesn't allow you to select the item on the menu. In these implementations,
  # the Optionmenu always starts out with the first item selected. As a workaround for this bug, I've reversed the order
  # of the menu, so that the first item is the one that /should/ be selected when the Optionmenu is rebuilt. 
  # This workaround required one other change in the code, which is also marked with a comment containing the text
  # "Reverse order as a workaround..."

  # This has to come /before/ the statement that creates the optionmenu, because the callback to make_scores gets triggered
  # when the optionmenu is first created.
  $self->{OPTIONS} = \@options; 

  # This is a flag that the optionmenu's callback routine checks (see below). Before I made this flag, it was
  # recalculating category totals twice whenever a new category was selected.
  $self->{DO_NOT_MAKE_SCORES} = 1;

  $self->{OPTIONS_MENU} = 
    $self->{OPTIONS_FRAME}->Optionmenu(
         -takefocus=>0,
         -font=>ExtraGUI::font('plain'),
         -options=>\@options,
         -variable=>\$self->{OPTION_VAR},
         -command=>sub{if (!$self->{DO_NOT_MAKE_SCORES}) {$self->make_scores()}}
    )->pack();

  $self->{DO_NOT_MAKE_SCORES} = 0;

  #$self->{OPTIONS_MENU}->setOption($self->{OPTION_VAR}); ...seems to have no effect
}

sub get_assignment_selection_info {
  my $self = shift;
  my $what = $self->type_of_assignment();
  my $cat = '';
  if ($what >= '2-category') {
    $cat = $self->{ASSIGNMENT};
    $cat =~ s/\..*//; # strip everything after the category, e.g. e.2 => e
  }
  my $ass = '';
  if ($what eq '3-assignment') {
    $ass = $self->{ASSIGNMENT};
    $ass =~ m/[^\.]*\.(.*)/;
    $ass = $1;
  }
  return ($what,$cat,$ass);
}

sub set_assignment {
  my $self = shift;
  my $key = shift;
  if ($key eq $self->{ASSIGNMENT}) {return}
  $self->{ASSIGNMENT} = $key;
  my $keys_ref = $self->{KEYS};
  my $n = @$keys_ref;
  my ($cat,$ass) = ('','');
  my $what;
  if ($key eq '') {
    $what = '1-overall';
  }
  else {
    if ($key =~ m/([^\.]*)\.(.*)/) {
      $what = '3-assignment';
      ($cat,$ass) = ($1,$2);
    }
    else {
      $what = '2-category';
      ($cat,$ass) = ($key,'');
    }
  }
  $self->make_options_menu($what,$cat,$ass);
  $self->make_scores($cat,$ass);
}

sub make_scores {
  my $self = shift;
  my ($what,$cat,$ass);
  if (@_) {
    $cat = shift;
    $ass = shift;
    $what =  $self->type_of_assignment();
  }
  else {
    ($what,$cat,$ass) = $self->get_assignment_selection_info();
  }

  my $gb = $self->{DATA}->{GB};
  if (!ref $gb) {return}
  my @k = $gb->student_keys("");
  my $n = @k;
  my $keys_ref = $self->{KEYS};

  my $assignments_in_this_cat;
  if ($what eq '2-category') {
    # For efficiency, compute this now, not once per student:
    $assignments_in_this_cat = $gb->array_of_assignments_in_category($cat);
  }

  for (my $j=0; $j<$n; $j++) { # loop over students
    my $who = $keys_ref->[$j];
    my $f;
    if (exists $self->{SCORE_ENTRIES}->[$j]) {
      $self->{SCORE_ENTRIES}->[$j]->destroy_widget();
      $f = $self->{SCORE_ENTRIES}->[$j]->frame();
    }
    else {
      $f = $self->{SCORE_FRAMES}->[$j];
    }
    my $score_info;
    my $gb = $self->{DATA}->{GB};
    if ($what eq '') {
      $score_info = ''; # shouldn't happen unless we're in an inconsistent state; see comment at bottom of type_of_assignment()
    }
    if ($what eq '1-overall') {
      my $total = Crunch::totals($gb,$who);
      $score_info = Report::fraction_to_display($gb,$total->{'all'});
    }
    if ($what eq '2-category') {
      my $total = Crunch::totals($gb,$who,$cat,$assignments_in_this_cat);
      $score_info = Report::fraction_to_display($gb,$total->{$cat});
    }
    if ($what eq '3-assignment') {
      $score_info = $self->{DATA}->get_a_grade($cat.".".$ass,$who);
    }
    $self->{SCORE_VARIABLES}->[$j] = $score_info;
    $self->{SCORE_ENTRIES}->[$j] = Score->new($f,$who,$what,\$self->{SCORE_VARIABLES}->[$j],$self,$j,$n);
  }

  return;

}


=head4 refresh()

Empties the roster and fills it up again.
This currently assumes only one student can be selected at once.

=cut

sub refresh {
  my $self = shift;
  $self->clear();
  my $data = $self->{DATA};
  my $spread = $self->{SPREADSHEET};
  my $canvas = $spread->Subwidget('canvas');
  if (exists $self->{FRAME_IN_CANVAS}) {  $self->{FRAME_IN_CANVAS}->destroy()}
  $self->{CANVAS} = $canvas;
  my $f = $canvas->Frame();
  $self->{FRAME_IN_CANVAS} = $f;
  $f->configure(-takefocus=>0);
  # If a file is open, fill it back up:
  if ($data->file_is_open()) {
    my $gb = $data->{GB};
    my @k = $gb->student_keys("");
    $self->{KEYS} = \@k;
    my @name_buttons = ();
    my @name_button_colors = ();
    my @score_frames = ();
    my $n = @k;
    for (my $j=0; $j<$n; $j++) {
      my $who = $k[$j];
      my $name = $f->Button(
        -relief=>'flat',
        -text=>$data->key_to_name(KEY=>$who)
      )->grid(-sticky=>'w',-row=>$j,-column=>0);
      push @score_frames,$f->Frame()->grid(-sticky=>'w',-row=>$j,-column=>1);
      push @name_buttons,$name;
      my $hover_color = Prefs::get_roster_hover_color();
      my $bg_color = Prefs::get_roster_bg_color_1();
      if (($j/3)%2 == 1) {$bg_color = Prefs::get_roster_bg_color_2()}
      push @name_button_colors,$bg_color;
      $name->configure(-activebackground=>$hover_color,
                       -takefocus=>0,
                       -anchor=>'w',
                       -background=>$bg_color,
                       -width=>16,
                       -height=>1,
                       -font=>ExtraGUI::font('plain'),
                       -pady=>-3);
      my $jj= $j; # Need this for closure, although it doesn't seem like we should!
      $name->bind('<ButtonRelease-1>',sub{       # $who and $name are closure-ized in this subroutine
           $self->clicked_on_student($who);
           $self->{KBSEL}='';
           $self->{SELECTED_STUDENT} = $who;
           $name->focus();
           $self->{COLUMN_FOCUS} = 0;
         }
        );
      $name->bind('<Key>',sub{key_pressed_in_roster($self,$who,$jj,$n)});
      $name->bind('<KeyRelease>',sub{key_released_in_roster($self,$who,$jj,$n)});
      $name->bind('<FocusIn>',sub{$self->{COLUMN_FOCUS}=0});
      # Note: bindings for the Score object are handled inside its constructor.
    }
    $self->{COLUMN_FOCUS} = 1;
    $self->{NAME_BUTTONS} = \@name_buttons;
    $self->{NAME_BUTTON_COLORS} = \@name_button_colors;
    $self->{SCORE_FRAMES} = \@score_frames;
    $self->make_scores();
  }
  $canvas->createWindow(0,0,-anchor=>'nw',-window=>$f);
  Browser::main_window()->update;
  $canvas->configure(-scrollregion=>[$canvas->bbox('all')]);
}

# In some exceptional situations, we may return a null
# string. Calling routines are responsible for not freaking out if we do that.
sub type_of_assignment {
  my $self = shift;
  # The following line gets executed when the file is first opened:
  if (!(exists $self->{OPTIONS} && exists $self->{OPTION_VAR})) {return ''}
  my $ops = $self->{OPTIONS};
  my $o = $self->{OPTION_VAR};
  my $what;
  my $most_specific_possibility = '0';
  for (my $j=0; $j<@$ops; $j++) {
    my $jj = (@$ops-1)-$j; # see the comment elsewhere in the code that reads "Reverse order as a workaround..."
    if ($jj==0) {$what='1-overall'}
    if ($jj==1) {$what='2-category'}
    if ($jj==2) {$what='3-assignment'}
    if ($what>$most_specific_possibility) {$most_specific_possibility=$what}
    if ($o eq $ops->[$j]) {
      return $what;
    }
  }
  # The following should really never happen. It used to happen before I fixed a
  # bug that left the Optionmenu in a temporarily inconsistent state. It still doesn't
  # hurt to have it here just in case.
  return ''; # make_scores() knows what this means
}

sub sel_color {
  return '#b0d0ff';
}

=head4 key_pressed_in_scores()

Handle a key pressed in the scores.

=cut

sub key_pressed_in_scores {
  my $self = shift;
  my $who = shift; # student's database key
  my $j = shift;   # index into the array of students
  my $n = shift;   # number of students
  my $key = $Tk::event->K;
  my $them = $self->{KEYS};
  if (!@$them) {return}        # no file open, or the roster is empty

  my $score = $self->{SCORE_VARIABLES}->[$j];

  # The logic for handling control characters is now partly obsolete, because this routine is no longer bound to those. Num_lock also doesn't seem to get passed here (?).
  # We do get an event when the control key is pressed, but, e.g., no W when control-W is pressed.

  # We want to beep if they type bogus characters. 
  # It's important not to use this list for anything other than beeping -- if I add a new feature, there
  # may be new keys that are valid.
  # This does not help with recognizing when the keypad is being used but the NumLock is off, because everything gets passed in
  # here the same way regardless of the state of the NumLock key.

  my $beep_because_key_is_goofy = !($key=~m/([0-9]|\-|\.|x|space|Tab|plus|minus|period|Caps_Lock|Shift_(L|R)|Return|Control_(L|R)|KP_(Left|Begin|Right|Prior|Home|End|Subtract|Add|Delete|Insert)|Down|Up|Num_Lock)/i);

  # Also, try to detect cases where NumLock is off, and they're trying to type keys on the keypad. This is actually really hard to do in general, but
  # I just look for cass where they hit a keypad key, but the grade is still null. That catches the most common situation, where you start
  # typing in a bunch of grades on the keypad without realizing the num lock is off.

  my $beep_because_of_num_lock = ($key=~m/^KP_/) && ($score eq '');

  if ($beep_because_key_is_goofy) {
    ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'goofy_key',$key);
  }
  if ($beep_because_of_num_lock) {
    ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'num_lock');
  }


  my %not_part_of_scores = ('space'=>' ');


  #print "in scores, key=$key=, score=$score=\n"; #qwe

  # Pressing the space bar gets you from the grades column to the names column.
  if ($key eq 'space') {
    $self->{COLUMN_FOCUS}=0;
    $self->clicked_on_student($who);
  }

  if ($key ne 'Tab' && !($key =~ m/^Control/)) {
    my $assignment = $self->{STAGE}->{ASSIGNMENTS}->selected();
    my $entry = $self->{SCORE_ENTRIES}->[$j]->{WIDGET};
    my $linked_var = \($self->{SCORE_VARIABLES}->[$j]);
    if (exists $not_part_of_scores{$key}) {my $remove=$not_part_of_scores{$key}; $score=~s/$remove$//}
    $self->{SCORE_VARIABLES}->[$j] = $score;

    # Queue up the latest version of the input string, including this keystroke:
    if ($self->{STAGE}->{ASSIGNMENTS}->specific_assignment_selected()) {
      # The following code is to deal with a nasty bug resulting from how Perl/Tk implements control key equivalents for menu items.
      # If the keyboard focus is in a grade space and you then hit control-W to close the file, Tk apparently does this:
      #          1. temporarily sets the widget's string to a null string (why!?!?!?)
      #          2. sends the keystroke to the widget, which results in a call to this routine
      #          3. calls the hook I gave for the File>Close menu item
      #          4. sets the entry widget's string back to what it should be
      # This results in the loss of that student's grade!
      # The following code detects that bug. The idea is that there's no way that hitting the W key (or any alphanumeric key)
      # can result in a null string in the entry widget, so if we get that combo, we should /not/ put the null string on the grade queue.
      # Note added 2004 Feb. 7: This is no longer an issue, because I figured out how to keep Control-Key events from being bound to this routine.
      my $menu_control_key_bug = ($score eq '') && ($key=~m/^[a-zA-Z0-9]$/);

			if (!$menu_control_key_bug) {
        $self->{STAGE}->{BROWSER_WINDOW}->grades_queue(
          ACTION=>'put',  KEY=>$who.".".$assignment,  SCORE=>$score, 
                  ENTRY=>$entry,  LINKED_VAR=>$linked_var,
        );
		  }
    }

    # If appropriate, flush the queue, and select a different student.
    my %enter_it = ('Return'=>1,'KP_Enter'=>1,'Down'=>1,'Up'=>-1,'space'=>0);
    if (exists $enter_it{$key}) {
      $self->{STAGE}->{BROWSER_WINDOW}->grades_queue(
                 ENTRY=>$entry,  LINKED_VAR=>$linked_var); # flush grades queue
      if ($enter_it{$key}!=0) {
        $j = $j+$enter_it{$key};
        if ($j<0) {$j=0}
        if ($j>$n-1) {$j=$n-1}
        $self->change_student($them->[$j]);
      }
    }
  }
}

sub key_pressed_in_roster {
  my $self = shift;
  my $who = shift; # student's database key
  my $j = shift;   # index into the array of students
  my $n = shift;   # number of students
  my $key = $Tk::event->K;
  my $sel = $self->{KBSEL};
  my $them = $self->{KEYS};
  #print "key=$key=\n"; #-----debugging
  if (!@$them) { # no file open, or the roster is empty
    ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'empty_roster');
    return;
  }                                
  if ($key =~ m/\d/) { # Trying to enter grade when focus is on names column. Warn them it's going in the bit bucket.
    ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'grade_in_name_column');
    return;
  }
  if ($key eq 'Up' || $key eq 'Down') {    # e.g., typed J to select Jones, then down-arrow to get Josephson
    $sel='';
  }                                
  if ($key eq 'Tab') {                 # about to switch to the grades column; forget keyboard input
    $sel='';
  }                                
  if ($key =~ m/^\w$/ && $key ne '_' && !$self->{KBCTRL}) { # not Tab, control sequence, or underbar; could be a-z, A-Z, Chinese character,...
    $sel=$sel.lc($key); # Matches are case-insensitive, and we may as well impress that fact on them so they don't think they have to enter initial caps.
  } else {                        # not a normal alphabetic character
    my $recognized = 0;
    if ($key eq 'Delete' || $key eq 'BackSpace') {
      $recognized = 1;
      if (length($sel)>0) {
        $sel = substr($sel,0,length($sel)-1)
			}
      else {
        ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'nothing_left_to_delete'); # nothing left to delete
      }
    } # end if delete or backspace
    if ($key =~ m/^Control/) {        # Actual strings are Control-L and Control-R
      $recognized = 1;
      $self->{KBCTRL} = 1;
    }
    if ($key eq 'Shift' || $key eq 'Up' || $key eq 'Down' || $key eq 'Tab' || $key =~ m/^Shift/ || $key eq 'Num_Lock' 
         || $key eq 'Caps_Lock') {
      $recognized = 1;
    }
    if (!$recognized  && !$self->{KBCTRL}) { # They hit _, Escape, or some other key that doesn't make sense here.
      ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'key_not_recognized',$key);
    }
  }
  $self->{KBSEL} = $sel;

  # The following cases are handled below:
  # 1. They've typed in a string that doesn't match anybody. $match is null.
  # 2. They've typed in a string that matches someone. $match is their key
  # 3. They've hit an arrow key. $match is set appropriately.
  # 4. They've hit a key that doesn't affect anything. $match is the currently selected student.
  my $n_matches = 0;
  my $match = '';                # ---- case 1 ----
  if ($sel ne "") {
    foreach my $who (@$them) {
      if ($who =~ m/^$sel/i) {
        $match = $who;                # ---- case 2 ----
        ++$n_matches;
      }
    }
  } else {
    if ($key eq 'Up' || $key eq 'Down') {
      # ---- case 3 ---- 
      if ($key eq 'Up') {$j--} else {$j++}
      if ($j<0) {$j=0}
      if ($j>$n-1) {$j=$n-1}
      $match = $them->[$j];
    } else {
      $match = $self->{SELECTED_STUDENT}; # ---- case 4 ----
    }
  }

  if ($n_matches==1 && $key=~m/^\w$/ && length($sel)>=3) {
		ExtraGUI::click(); # confirm that they got a unique match
  }

  if ($match eq '' && $key=~m/^\w$/) {ExtraGUI::beep_if_allowed(Browser::main_window(),$self->{STAGE}->{BROWSER_WINDOW},'no_such_student',$sel)} # no such student

  $self->change_student($match); # null string is ok
}

# The point of the following is to keep track of whether the control key is being held down. If it
# is, we don't want to select students based on the alphabetic key they hit.
sub key_released_in_roster {
  my $self = shift;
  my $who = shift; # student's database key
  my $j = shift;   # index into the array of students
  my $n = shift;   # number of students
  my $key = $Tk::event->K;
  if ($key =~ m/^Control/) {        # Actual strings are Control-L and Control-R for left and right control keys
    $self->{KBCTRL} = 0;
  }
}

# This gets called by refresh(). Don't call it directly and expect the GUI
# to get redrawn.
sub clear {
  my $self = shift;
  $self->{KEYS} = [];
  $self->{NAMES} = [];

  my $buttons = $self->{NAME_BUTTONS};
  foreach my $button(@$buttons) {
    $button->destroy();
  }
  $self->{NAME_BUTTONS} = [];

  my $scores = $self->{SCORE_ENTRIES};
  foreach my $score(@$scores) {
    $score->destroy_widget();
  }
  $self->{SCORE_ENTRIES} = [];

}

# null string is ok
sub change_student {
  my $self = shift;
  my $new = shift;
  $self->{SELECTED_STUDENT} = $new;
  $self->clicked_on_student($new); # ok if null
}

sub select_next_student {
  my $self = shift;
  my $what = shift; # 1 or -1
  if ($what==0) {return}
  my $foo = $self->{KEYS};
  my $n = @$foo;
  my $old;
  for (my $j=0; $j<=$n-1; $j++) {
    if ($self->{KEYS}->[$j] eq $self->{SELECTED_STUDENT}) {$old=$j}
  }
  if ($old eq '') {return}
  my $new = $old+$what;
  if ($new>=0 && $new<=$n-1) {
    $self->{SELECTED_STUDENT} = $self->{KEYS}->[$new];
    $self->clicked_on_student($self->{SELECTED_STUDENT});
  }
}


=head4 selected()

Get or set the key of the student who is currently selected.

=cut

sub selected {
  my $self = shift;
  if (@_) {
    $self->{SELECTED}=shift;
    $self->{STAGE}->roster_has_set_student($self->{SELECTED});
  }
  return $self->{SELECTED};
}



sub clicked_on_student {
  my $self = shift; # Roster
  my $key = shift;

  $self->selected($key);

  # Fiddle with colors, and only let the selected one have the focus:
  my $r = $self->{KEYS};
  my @keys = @$r;
  for (my $j=0; $j<@keys; $j++) {
    if ($keys[$j] eq $key) {
      $self->{NAME_BUTTONS}->[$j]->configure(-font=>ExtraGUI::font('bold'),-takefocus=>1);
      $self->{SCORE_ENTRIES}->[$j]->highlighted(1);
      if ($self->{STAGE}->{ASSIGNMENTS}->specific_assignment_selected()) {
        $self->{SCORE_ENTRIES}->[$j]->enabled(1);
        if ($self->{COLUMN_FOCUS}==0) {
          $self->{NAME_BUTTONS}->[$j]->focus();
        }
        else {
          $self->{SCORE_ENTRIES}->[$j]->focus();
        }
      }
    }
    else {
      $self->{NAME_BUTTONS}->[$j]->configure(-font=>ExtraGUI::font('plain'),-takefocus=>0);
      $self->{SCORE_ENTRIES}->[$j]->highlighted(0);
      $self->{SCORE_ENTRIES}->[$j]->enabled(0);
    }
  }

}

#---------------------------------------------------
# Score class
#---------------------------------------------------

=head3 Score class

This represents an Entry for a score, or a piece of
text describing the student's overall score in a certain
category, or in the whole course.

=cut

package Score;

sub new {
  my $class = shift;
  my $frame = shift;       # the frame it goes into; it's the sole occupant
  my $who = shift;         # student key
  my $what = shift;        # ='1-overall', '2-category', or '3-assignment'; numbers are so comparisons work
  my $textvar_ref = shift; # the text variable tied to the Entry
  my $roster = shift;
  my $j = shift;           # index into the roster
  my $n = shift;           # number of students in the roster
  my $self = {};
  bless($self,$class);
  $self->{FRAME} = $frame;
  $self->{WHO} = $who;
  $self->{WHAT} = $what;
  $self->{ROSTER} = $roster;
  my $relief = 'flat';
  if ($what eq '3-assignment') {$relief='sunken'}
  my $gb = $roster->{DATA}->{GB};
  my $justify = {0=>'left',1=>'right',''=>'left'}->{$gb->preferences()->get('justify')};
    # ...default to left because some implementation of Perl/Tk have a bug in displaying right-justified entries
  $self->{WIDGET} = $frame->Entry(
        -width=>9,
        -takefocus=>0,
#        -state=>'disabled',
            # ...Removed this 2004 jan 25 because in the latest version of Perl/Tk, it was causing the scores to dim out and become hard
            # to read. Not sure why I did this in the first place. Similar change at one other place in the code, marked with same date.
        -textvariable=>$textvar_ref,
#        -font=>ExtraGUI::font('fixed_width'), # hurts performance a lot for some reason
        -relief=>$relief,
        -justify=>$justify,
      )->pack();
  my $w = $self->{WIDGET};
  $w->bind('<ButtonRelease-1>',sub{       # $who and $name are closure-ized in this subroutine
           $roster->clicked_on_student($who);
           $roster->{KBSEL}='';
           $roster->{SELECTED_STUDENT} = $who;
           $self->focus();
           $roster->{COLUMN_FOCUS} = 1;
         }
        );
  $w->bind('<Key>',sub{Roster::key_pressed_in_scores($roster,$who,$j,$n)});
  $w->bind('<Control-Key>',sub{}); # This overrides the <Key> binding in the special case of Control-Key, so we don't call key_pressed_in_scores on that key.
  $w->bind('<FocusIn>',
    sub{
      $roster->{COLUMN_FOCUS}=1;
      $w->selectionClear(); # This is extremely important, because otherwise you hit tab to go to a preexisting score, it selects the whole score, and then if you
                            # hit space, it clears the score!
      $w->icursor('end');
    }
  );
  $w->bind('<FocusOut>',sub{$self->tidy_when_leaving()});

  return $self;
}

sub widget {
  my $self = shift;
  return $self->{WIDGET};
}

sub frame {
  my $self = shift;
  return $self->{FRAME};
}

sub tidy_when_leaving {
  my $self = shift;
  $self->{WIDGET}->selectionClear();
}

sub focus {
  my $self = shift;
  $self->{WIDGET}->focus();
}

sub highlighted {
  my $self = shift;
  my $want_highlighted = shift;
  if ($want_highlighted) {
    $self->{WIDGET}->configure(-font=>ExtraGUI::font('bold'));
  }
  else {
    $self->{WIDGET}->configure(-font=>ExtraGUI::font('plain'));
  }
}

=head4 is_editable()

Tells whether this is a specific assignment that we can edit the score on.
Doesn't know or care whether it's actually enabled (e.g., whether it actually
has the focus).

=cut

sub is_editable {
  my $self = shift;
  return $self->{WHAT} eq '3-assignment';
}

=head4 enabled()

Set whether it's enabled for editing and can take the focus.
Ignored if not editable.

=cut

sub enabled {
  my $self = shift;
  my $want_enabled = shift;
  if ($want_enabled && $self->is_editable()) {
    $self->{WIDGET}->configure(-takefocus=>1,-state=>'normal');
  }
  else {
    #$self->{WIDGET}->configure(-takefocus=>0,-state=>'disabled');
        # ... Commented out 2004 jan 25. See comment above with same date.
  }
}

sub destroy_widget {
  my $self = shift;
  $self->{WIDGET}->destroy();
}


1;
