#----------------------------------------------------------------
# Copyright (c) 2002 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------

=head2 ExtraGUI.pm

This module contains a bunch of miscellaneous GUI code for making
windows other than the main Browser.pm window, e.g. dialog boxes.
For the most part, Browser.pm calls routines in ExtraGUI.pm,
not the other way around.
The only exceptions are (1) we do calls to
 Browser::empty_toplevel_window(); (2) we make use
of $Browser::words.

=cut

use strict;

use Cwd;
use Tk;
use Browser;
use Words;
use MyWords;
use Input;
use Preferences;
BEGIN {
  eval "use Audio::Data";
  eval "use Audio::Play";
}

package ExtraGUI;

# See the Words and MyWords modules for info on the following.
our $words_prefix;                # initialized in each routine
sub w {
    my $key = shift;
    my $newline = "";
    if ($key =~ m/\n$/) {$newline="\n"}
    chomp $key;
    return $Browser::words->get($words_prefix.".".$key).$newline;
}


BEGIN {
  my %fonts = ();
  my %size    = ('plain' => 11,          'bold' => 11,          'small_plain' => 10,         'fixed_width'=>11);
  my %family  = ('plain' => 'helvetica', 'bold' => 'helvetica', 'small_plain' => 'helvetica','fixed_width'=>'courier');
  my %weight  = ('plain' => 'normal',    'bold' => 'bold',      'small_plain' => 'normal',   'fixed_width'=>'normal');
  my %slant   = ('plain' => 'roman',     'bold' => 'roman',     'small_plain' => 'roman',    'fixed_width'=>'roman');
  my $widget; # Gets initialized by calling fonts_init. We just need to have some widget
              # that we can call the fontCreate() method on.
  sub fonts_init {
    $widget = shift;
  }
  sub font {
    my $style = shift;
    if (!exists($fonts{$style})) {
      $fonts{$style} = $widget->fontCreate(-size=>$size{$style},-family=>$family{$style},-weight=>$weight{$style},-slant=>$slant{$style});
    }
    return $fonts{$style};
  }
}

=head3 preferred_location

Returns a geometry string describing where dialog boxes should be placed: in front of the main window,
and just a little lower at the top, like ``sheets'' in MacOS X.

=cut

sub preferred_location {
  my ($x,$y) = preferred_location_coords();
  # For some reason, the main window's geometry is returned as the coordinates of the /interior/
  # of the window. Therefore, we don't really need to add anything to $y.
  return coords_to_geometry_string($x,$y);
}

sub offset_location {
  my $xoffset = shift; my $yoffset = shift;
  my ($x,$y) = preferred_location_coords();
  return coords_to_geometry_string($x+$xoffset,$y+$yoffset);
}

sub coords_to_geometry_string {
  my $x = shift; my $y = shift;
  if (! ($y =~ m/^[\+\-]/)) {$y='+'.$y}
  if (! ($x =~ m/^[\+\-]/)) {$x='+'.$x}
  return $x.$y;
}

sub preferred_location_coords {
  Browser::main_window()->geometry() =~ m/^=?(\d+)x(\d+)([+-]\d+)([+-]\d+)$/;
  return ($3,$4);
}

=head3 choose_file()

Displays a file dialog box.

=cut

sub choose_file {
  my %args = (
    CALLBACK=>sub{},
    TITLE=>'',
    CREATE=>0,
    PATH=>Cwd::getcwd(),
    INITIALFILE=>'',
    FILETYPES=>[['OpenGrade Files','.gb'],['All Files','*']],
    DEFAULT_EXTENSION=>'.gb',
    WHAT=>'input', # can also be 'output'
    @_,
  );
  my $callback = $args{CALLBACK};
  my $filetypes = $args{FILETYPES};
  my $title = $args{TITLE};
  my $default_extension = $args{DEFAULT_EXTENSION};
  my $path = $args{PATH};
  my $initialfile = $args{INITIALFILE};
  $path =~ s|/$||;

  my $file_name;
  if ($args{WHAT} eq 'input') {
    # choosing file to open:
    $file_name = Browser::main_window()->getOpenFile(-initialdir=>$path,-title=>$title,-filetypes=>$filetypes);
  }
  if ($args{WHAT} eq 'output') {
    # choosing file to save as:
    $file_name = Browser::main_window()->getSaveFile(-initialdir=>$path,-initialfile=>$initialfile,-title=>$title,-defaultextension=>$default_extension);
  }
  &$callback($file_name);
}


=head3 error_message()

Takes one parameter, which is the error message.

=cut

sub error_message {
  my $message = shift;
  local $words_prefix = "b.error_message";
  my $box = Browser::empty_toplevel_window(w('error'));
  $box->Label(-text=>$message,-justify=>'left',-font=>font('plain'))->pack();
  $box->Button(-text=>w("ok"),-command=>sub{$box->destroy()})->pack();
  $box->geometry(preferred_location());
  beep_if_allowed($box);
}


=head3 message()

Takes one parameter, which is the message. Cf. error_message().


=cut

sub message {
  my $message = shift;
  local $words_prefix = "b.error_message";
  my $box = Browser::empty_toplevel_window('');
  $box->geometry(preferred_location());
  $box->Label(-text=>$message,-justify=>'left',-font=>font('plain'))->pack();
  $box->Button(-text=>w("ok"),-command=>sub{$box->destroy()})->pack();
}

=head3 confirm()

First argument is a question, e.g., "Drop Al Einstein?"
Second argument is a callback, which gets
1 if they hit ok, 0 if they cancel.
This can also be used for other kinds of two-choice dialog boxes:
the third and fourth arguments, if present, are text to use
instead of OK and Cancel.

=cut

sub confirm {
  my $question = shift;
  my $callback = shift;
  local $words_prefix = "b.dialog";
  my $ok_text = w('ok');
  if (@_) {$ok_text = shift}
  my $cancel_text = w('cancel');
  if (@_) {$cancel_text = shift}
  my $box = Browser::empty_toplevel_window(w('confirm'));
  $box->geometry(preferred_location());
  $box->Label(-text=>$question,-justify=>'left',-font=>font('plain'))->pack();
  $box->Button(-text=>$ok_text,-command=>sub{$box->destroy();&$callback(1)})->pack();
  $box->Button(-text=>$cancel_text,-command=>sub{$box->destroy();&$callback(0)})->pack();
}

=head3 ask()

Get a string as input in a dialog box.

    PROMPT=>"?",
    TITLE=>"",
    WIDTH=>40,
    CALLBACK=>sub{},
    ACTION_BUTTON=>w("ok"),
    DEFAULT=>'',
    PASSWORD=>0

The callback routine gets the result, or gets no argument if the
user hit cancel.

=cut

sub ask {
  local $words_prefix = "b.dialog";
  my %args = (
    PROMPT=>"?",
    TITLE=>"",
    WIDTH=>40,
    CALLBACK=>sub{},
    ACTION_BUTTON=>w("ok"),
    DEFAULT=>'',
    PASSWORD=>0,
    BEEP=>0, # If set to 1, /and/ if user's prefs allow it, beep.
    @_
  );
  my $callback = $args{CALLBACK};
  my $result = $args{DEFAULT};
  my $pwd = $args{PASSWORD};
  my $beep = $args{BEEP};
  my @pwd_option = ();
  if ($pwd) {@pwd_option=(-show=>'*')}
  my $box = Browser::empty_toplevel_window($args{TITLE});
  $box->geometry(preferred_location());
  $box->Label(-anchor=>'w',-justify=>'left',-text=>$args{PROMPT})->pack();
  my $f = $box->Frame->pack(-side => 'right');
  my $e = $box->Entry(
              -takefocus=>1,
              -width=>$args{WIDTH},
              -textvariable=>\$result,
              @pwd_option,
  )->pack(-side=>'left');
  $e->focus();
  $e->icursor('end');
  $f->Button(-text=>w("cancel"),-command=>sub{$box->destroy();&$callback()})->pack(-side=>'right');
  $e->bind('<KeyRelease-Return>'=>sub{$box->destroy();&$callback($result)});
  $f->Button(-text=>w("ok"),-command=>sub{$box->destroy();&$callback($result)})->pack(-side=>'right');
  if ($beep) {beep_if_allowed($box)}
}

sub save_plain_text_to_file {
  my $text = shift;
  my $path = '.';
  if (@_) {$path = shift}
  my $initialfile = ''; # just the tail of the filename
  if (@_) {$initialfile = shift}
  local $words_prefix = "b.dialog";
  my $filename = $path;
  if ($path eq '.') {
    $filename = '';
  }
  if ($filename ne '' && !($filename =~ m@/$@)) {$filename="$filename/"} # put a slash on the end if it doesn't already have one

  my $do_it = sub {
              local $words_prefix = "b.dialog";
              my $really_do_it = shift;
              if ($really_do_it == 1) {
                my $err = '';
                if ($err eq '') {
                  if (open(FILE,">$filename")) {
                    print FILE $text;
                    close(FILE);
                  }
                  else {
                    $err = sprintf w('error_opening_for_output'),$filename;
                  }
                }
                if ($err ne '') {error_message($err)}
              }
  };
  my $callback = sub {
              local $words_prefix = "b.dialog";
              $filename = shift;
              if ($filename eq '') {return} # user canceled
              my $err = '';
              if ($filename =~ m/\.gb$/) {$err = w('save_on_top_of_gb')}
              # If we're overwriting a preexisting file, getOpenFile will already have asked for confirmation.
              if ($err eq '') {
                                                                &$do_it(1);
              }
              else {
                error_message($err);
              }
  };

  ExtraGUI::choose_file(TITLE=>w('save_as'),WHAT=>'output',CREATE=>1,DEFAULT_EXTENSION=>'',CALLBACK=>$callback,PATH=>$filename,INITIALFILE=>$initialfile);

}

=head3 show_text()

Show some text, and give the user the option of saving it in a file.
    TITLE=>"",
    TEXT=>"",
    WIDTH=>60,
    MIN_HEIGHT=>3,
    MAX_HEIGHT=>50,
    BGCOLOR=>"white",
    PATH=>'.', # default directory for saving

=cut

sub show_text {
  local $words_prefix = "b.dialog";
  my %args = (
    TITLE=>"",
    TEXT=>"",
    WIDTH=>60,
    MIN_HEIGHT=>3,
    MAX_HEIGHT=>50,
    BGCOLOR=>"white",
    PATH=>'.', # default directory for saving
    FILENAME=>'', # default filename for saving
    @_,
  );
  my $text = $args{TEXT};
  my $width = $args{WIDTH};
  my $nlines = 0;
  my @lines = split /\n/,$text; # in scalar context, we get the number of elements in the array
  foreach my $line(@lines) {
    $nlines = $nlines + int(length($line)/$width) + 1;
  }
  my $height = $nlines;
  if ($height>$args{MAX_HEIGHT}) {$height=$args{MAX_HEIGHT}}
  if ($height<$args{MIN_HEIGHT}) {$height=$args{MIN_HEIGHT}}
  my $box = Browser::empty_toplevel_window($args{TITLE});
  $box->geometry(preferred_location());
  my $t = $box->Scrolled("Text",
        -scrollbars=>'e',
        -width=>$width,
        -height=>$height,
        -background=>$args{BGCOLOR},
  )->pack();
  my $text_without_newline_at_end = $text;
  $text_without_newline_at_end =~ s/\n$//; # newline at the end is shown as an extraneous blank line
  $t->insert('end',$text_without_newline_at_end);
  # Note: The following is commented out because even though this routine is meant to be used
  # with immutable text, if you disable the Text object, they can't copy it to the clipboard.
  #  $t->configure(-state => 'disabled'); # so they can't type in it
  my $f = $box->Frame();
  $f->Button(-text=>w('ok'),-command=>sub{$box->destroy()})->pack(-side=>'left');
  $f->Button(-text=>w('save'),-command=>sub{save_plain_text_to_file($args{TEXT},$args{PATH},$args{FILENAME}); $box->destroy()})->pack(-side=>'left');
  $f->pack();
}

=head3 fill_in_form

Arguments:

      INPUTS=>[], # ref to an array of Input objects
      TITLE=>"",
      INFO=>'',   # information displayed at the top
      CALLBACK=>sub{},
      CANCEL_CALLBACK=>sub{},
      COLUMNS=>2, # 2=prompts side by side with blanks, 1=vertical arrangement
      WIDTH=>80,

This routine creates a dialog box and asks for the inputs referred to in the INPUTS array.
Then it calls the callback routine, giving it a
reference to a hash containing the strings given by the
user. The hash keys are defined by the {KEY} instance variables of the Input objects.

=cut

sub fill_in_form {
  my %args = (
        INPUTS=>[], # ref to an array of Input objects
        TITLE=>"",
        INFO=>'',   # information displayed at the top
        CALLBACK=>sub{},
        CANCEL_CALLBACK=>sub{},
        COLUMNS=>2, # 2=prompts side by side with blanks, 1=vertical arrangement
        WIDTH=>80,
        HEIGHT=>'',
        MAX_HEIGHT=>30,
        XOFFSET=>0,
        YOFFSET=>0,
        @_,
  );
  my $callback = $args{CALLBACK};
  my $cancel_callback = $args{CANCEL_CALLBACK};
  my $inputs = $args{INPUTS};
  my $info = $args{INFO};
  my $xoffset = $args{XOFFSET};
  my $yoffset = $args{YOFFSET};
  my $n_inputs = @$inputs;
  my $max_height = $args{MAX_HEIGHT};
  my $height;
  if ($args{HEIGHT} ne '') {
    $height=$args{HEIGHT}
  }
  else {
    # They get a scrollbar if it's too small, but too big can be a pain, because the OK button is hidden.
    $height = 10+6*$n_inputs;
    if ($height>$max_height) {
      $height = $max_height;
    }
  }
  my %results = ();
  local $words_prefix = "b.dialog";
  my $box = Browser::empty_toplevel_window($args{TITLE});
  $box->geometry(offset_location($xoffset,$yoffset));

  if ($info ne '') {
    $box->Label(-anchor=>'w',-justify=>'left',-text=>$info,-width=>$args{WIDTH})->pack(-side=>'top');
  }

  my $f = $box->Frame->pack(-side => 'bottom');
  $f->Button(-text=>w("ok"),
     -command=>
       sub{
         my $has_errors = 0;
         my $error_text = '';
         foreach my $input(@$inputs) {
           my $key = $input->{KEY};
           my $retrieve = $input->{RETRIEVE};
           my $value = &$retrieve;
           $results{$key} = $value;
           my @err_stuff = $input->check($value);
           my $is_an_error = @err_stuff;
           $has_errors = $has_errors || $is_an_error;
           if ($is_an_error) {
             my $fmt = shift @err_stuff;
             my $this_error = (sprintf $fmt,(@err_stuff));
             $error_text = $error_text . $key.": ".$this_error."\n";
           }
         }
         if (!$has_errors) {
                 $box->destroy();
                 &$callback(\%results)
         }
               else {
                 error_message($error_text);
         }
       }
  )->pack(-side=>'left'); 
  $f->Button(-text=>w("cancel"),
     -command=>[
       sub{
         my $box = shift;
         my $callback = shift;
         my $results = shift;
         $box->destroy();
         &$cancel_callback()
       }
       ,$box,$cancel_callback
     ]
  )->pack(-side=>'left');

  my $ncols = $args{COLUMNS};
  my $t;
  if ($ncols==2) {
    $t = $box->Scrolled("Text",
         -scrollbars=>'oe',
         -width=>$args{WIDTH},
         -height=>$height,
    )->pack();
  }
  else { # one column
    $t = $box->Scrolled("Text",
        -scrollbars=>'oe',
        -width=>$args{WIDTH},
        -height=>$height,
    )->pack();

  }
  my $first_one = 1;
  foreach my $input(@$inputs) {
    my $which = $input->{KEY};
    my $w = $t->Label(-anchor=>'w',-justify=>'left',-text=>$input->{PROMPT},-width=>$args{WIDTH}/$ncols);

    $t->windowCreate('end',-window=>$w);
    if ($args{COLUMNS}==1) {$t->insert('end',"\n");}
    $results{$which} = $input->{DEFAULT};
    if ($input->{WIDGET_TYPE} eq 'entry') {
      $w = $t->Entry(
                   -textvariable=>\$results{$which},
       -width=>$args{WIDTH}/$ncols,
      );
      $input->{RETRIEVE} = sub{return $results{$which}}
    }
    if ($input->{WIDGET_TYPE} eq 'radio_buttons') {
      $w = $t->Frame;
      my $radio_buttons = $input->{RADIO_BUTTONS};
      my $order = $input->{RADIO_BUTTON_ORDER};
      foreach my $value(@$order) {
        $w->Radiobutton(-text=>$radio_buttons->{$value},-value=>$value,-variable=>\$results{$which})
                        ->pack(-side=>'left');
      }
      $results{$which} = $input->{DEFAULT};
      $input->{RETRIEVE} = sub{return $results{$which}};
    }
    if ($input->{WIDGET_TYPE} eq 'text') {
      $w = $t->Scrolled('Text',
        -scrollbars=>'e',
        -width=>$args{WIDTH}/$ncols,
        -height=>90);
      $input->{RETRIEVE} = sub{return $w->get('1.0','end');}
    }
    $t->windowCreate('end',-window=>$w);
    if ($first_one) {$w->focus}
    $first_one = 0;
    $t->insert('end',"\n");
  }
  $t->configure(-state => 'disabled'); # so they can't type in it
}


=head3 beep_if_allowed

Called with one argument, just beeps (if allowed). Use this only when the actual error is being displayed in
its own dialogbox.

Normally you call it with three arguments. The second argument is the BrowserWinudow, 
 and the third is a key of a message to display at the
bottom of the window, e.g., 'luser' for 'b.beep.luser'.

You can call it with four arguments, in which case the fourth is used as a %s argument to sprinf.

=cut

sub beep_if_allowed {
  my $widget = shift; # Browser::main_window()
  my $browser_window = undef;
  if (@_) {$browser_window = shift}
  my $message_code = '';
  if (@_) {$message_code = shift}
  my $message_info = '';
  if (@_) {$message_info = shift}
  local $words_prefix = 'b.beep';
  if (want_beeps()) {$widget->bell}
  if ($message_code ne '') {
    my $message = w($message_code);
    if ($message_info ne '') {$message = sprintf $message,$message_info}
    $browser_window->set_footer_text($message);
  }
}

sub want_beeps {
  return Preferences->new()->get('beep');
}

BEGIN {
  my ($server,$sound);
  eval { # Don't crash and burn if they don't have Audio installed.

    $server = Audio::Play->new;

    # Synthesize a little "shk" sound. 

    my $silent = Audio::Data->new(rate=>44100.);
    my $shk = Audio::Data->new(rate=>44100.);

    my $delay = 0.06; # so it doesn't occur at the same instant as the sound of the clicking key
    my $duration = 0.05;

    $silent->silence($delay);

    $shk->noise($duration,1.6);
    my $n = $shk->samples;
    my $attack_frac = .1;
    for (my $i=0; $i<=$n; $i++) {
      my $attack = 1.;
      if ($i<$n*$attack_frac) {$attack=$i/($n*$attack_frac)}
      my $tail = ($n-$i)/$n;
      my $env = $attack*$attack*$tail*$tail*$tail;
      $shk->[$i] *= $env;
    }

    $sound = Audio::Data->new(rate=>44100.);
    $sound = $silent . $shk;

  };
  sub click {
    return if ! want_beeps();
    eval { # Don't crash and burn if they don't have Audio installed, in which case $server and $sound are undefs.
      $server->play($sound);
    }
  }
}

1;
