#----------------------------------------------------------------
# Copyright (c) 2002 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------

=head2 GradeBook.pm

This package defines the GradeBook class, which is essentially just
some hashes that implement the gradebook file format in memory.
In addition to these hashes, it has a few more data structures:

=over

=item *

An array of assignment categories, to keep track of the order in which
     the categories are supposed to be displayed in reports.
     Interface: category_list(), add_category(),...

=item *

An array of assignments. Interface: assignment_list(),...

=back

To make sure that these arrays stay in sync with the hashes, it's
important not to use any of the class's private methods from outside
the package!

=cut

use strict;
use POSIX;
use File::Copy;
use LineByLine;
use Preferences;
use NetOG;
use Words;
use MyWords;
use Fcntl qw/:flock/; # brings in LOCK_EX, etc.

use Clone;
push @Gradebook::ISA, 'Clone';

package GradeBook;

# See the Words and MyWords modules for info on the following.
our $words;                        # initialized in main_loop()
our $words_prefix;                # initialized in each routine
sub w {
    my $key = shift;
    my $newline = "";
    if ($key =~ m/\n$/) {$newline="\n"}
    chomp $key;
    return $words->get($words_prefix.".".$key).$newline;
}


sub auto_save_if_they_want_it {
    my $self = shift;
    my $s = $self->when_last_autosaved();
    my $m = $self->when_last_modified();
    if ($self->want_auto_save() && (defined $m) && ((!(defined $s)) || ($s<=$m)) && time>$m) {
        # Testing for $s<=$m, as opposed to $s<$m, is necessary for safety, since the clock has a granularity of one
        # second, and the check for the autosave could happen in the same second as the modification.
        # It does raise the possibility of saving twice in a row, which is the reason for the final condition on time>$m.
        # This may have the effect of delaying the autosave for one cycle, which is preferable to double-saving or not saving at all.
        $self->auto_save();
    }
}

sub auto_save {
    my $self = shift;
    $self->write_to_named_file($self->autosave_filename());
    $self->mark_autosaved_now();
}

sub auto_save_file_exists {
    my $self = shift;
    return -e ($self->autosave_filename());
}


sub want_auto_save {
    my $self = shift;
    #return $self->boolean_pref("auto_save");
    return 1;
}

sub want_auto_backup {
    my $self = shift;
    #return $self->boolean_pref("auto_backup");
    return 1;
}

=head3 when_last_modified()

Get or set the time when the gradebook object was last modified. Returns
undef if never modified.

=cut

sub when_last_modified {
    my $self = shift;
    if (@_) {
      $self->{WHEN_LAST_MODIFIED} = shift;
    }
    return $self->{WHEN_LAST_MODIFIED};
}

=head3 mark_modified_now()

Mark the gradebook as having been modified now.

=cut

sub mark_modified_now {
    my $self = shift;
    $self->when_last_modified(time);
}

=head3 when_last_autosaved()

Get or set the time when the gradebook object was last auto-saved.
Returns undef if never autosaved.

=cut

sub when_last_autosaved {
    my $self = shift;
    if (@_) {
      $self->{WHEN_LAST_AUTOSAVED} = shift;
    }
    return $self->{WHEN_LAST_AUTOSAVED};
}

=head3 mark_autosaved_now()

Mark the gradebook as having been autosaved now.

=cut

sub mark_autosaved_now {
    my $self = shift;
    $self->when_last_autosaved(time);
}
=head3 write()

Returns a null string normally, or an error message otherwise.
Writes to the remembered filename, and does an emacs-style auto-backup
with a tilde on the end, if the auto_backup preference is set.
At this point, the file is presumably locked, but that's OK; it was
locked, it stays locked, and we still have it open for purposes of
locking.

=cut

sub write {
    my $self = shift;
    my $auto_backup = $self->want_auto_backup();
    my $file_name = $self->file_name();
    $self->close();
    my $temp_file_name = "ogr_tmp_$$";
    my $result = $self->write_to_named_file($temp_file_name);
    if ($result ne "") {return $result} # if there's an error, return
    if ($auto_backup && !$self->{WROTE_TILDE_FILE}) {
      File::Copy::copy($file_name,"$file_name~");
      $self->{WROTE_TILDE_FILE} = 1;
    }
    unlink $file_name; # ignore errors, which just mean file is being created
    rename $temp_file_name, $file_name or return $!;
    unlink $self->autosave_filename(); # ignore errors, since it may not exist
    return "";
}

sub autosave_filename {
  my $self = shift;
  my $name = $self->file_name();
  $name =~ s|([^/]*)$|\#$1\#|;
  return $name;
}

=head3 write_to_named_file()

Returns a null string normally, or an error message otherwise.
Shouldn't normally call this directly, because it doesn't do
any of the safety features. Call write() instead.

=cut

sub write_to_named_file {
    my $self = shift;
    my $file_name = shift;
    open(OUTFILE,">$file_name") or return "Error opening $file_name file for output";

    my $mac = LineByLine->new(KEY=>$self->password(),ITERATIONS=>iterations_for_line_by_line());
    print OUTFILE $mac->head();

    print OUTFILE $mac->line("class");
    print OUTFILE $mac->line("  .title \"".$self->title()."\"");
    print OUTFILE $mac->line("  .staff \"".$self->staff()."\"");
    print OUTFILE $mac->line("  .days  \"".$self->days()."\"");
    print OUTFILE $mac->line("  .term  \"".$self->term()."\"");
    print OUTFILE $mac->line("  .dir   \"".$self->dir()."\"");
    print OUTFILE $mac->line("  .standards   ".hash_to_comma_delimited($self->standards()));
    print OUTFILE $mac->line("");
    print OUTFILE $mac->line("preferences");
    my $h = $self->preferences_private_method();
    foreach my $p(keys(%$h)) {
        print OUTFILE $mac->line("  .".lc($p)." \"".$h->{$p}."\"");
    }
    print OUTFILE $mac->line("");
    print OUTFILE $mac->line("categories");
    my @cl = split(",",$self->category_list());
    foreach my $c(@cl) {
        print OUTFILE $mac->line("  .".$c." ".$self->category_properties_comma_delimited($c));
    }
    print OUTFILE $mac->line("");

    print OUTFILE $mac->line("roster");
    my @student_keys = $self->student_keys("all");
    $h = $self->roster_private_method();
    foreach my $k (@student_keys) {
        my $v = $h->{$k};
        $v =~ s/^,//; # strip leading comma
        $v =~ s/,$//; # strip trailing comma
        print OUTFILE $mac->line("  .$k $v");

    }
    print OUTFILE $mac->line("");

    print OUTFILE $mac->line("assignments");
    my @a = split(",",$self->assignment_list());
    my $ass = $self->assignments_private_method();
    foreach my $a(@a) {
        print OUTFILE $mac->line("  .$a ".$self->strip_redundant_properties($a,$ass->{$a}));
    }
    print OUTFILE $mac->line("");

    print OUTFILE $mac->line("grades");
    my $g = $self->grades_private_method();
    my $z = $self->roster_private_method();
    my @r = sort keys %$z;

    # The hard-coded 1000 in the following may be ugly, but I doubt it will
    # ever be an issue, and this /is/ an efficient way to do the sort.
    my @grades_to_sort = ();
    while (my ($k,$v) = each (%$g)) {
      $k =~ m/([^\.]+)\.([^ ]+)/;
      my $who = $1;
      my $cat = $2;
      push @grades_to_sort, (sprintf "%-1000s%-1000s%s",$cat,$who,$v);
    }
    @grades_to_sort = sort @grades_to_sort;
    foreach my $line(@grades_to_sort) {
        my $cat = substr($line,0,1000);
        my $who = substr($line,1000,1000);
        my $v   = substr($line,2000);
        $cat =~ s/ +$//; # delete trailing blanks
        $who =~ s/ +$//; # delete trailing blanks
        #print "$cat,$who,$v\n";
        print OUTFILE $mac->line("  .$who.$cat $v");
    }

    if (my $stuff = $self->stuff_unable_to_parse()) {
      print OUTFILE $mac->line("");
      while ($stuff =~ m/([^\n]*)\n/g) {
        my $line = $1;
        print OUTFILE $mac->line($line);
      }
    }

    print OUTFILE $mac->tail();

    close(OUTFILE) or return "Error closing output file";
    return "";
}

=head3 new()

GradeBook->new() creates a new GradeBook object.

=cut

sub new {
    my $class = shift;
    my %args = (
                TITLE=>"",
                STAFF=>"staff",
                DAYS=>"MTWRF",
                TERM=>"",
                DIR=>"web_reports",
                #FTP=>"",
                FILE_NAME=>"",
                STANDARDS=>"",
                PASSWORD=>"",
                @_
                );
    my $self = {};
    bless($self,$class);
    $self->class_private_method({
        "title"=>$args{TITLE},
        "staff"=>$args{STAFF},
        "days"=>$args{DAYS},
        "term"=>$args{TERM},
        "dir"=>$args{DIR},
        #"ftp"=>$args{FTP},
        "standards"=>$args{STANDARDS},
    });
    $self->preferences_private_method({"backups_on_server"=>'"true"',
                                   "auto_save"=>'"true"',
                                   "auto_backup"=>'"true"'});
    $self->modified(1);
    $self->file_name($args{FILE_NAME});
    $self->password($args{PASSWORD});
    $self->category_list("");
    $self->clear_assignment_list;
    $self->misc_initialization();
    return $self;
}

sub clear_grades {
  my $self = shift;
  $self->grades_private_method({});
}

sub clear_assignment_list {
  my $self = shift;
  $self->assignment_list("");
}

sub clear_roster {
  my $self = shift;
  $self->roster_private_method({});
}


=head3 union()

$a->union($b,$ask) adds scores and students from gradebook $b into gradebook $a. Nothing is done
with the preferences or grading standards; these are left as they were (in $a).
When a student exists in both gradebooks, the student's properties are left as they were in $a, and likewise
for category and assignment properties.

When a grade is present in both gradebooks, and the grades
are unequal, call the subroutine $ask to decide which to use. This subroutine can interact with the
user, or work according to some predefined rule. If no second argument is provided, the default $ask
routine is one that just returns 0, causing grades from $a to be retained in all cases.

Returns text describing everything that was done.

=cut

sub union {
  my $a = shift;
  my $b = shift;
  my $ask = sub {return 0};
  if (@_) {$ask = shift}
  my $log = '';

  # Copy students:
  foreach my $b_student(sort ($b->student_keys('all'))) {
    if (!(exists($a->roster_private_method->{$b_student}))) {
      $a->roster_private_method->{$b_student} = $b->roster_private_method->{$b_student};
      my ($first,$last) = $a->name($b_student);
      $log = $log . "Added student $last, $first.\n";
    }
  }

  # Copy categories:
  my $b_cats = $b->category_array();
  foreach my $b_cat(@$b_cats) {
    if (!($a->category_exists($b_cat))) {
      $a->add_category($b_cat);
      $a->categories_private_method->{$b_cat} = $b->categories_private_method->{$b_cat}; # copy properties
      $log = $log . "Added category ".$a->category_name_plural($b_cat).".\n";
    }
  }

  # Copy assignments:
  my $cats = $a->category_array();
  foreach my $cat(@$cats) {
    if ($b->category_exists($cat)) {
      my $a_assignments = $a->array_of_assignments_in_category($cat); # array ref
      my $b_assignments = $b->array_of_assignments_in_category($cat);
      foreach my $b_ass(@$b_assignments) {
        if (!($a->assignment_exists("$cat.$b_ass"))) {
          $a->add_assignment(CATEGORY=>$cat,ASS=>$b_ass,PROPERTIES=>($b->assignments_private_method->{".$cat.$b_ass"}));
          $log = $log . "Added assignment ".$a->assignment_name($cat,$b_ass).".\n";
        }
      }
    }
  }

  # Copy scores:
  foreach my $cat(@$cats) {
    my $assignments = $a->array_of_assignments_in_category($cat); # array ref
    foreach my $ass(@$assignments) {
      foreach my $student(sort ($a->student_keys('all'))) {
        my $key = "$cat.$ass";
        if ($b->category_exists($cat) && $b->assignment_exists($key)) {
          my $a_grade = $a->get_current_grade($student,$cat,$ass);
          my $b_grade = $b->get_current_grade($student,$cat,$ass);
          my $ass_name = $a->assignment_name($cat,$ass);
          my ($first,$last) = $a->name($student);
          my $copy = 0;
          if ($a_grade eq '' && $b_grade ne '') {
            $log = $log . "Added grade $b_grade on assignment $ass_name for $last, $first.\n";
            $copy = 1;
          }
          if ($a_grade ne '' && $b_grade ne '' && $a_grade!=$b_grade) {
            if (&$ask($ass_name,"$last, $first",$a_grade,$b_grade)) {$copy = 1}
            if ($copy) {$log = $log . "Changed grade from $a_grade to $b_grade on assignment $ass_name for $last, $first.\n"}
          }
          if ($copy) {
            $a->set_grades_on_assignment(CATEGORY=>$cat,ASS=>$ass,GRADES=>{$student=>$b_grade});
          }
				}
      }
    }
  }
  
  return $log;

}

=head3 clone()

Make an identical copy of the GradeBook object, but sets its filehandle and file_name to undef. The
list of assignments and student roster are copied.

=cut

sub clone {
  my $self = shift;
  my $clone = GradeBook->Clone();
  $clone->filehandle(undef);
  $clone->file_name(undef);
}

=head3 filehandle()

Sets or gets a reference to the filehandle being used to access the file. Returns undef if there
is no filehandle associated with the file. The only purpose of maintaining this filehandle is
so we can keep a lock on the file.

=cut
sub filehandle {
  my $self = shift;
  if (@_) {$self->{FILEHANDLE} = shift}
  return $self->{FILEHANDLE};
}

=head3 close()

Close the file associated with this GradeBook's filehandle. This also releases the lock.
This does NOT save the file or anything like that. The only real reason you want to call
this routine is to release the lock on the file, or to make sure Windows will not be upset
because it's been left open.

=cut

sub close {
  my $self = shift;
  if (ref $self->{FILEHANDLE}) {
    close($self->filehandle()); # releases lock
  }
}

=head3 lock()

Try to lock the file associated with this GradeBook's filehandle. Pass back the return value.
The unlocking happens automatically when the file is closed.

=cut
sub lock {
  my $self = shift;
  if (Portable::os_type() ne 'windows') {
    return flock($self->filehandle(),(2 | 4));
           # ... 2=exclusive, 4=non-breaking.
           # Can't use symbolic names because of use strict. 
  }
  else {
    return 1;
    # If we're on Windows, don't try to do file locking. I don't understand how Perl implements
    # file locking on Windows, but it seemed to keep me from being able to
    # read files. This may be related to the fact that Perl on Windows is pickier than Unix
    # about opening a file again without closing it.
  }
}

=head3 read()

GradeBook->read("foo.gb",$password) creates a new GradeBook object by reading a file.
In case of an error, returns a string containing an error message.
Should call it like this: $gb=GradeBook->read(...); if (!ref $gb) {..error..}
This also sets filehandle() and locks the file.

=cut

sub read {
    my $class = shift;
    my $file_name = shift;
    my $password = "";
    if (@_) {$password = shift}
    my $self = {};
    bless($self,$class);
    $self->password($password);
    $self->modified(0);
    $self->file_name($file_name);
    my $buff = "";
    my $mode = ""; # e.g. after encountering "grades", we expect some grades
    my %recognized_modes = ('class'=>1,'preferences'=>1,'categories'=>1,'roster'=>1,
                            'assignments'=>1,'grades'=>1);
    my $mode_is_recognized; # anything not recognized is just preserved without parsing it
    my $one_liner = ""; # Some modes are one-liners.
    my @names; # for use in roster mode
    my $children={};
    my $mac = LineByLine->new(KEY=>$password,INPUT=>$file_name);
    $self->{AUTHENTICITY} = $mac->authenticity(); # This line has to come before the open(), because otherwise we have the file open twice, and that makes
                                                  # it not work on Windows.
    open (INFILE,"<$file_name") or return "Cannot open $file_name";
    $self->filehandle(\*INFILE); # Perl idiom for getting a reference to a filehandle, see camel book, p. 249
    $self->lock() or return "Unable to obtain a lock on file $file_name. This can happen if you are running two copies of OpenGrade at once.";
    LINE: while (<INFILE>) {
        next LINE unless $_ = $mac->strip_line($_); # gets rid of MAC part
        tr/\r/\n/;    # Mac newlines must die.
        chomp;        # Get rid of trailing newline.
        s/\#[^\"]*//; # Get rid of comments.
        s/\s+$//;     # Get rid of trailing whitespace.
        s/^\s+//;     # Get rid of leading whitespace.
        next LINE unless $_;
        $buff = $buff . $_;
        if (substr($buff,-1) ne ",") {
          #print "line: $buff\n";
          if ($buff) {
              if (substr($buff,0,1) ne ".") { #mode line
                  flush_children($self,$children);
                  $buff =~ m/^([a-zA-Z_\-]+)\s*(.*)/;
                  $mode = $1;
                  $children={"."=>$mode};
                  my $tail = $2;
                  $mode = lc($mode);
                  $mode_is_recognized = exists $recognized_modes{$mode};
                  $one_liner = "";
                  my $children = {}; # for later lines under this mode
                  $tail =~ s/\"//g;
              } # end if mode line
              else { #not mode line
                  $buff =~ m/^\.([a-zA-Z_\-\.\d]+)\s*(.*)/;
                  my ($head,$tail) = ($1,$2);
                  if ($mode eq "grades") {$tail =~ s/\"[^\:]+\:\?\",?//g} # MicroGrade uses ?
                  $head = lc($head);
                  $children->{$head}=$tail;
                  if ($mode eq "assignments") {$self->assignment_list($self->assignment_list.",".$head);}
                  if ($mode eq "categories") {$self->category_list($self->category_list.",".$head);}
              }
              if (!$mode_is_recognized) {
                $self->stuff_unable_to_parse($buff);
              }
          } # end if not an empty line
          $buff = "";
        } # end if done with line plus continuation lines
    } # end loop over input lines
    flush_children($self,$children);
    $self->use_defaults_for_assignments();
    #close INFILE; # Used to close after reading. No longer do this, because we want locking.
    $self->misc_initialization();
    return $self;
}

sub misc_initialization {
    my $self = shift;
    $self->{WROTE_TILDE_FILE} = 0;
    $self->preferences(Preferences->new(GB=>$self));
}

sub preferences {
    my $self = shift;
    if (@_) {$self->{PREFERENCES} = shift}
    return $self->{PREFERENCES};
}

sub file_name {
    my $self = shift;
    if (@_) {
      $self->{FILE_NAME} = shift;
    }
    return $self->{FILE_NAME};
}

sub modified {
    my $self = shift;
    if (@_) {$self->{MODIFIED} = shift}
    return $self->{MODIFIED};
}

sub password {
    my $self = shift;
    if (@_) {$self->{PASSWORD} = shift}
    return $self->{PASSWORD};
}

sub pref {
    my $self = shift;
    my $what = shift;
    my $it = ($self->preferences_private_method())->{$what};
    $it =~ m/^\"([^\"]*)\"$/;
    return $1;
}

sub boolean_pref {
    my $self = shift;
    my $what = shift;
    return ($self->pref($what) eq "true");
}


=head3 stuff_unable_to_parse()

A read-write method. Returns undef if it has never been set. The write method
appends the argument, plus a newline, to what's already there.

The idea here is to ensure forward-compatibility by preserving
any sections in the gradebook file that we were unable to interpret.

=cut

sub stuff_unable_to_parse {
    my $self = shift;
    if (@_) {
      my $new_stuff = shift;
      if (! exists $self->{STUFF_UNABLE_TO_PARSE}) {$self->{STUFF_UNABLE_TO_PARSE} = ''}
      $self->{STUFF_UNABLE_TO_PARSE} = $self->{STUFF_UNABLE_TO_PARSE} . "$new_stuff\n";
    }
    if (! exists $self->{STUFF_UNABLE_TO_PARSE}) {return undef}
    return $self->{STUFF_UNABLE_TO_PARSE};
}

=head3 set_grades_on_assignment()

Add new grades. Any preexisting ones are retained.
      CATEGORY => "",
      ASS => "",
      GRADES => {}, 

=cut

sub set_grades_on_assignment {
    my $self = shift;
    my %args = (
      CATEGORY => "",
      ASS => "",
      GRADES => {}, # hash ref
      @_,
    );
    my $category = $args{CATEGORY};
    my $ass = $args{ASS};
    my $new_grades = $args{GRADES};
    my $grades = $self->grades_private_method();
    while (my ($student,$score) = each(%$new_grades)) {
        #print "set_grades_on_assignment $ass,$student,$score\n";
        my $c = $student.".".$category;
        my $record = "\"$ass:$score\"";
        if (exists($grades->{$c})) {
            $grades->{$c} = set_property($grades->{$c},$ass,$score);
        }
        else {
            $grades->{$c} = $record;
        }
    }
    $self->grades_private_method($grades);
}

=head3 drop_student()

Takes one argument, the key of the student to be dropped.

=cut

sub drop_student {
    my $self = shift;
    my $who = shift; # student's key
   $self->set_student_property($who,"dropped","true");
}

=head3 reinstate_student()

Takes one argument, the key of the student to be reinstated.
If the key is a null string, does nothing.

=cut


sub reinstate_student {
    my $self = shift;
    my $who = shift; # student's key
    $self->set_student_property($who,"dropped","false") unless $who eq "";
}



=head3 set_student_property()

  set_student_property($key,$property,$value)

=cut

sub set_student_property {
    my $self = shift;
    my $who = shift; # student's key
    my $prop = shift;
    my $value = shift;
    my $students = $self->roster_private_method();
    $students->{$who} = set_property($students->{$who},$prop,$value);
    $self->roster_private_method($students);
}


=head3 get_student_property()

  get_student_property($key,$property)

=cut

sub get_student_property {
    my $self = shift;
    my $who = shift; # student's key
    my $prop = shift;
    return get_property($self->roster_private_method()->{$who},$prop);
}

=head3 get_property()

Avoid using this routine, and use get_property2() instead.
Same as get_property2(), but
returns a null string if the parameter doesn't exist.
This was bad design, since it also
returns a null string if the parameter has a null-string value -- yech.
This is only here in order to keep from breaking old code that depends
on this behavior.

=cut

sub get_property {
    my @args = @_;
    my $x = get_property2(@args);
    if (defined $x) {
      return $x;
    }
    else {
      return '';
    }
}


=head3 get_property2()

Given a string like "x:1","y:2", and a param like "x", return 1.
First argument is the string. Second argument is the key, e.g., "x".
Returns undef if the parameter doesn't exist. This is a plain function,
not a GradeBook method.

=cut

sub get_property2 {
    my $x = shift;
    my $key = shift;
    my @p = split_comma_delimited_values($x);
    foreach my $p(@p) {
        $p =~ m/\"?([^\:]+)\:([^\"]*)\"?/;
        my ($a,$b) = ($1,$2);
        if ($a eq $key) {return $b}
    }
    return undef;
}

=head3 delete_property()

Given a string like "x:1","y:2", and a param like "x", delete the part
of the string referring to x.
First argument is the string. Second argument is the key, e.g., "x".
Returns the original string if the parameter doesn't exist. This is a plain function,
not a GradeBook method.

=cut

sub delete_property {
    my $x = shift;
    my $key = shift;
    $x =~ s/\"$key\:([^\"]*)\"\,?//g;
    $x =~ s/\,$//; # If the property was the last one, we need to delete the comma before it.
    return $x;
}

=head3 property_exists()

Given a string like "x:1","y:2", and a param like "x", tells whether
x occurs in the string.

=cut

sub property_exists {
  return defined get_property2(@_);
}

=head3 get_properties()

Given a hash with keys like "x:1","y:2", extract a particular
key, and make it into a hash ref like {x=>1, y=>2}.

=cut

sub get_properties {
    my $self = shift;
    my $which = shift;
    my $c = shift;
    my @p = split_comma_delimited_values($c->{$which});
    my %result = ();
    foreach my $p(@p) {
        $p =~ m/\"?([^\:]+)\:([^\"]*)\"?/;
        my ($a,$b) = ($1,$2);
        $result{$a} = $b;
    }
    return \%result;
}


=head3 set_property()

  set_property($list,$property,$value)

This just returns a modified version of the original list -- doesn't modify it in place.

=cut

sub set_property {
    my $stuff = shift; # .foo "bar:37","glub:92"
    my $prop = shift;
    my $value = shift;
    if ($stuff =~ m/\"$prop\:[^\"]*\"/) {
        $stuff =~ s/\"$prop\:[^\"]*\"/\"$prop\:$value\"/;
    }
    else {
        $stuff = $stuff . ",\"" . $prop . ":" . $value . "\"";
    }
    return $stuff;
}


=head3 heritable_from_cat()

Some properties are inherited by assignments
from the parent category. See use_defaults_for_assignments.
This is a plain old function, not a method for GradeBook objects.
See also list_properties_heritable_from_cat().

=cut

sub heritable_from_cat {
  my $what = shift;
  return ($what eq "max" || $what eq "ignore");
}

sub list_properties_heritable_from_cat {
  return ('max','ignore');
}

=head3 use_defaults_for_assignments()

Goes through the list of assignments, and supplies defaults
from the parent category's properties.

=cut


sub use_defaults_for_assignments {
    my $self = shift;
    my @a = split(",",$self->assignment_list());
    my @c = split(",",$self->category_list());
    foreach my $c(@c) {
        my $cp = $self->category_properties($c);
        foreach my $p(keys(%$cp)) {
            if (!heritable_from_cat($p)) {
                delete($cp->{$p});
            }
        }
        my %cp = %$cp;
        foreach my $a(@a) {
            if (first_part_of_label($a) eq $c) {
                my $ap = $self->assignment_properties($a);
                my %ap = %$ap;
                %ap = (%cp,%ap); # merge the hashes
                $self->assignment_properties($a,\%ap);
            }
        }
    }
}

=head3 strip_redundant_properties

First argument is the assignment key. Second argument is its list of properties.
This is meant for use only when writing the gradebook file. If an assignment
has properties that are the defaults for its category, we strip those out.
All this does is make the gradebook file a little cleaner and more readable.
When we read the file back in, the defaults are re-applied
by use_defaults_for_assignments().

=cut

sub strip_redundant_properties {
  my $self = shift;
  my $key = shift;
  my $props = shift;
  my ($cat,$ass) = $self->split_cat_dot_ass($key);
  foreach my $prop(list_properties_heritable_from_cat()) {
    my $default = $self->category_property2($cat,$prop);
    my $actual = get_property2($props,$prop);
    if ((defined $actual) && ($actual eq $default)) {
      $props = delete_property($props,$prop);
    }
  }
  return $props;
}

=head3 split_cat_dot_ass()

Takes one argument, of the form "cat.ass", and returns a list
(cat,ass).

=cut

sub split_cat_dot_ass {
  my $self = shift;
  my $x = shift;
  $x =~ m/^([^.]*)\.([^.]*)$/;
  my ($cat,$ass) = ($1,$2);
  return ($cat,$ass);
}

=head3 student_keys()

Returns an array, sorted in alphabetical order. Bug:
probably doesn't work right when one student's name is
the same as the beginning of another's.

 Optional argument:
   ""          only return active students (default)
   "all"       return all students
   "dropped"   return a list of dropped students

=cut

sub student_keys {
    my $self = shift;
    my $criteria = "";
    if (@_) {$criteria = shift}
    my $h = $self->roster_private_method();
    my @raw = sort keys %$h;
    
    my @s = ();
    foreach my $s(@raw) {
        my $is_dropped = "";
        my $info = $h->{$s};
        my $is_dropped = get_property($info,"dropped") eq "true";
        if ($is_dropped && $criteria ne "" or !$is_dropped && $criteria ne "dropped") {
            push @s,$s;
        }
    }

    return @s;
}

sub first_part_of_label {
    my $x = shift;
    $x =~ m/^([^\.]*)\./;
    return $1;
}

sub second_part_of_label {
    my $x = shift;
    $x =~ m/^([^\.]*)\.([^\.]*)/;
    return $2;
}

=head3 get_current_grade()

get_current_grade($student_key,$cat,$ass)

=cut

# Returns null if no grade recorded.
sub get_current_grade {
    my $self = shift;
    my $student_key = shift;
    my $cat = shift;
    my $ass = shift;
    my $h = $self->grades_private_method();
    if (!exists($h->{$student_key.".".$cat})) {return ""}
    my $list = $h->{$student_key.".".$cat};
    return get_property($list,$ass);
}

=head3 weights_enabled()

Shouldn't normally be necessary to call this routine, except
when setting up a new category to see if we need to ask them
for a weight. It's inefficient, so avoid calling it.
Return values:
  0 unweighted
  1 weighted
  2 no categories have been put in yet, so could be either

=cut

sub weights_enabled {
    my $self = shift;
    my $h = $self->categories_private_method();
    my $any_cats = 0;
    foreach my $c(keys(%$h)) {
        if (get_property($h->{$c},"weight") ne "") {return 1}
        $any_cats = 1;
    }
    if (!$any_cats) {return 2}
    return 0;
}

# Returns 100 if this weights are not enabled. Otherwise
# returns the weight.
sub category_weight {
    my $self = shift;
    my $which = shift;
    my $h = $self->categories_private_method();
    if (! exists($h->{$which})) {return ""}
    my $w = get_property($h->{$which},"weight");
    if ($w ne "") {return $w}
    return 100;
}

=head3 category_name_singular()

Takes one argument, the category key.

=cut

sub category_name_singular {
    my $self = shift;
    my $which = shift;
    my $h = $self->categories_private_method();
    if (! exists($h->{$which})) {return ""}
    my $n = get_property($h->{$which},"catname");
    $n =~ m/([^\,]+),(.*)/;
    my ($sing,$pl) = ($1,$2);
    return $sing;
}

=head3 category_name_plural()

Takes one argument, the category key.

=cut

sub category_name_plural {
    my $self = shift;
    my $which = shift;
    my $h = $self->categories_private_method();
    if (! exists($h->{$which})) {return ""}
    my $n = get_property($h->{$which},"catname");
    $n =~ m/([^\,]+),(.*)/;
    my ($sing,$pl) = ($1,$2);
    return $pl;
}

=head3 category_exists()

Takes one argument, the category key.

=cut

sub category_exists {
    my $self = shift;
    my $which = shift;
    my $h = $self->categories_private_method();
    return exists($h->{$which});
}

=head3 assignment_exists()

Takes one argument, cat.assmt.

=cut

sub assignment_exists {
    my $self = shift;
    my $which = shift;
    my $h = $self->assignments_private_method();
    return exists($h->{$which});
}

=head3 category_property_boolean()

Arguments are category, property. This is a read-only method.

=cut

sub category_property_boolean {
  my $self = shift;
  my $x = $self->category_property(@_);
  return (lc($x) eq 'true');
}

=head3 category_property2()

To get a property, arguments are category, property; returns undef
if there's no such property.
To set a property, arguments are category, property, value.

=cut

sub category_property2 {
    my $self = shift;
    my $which = shift;
    my $prop = shift;
    my $h = $self->categories_private_method();
    if (!exists($h->{$which})) {return undef}
    if (@_) {
      my $value = shift;
      $h->{$which} = set_property($h->{$which},$prop,$value);
    }
    return get_property2($h->{$which},$prop);
}

=head3 category_property()

Avoid using this. Use category_property2() instead. This is only
here for the sake of old code that depends on receiving a null
string when the property isn't found.

=cut

sub category_property {
    my $self = shift;
    my $which = shift;
    my $prop = shift;
    my $h = $self->categories_private_method();
    if (!exists($h->{$which})) {return ""}
    if (@_) {
      my $value = shift;
      $h->{$which} = set_property($h->{$which},$prop,$value);
    }
    return get_property($h->{$which},$prop);
}

sub category_properties_comma_delimited {
    my $self = shift;
    my $which = shift;
    my $h = $self->categories_private_method();
    return $h->{$which};
}

# Get or set the properties of the category as a hash reference.
# category must already exist.
sub category_properties {
    my $self = shift;
    my $which = shift;
    if (@_) {
        my $h = shift;
        my $x = hash_to_comma_delimited($h);
        my $c = $self->categories_private_method();
        $c->{$which} = $x;
        $self->categories_private_method($a);
    }
    return $self->get_properties($which,$self->categories_private_method());
}


=head3 add_assignment()

 add_assignment {
        CATEGORY => "",
        ASS => "",          # assmt name only, not category.assmt
        PROPERTIES=>"",     # comma-delimited list
        COMES_BEFORE=>"",   # if null, put it at end of list for this category
        @_,
    );

=cut

sub add_assignment {
    my $self = shift;
    my %args = (
        CATEGORY => "",
        ASS => "",          # assmt name only, not category.assmt
        PROPERTIES=>"",     # comma-delimited list
        COMES_BEFORE=>"",   # if null, put it at end of list for this category
        @_,
    );
    local $words_prefix = "add_assignment";
    my $category = lc($args{CATEGORY});
    my $name = lc($args{ASS});
    my $properties = $args{PROPERTIES};
    my $comes_before = $args{COMES_BEFORE};
    $name =~ s/[ :"]/_/g; # Otherwise you lose data!
    my @a = split(",",$self->assignment_list());
    my $comes_before_index = -999;
    if (!@a) {
      $comes_before_index=0
    }
    else {
    if ($comes_before) {
        for (my $j=0; $j<=$#a; $j++) {
            if ($a[$j] eq $category.".".$comes_before) {$comes_before_index=$j}
        }
    }
    else {
        # Typically this category already has assignments, so we add this one after them:
        for (my $j=0; $j<=$#a; $j++) {
            my $c = $a[$j];
            $c =~ m/([^\.]*)\.([^\.]*)$/;
            if ($1 eq $category) {$comes_before_index=$j+1}
        }
        # This category doesn't have any assignments yet; this will be the first one we add to it:
        if ($comes_before_index == -999) {
            my $cl = $self->category_list();
            $comes_before_index = $#a+1;
               # ...If we can't find any assignments that come after it, it's the last. This happens
               # if it's the last assignment in the last category, or if all the later categories are empty.
            for (my $j=$#a; $j>=0; $j--) {
              my $c = $a[$j];
              $c =~ s/\..*//;
              if ($cl =~ m/$category\,([^,]+\,)*$c/) {$comes_before_index=$j}
            }

        }
    }
    }
    if ($comes_before_index == -999) {return w('where_to_insert')}
    for (my $j=$#a+1; $j>=$comes_before_index+1; $j--) {
        $a[$j] = $a[$j-1];
    }
    my $full_name = $category.".".$name;
    my $h = $self->assignments_private_method();
    if (exists($h->{$full_name})) {return (sprintf w('exists'),$name)}
    $a[$comes_before_index] = $full_name;
    $self->assignment_list(@a);
    $h->{$full_name} = $properties;
    $self->assignments_private_method($h);
    $self->use_defaults_for_assignments();
    return "";
}

=head3 delete_assignment()

 $gb->delete_assignment("exam.1");

Returns null on success.

=cut

sub delete_assignment {
    my $self = shift;
    my $key = shift;

    my $h = $self->assignments_private_method();
    if (!exists($h->{$key})) {return "Assignment $key does not exist."}
    delete $h->{$key};
    $self->assignments_private_method($h);

    my @a = $self->assignment_array();
    my $victim = $#a+999;
    for (my $i=0; $i<=$#a; $i++) {
      if ($a[$i] eq $key) {$victim=$i; last}
    }
    for (my $i=$victim; $i<=$#a-1; $i++) {
      $a[$i] = $a[$i+1];
    }
    $#a = $#a-1; # reduce size by 1
    @a = $self->assignment_array(@a);

    $key =~ m/^([^.]*)\.([^.]*)$/;
    my $category = $1;
    my $which = $2;
    my $g = $self->grades_private_method();
    foreach my $student_and_cat(keys %$g) {
      if ($student_and_cat =~ m/$category$/) {
        $g->{$student_and_cat} =~ s/\"$which\:[^"]*\"\,?//;
        $g->{$student_and_cat} =~ s/\,$//;
      }
    }
    $self->grades_private_method($g);

    return "";
}



=head3 assignment_name()

If given only one argument, it's assumed to be category.assignment.
If given two arguments, they're assumed to be the category and the assignment.

=cut

sub assignment_name {
  my $self = shift;
  my $x = shift;
  my ($cat,$ass);
  if (@_) {
    $cat = $x;
    $ass = shift;
  }
  else {
    $x =~ m/(.*)\.(.*)/;
    ($cat,$ass) = ($1,$2);
  }
  if ($self->assignment_property_exists("$cat.$ass",'name')) {
    return $self->assignment_property("$cat.$ass",'name');
  }
  else {
    return $self->default_assignment_name($cat,$ass);
  }
}


=head3 default_assignment_name()

Arguments are the category and the assignment. This
returns the name the assignment would have simply based on
its category and database key, regardless of whether it has
a "name" property. It would be a bug to call this routine just
to find out a name to display -- that's what assignment_name()
is for.

=cut

sub default_assignment_name {
  my $self = shift;
  my $cat = shift;
  my $ass = shift;
  my $cat_name = $self->category_name_singular($cat);
  if (!$self->category_property_boolean($cat,'single')) {
    return "$cat_name $ass";
  }
  else {
    return $cat_name;
  }
}


=head3 rekey_assignment()

 $gb->rekey_assignment($category,$old,$new);

Returns a null string on success. This only changes the database key.
If the assignment has a "name" tag, then the display name is unaffected.
If the assignment doesn't have a "name" tag, then the display name is
affected.

=cut

sub rekey_assignment {
    my $self = shift;
    my $category = shift;
    my $old = shift;
    my $new = shift;

    my $new_key = $category.".".$new;
    my $old_key = $category.".".$old;
    my $h = $self->assignments_private_method();
    if ($new_key ne $old_key && exists($h->{$new_key})) {return "Assignment $new already exists"}
    my $stuff = $h->{$old_key};
    delete $h->{$old_key};
    $h->{$new_key} = $stuff;
    $self->assignments_private_method($h);

    my @a = $self->assignment_array();
    for (my $i=0; $i<=$#a; $i++) {
      if ($a[$i] eq $old_key) {$a[$i]=$new_key}
    }
    @a = $self->assignment_array(@a);

    my $g = $self->grades_private_method();
    foreach my $key(keys %$g) {
      if ($key =~ m/$category$/) {
        $g->{$key} =~ s/$old\:/$new\:/;
      }
    }
    $self->grades_private_method($g);

    return "";
}

=head3 assignment_property()

First argument is cat.assmt.
Second argument is name of property.
This is a read-only method.
If you're going to extract more than one property, it's
more efficient to call assignment_properties() rather
than this routine.

=cut

sub assignment_property {
    my $self = shift;
    my $which = shift;
    my $prop = shift;
    my $h = $self->assignments_private_method();
    if (!exists($h->{$which})) {return ""}
    return get_property($h->{$which},$prop);
}

=head3 assignment_property_exists()

Like assignment_property(), but tells whether the
property exists.

=cut

sub assignment_property_exists {
    my $self = shift;
    my $which = shift;
    my $prop = shift;
    my $h = $self->assignments_private_method();
    if (!exists($h->{$which})) {return 0}
    return defined get_property2($h->{$which},$prop);
}

=head3 assignment_properties()

Get or set the properties of the assignment as a hash reference.
First argument is cat.assmt. Optional second argument is 
hash ref.

=cut

sub assignment_properties {
    my $self = shift;
    my $which = shift;
    if (@_) {
        my $h = shift;
        my $x = hash_to_comma_delimited($h);
        my $a = $self->assignments_private_method();
        $a->{$which} = $x;
        $self->assignments_private_method($a);
    }
    return $self->get_properties($which,$self->assignments_private_method());
}

=head3 add_student()

                LAST=>"",
                FIRST=>"",
                KEY=>"",
                ID=>"",

Returns ($new_key,$error). You can provide the key,
the last and first names, or both.


=cut


sub add_student {
    my $self = shift;
    my %args = (
                LAST=>"",
                FIRST=>"",
                KEY=>"",
                ID=>"",
                @_,
                );
    my $last = $args{LAST};
    my $first = $args{FIRST};
    my $key = $args{KEY};
    my $id = $args{ID};
    if ($last eq "" && $first eq "" && $key eq "") {return ""}
    if ($key eq "") {$key = $last."_".$first}
    $key = lc($key);
    $key =~ s/[^a-z\-_]//g;
    $key=~m/_(.*)/;
    $first = $1;
    $key=~m/([^_]*)/;
    $last = $1;
    my $stuff = "";
    if (ucfirst($args{FIRST}) ne ucfirst($first)) {
        $stuff = $stuff . ',"first:' . $args{FIRST} . '"';
    }
    if (ucfirst($args{LAST}) ne ucfirst($last)) {
        $stuff = $stuff . ',"last:' . $args{LAST} . '"';
    }
    if ($id ne "") {
        $stuff = $stuff . ',"id:' . $id . '"';
    }
    $stuff =~ s/^,//;
    my $r = $self->roster_private_method();
    if (exists($r->{$key})) {return ('','already_exists')}
    $r->{$key} = $stuff;
    $self->roster_private_method($r);
    return ($key,'');
}

# given a student's key, returns the id
sub id {
    my $self = shift;
    my $key = shift;
    my $h = $self->roster_private_method();
    if (!exists($h->{$key})) {return ()}
    my $info = $h->{$key};
    return get_property($info,"id");
}

=head3 compare_names()

Given two students' keys, returns a <=>-style comparison of their names,
in alphabetical order by last name. Punctuation gets stripped out, and
case is ignored.

=cut

sub compare_names {
    my $self = shift;
    my $key1 = shift;
    my $key2 = shift;
    my ($first1,$last1) = $self->name($key1);
    my ($first2,$last2) = $self->name($key2);
    $last1 = strip_name($last1);
    $last2 = strip_name($last2);
    if ($last1 ne $last2) {return $last1 <=> $last2}
    $first1 = strip_name($first1);
    $first2 = strip_name($first2);
    return $first1 <=> $first2;
}

sub strip_name {
    my $name = shift;
    $name = lc($name);
    $name =~ s/[^\w]//g;
    return $name;
}

=head3 name()

Given the student's key, returns an array containing first and last name.

=cut

sub name {
    my $self = shift;
    my $key = shift;
    my $h = $self->roster_private_method();
    if (!exists($h->{$key})) {return ()}
    my $info = $h->{$key};
    my $first = get_property($info,"first");
    if (!$first) {$key=~m/_(.*)/; $first = $1}
    my $last = get_property($info,"last");
    if (!$last) {$key=~m/([^_]*)/; $last = $1}
    $first = ucfirst($first);
    $last = ucfirst($last);
    return ($first,$last);
}


sub flush_children {
    my $self = shift;
    my $h = shift;
    my $what = $h->{"."};
    delete($h->{"."});
    while ( my ($k,$v)=each %$h) {
        $v =~ s/,+/,/g;
        $v =~ s/^,//;
        if ($what eq "grades") {
            # Change keys to lc. Otherwise keys may mismatch in grade records, e.g. Exam:97 rather than exam:97.
          # We only do this for grades. It's not appropriate in general because, for instance, it would lead to
          # letter grades being forced to lower case in the standards section.
          my $lcv = $v;
          while ($v =~ m/(\"[A-Za-z]+\:)/g) {
            my $x = $1;
            my $y = lc($x);
            $lcv =~ s/$x/$y/;
          }
          $v = $lcv;
        }
        $h->{$k} = $v;
    }
    #print "flushing\n";
    #show_hash($what,$h);
    if ($what eq "class") {$self->class_private_method($h);}
    if ($what eq "preferences") {$self->preferences_private_method($h);}
    if ($what eq "categories") {$self->categories_private_method($h);}
    if ($what eq "roster") {$self->roster_private_method($h);}
    if ($what eq "assignments") {$self->assignments_private_method($h);}
    if ($what eq "grades") {$self->grades_private_method($h);}
}

# Change "Newton, Isaac" to "newton_isaac",
#        "Isaac Newton" to "newton_isaac",
#        "Franklin D. Roosevelt" to "roosevelt_franklin_d"
sub name_to_label {
    my $x = shift;
    my ($name,$explicit) = split(/:/,$x);
    my $result = "";
    if ($explicit) {$result= $explicit}
    # lastname, firstname:
    if (!$result) {
        if ($name =~ m/^\s*([^,])\s*,\s*(.*)$/) {
            $result = $1."_".$2;
        }
    }
    # firstname lastname
    if (!$result) {
        my $r = $name;
        $r = reverse($r);
        $r =~ m/^\s*([^\s]+)\s*(.*)$/;
        my ($lastrev,$firstrev) = ($1,$2);
        $firstrev =~ s/\s/_/g;
        $firstrev =~ s/_+$//;
        $firstrev =~ s/^_+//;
        $result = reverse($lastrev)."_".reverse($firstrev);
    }
    # Get rid of all whitespace:
    $result =~ tr/[\s",\.]//;
    $result = lc($result);
    return $result;
}

# pass it a hash ref
sub hash_to_comma_delimited {
    my $h = shift;
    my %h = %$h;
    my $result = "";
    my @k = sort keys(%h); # Sorting never hurts, and sometimes helps.
    foreach my $p(@k) {
        $result = $result . "\"" . $p . ":" . $h{$p} . "\",";
    }
    $result =~ s/,$//;
    return $result;
}

# returns a hash (not a hash ref)
sub comma_delimited_to_hash {
    my @a = split_comma_delimited_values(shift);
    my %result = ();
    foreach $a(@a) {
        $a =~ m/([^\:]*)\:(.*)/;
        $result{$1} = $2;
    }
    return %result;
}

# The following is nontrivial because there may be delimiters inside quotes.
sub split_comma_delimited_values { # from Perl Cookbook, p. 31
    my $text = shift;
    my @new = ();
    #$text =~ s/,+/,/g; # Removed because could have ,, inside quotes.
    $text =~ s/^,//;
    $text =~ s/,$//;
    push(@new,$+) while $text =~ m{
        "([^\"\\]*(?:\\.[^\"\\]*)*)",?
            | ([^,]+),?
            | ,
            }gx;
    return @new;
}

sub debug_print {
    my $self = shift;
    print "title: ".$self->title()."\n";
    print "staff: ".$self->staff()."\n";
    print "days:  ".$self->days()."\n";
    print "term:  ".$self->term()."\n";
    print "dir:   ".$self->dir()."\n";
    show_hash("categories",$self->categories_private_method(),$self->category_list());
    show_hash("roster",$self->roster_private_method());
    show_hash("assignments",$self->assignments_private_method(),$self->assignment_list());
    show_hash("grades",$self->grades_private_method());
}

sub show_hash {
    my $title = shift;
    my $h = shift;
    my %h = %$h;
    my @k = keys(%h); 
    if (@_) {
        @k = split(/,/,shift);
    }

    print "$title------------------------------------------\n";
    foreach my $k(@k) {print "  $k => $h{$k}\n";}
    print "-----------\n";
}

=head3 class_data()

Get or set a property from the class data. The argument tells which property we want to get.
The property's name is not surrounded with double quotes. To get a property, call with one
argument. To set a property, give a second argument as well.

=cut

sub class_data {
    my $self = shift;
    my $what = shift;
    my $h = $self->class_private_method();
    if (@_) {my $val = shift; $h->{$what}='"'.$val.'"'}
    if (!exists($h->{$what})) {return ""}
    my $result = $h->{$what};
    $result =~ s/^\"([^\"]*)\"$/$1/;
    return $result;
}

sub set_standards {
  my $self = shift;
  my $h = $self->class_private_method();
  $h->{'standards'} = shift;
}

sub title {
    my $self = shift;
    return $self->class_data("title");
}

sub staff {
    my $self = shift;
    return $self->class_data("staff");
}

sub days {
    my $self = shift;
    return $self->class_data("days");
}

sub term {
    my $self = shift;
    return $self->class_data("term");
}

=head3 dir()

Get or set the subdirectory for web grades.

=cut

sub dir {
    my $self = shift;
    return $self->class_data("dir",@_);
}




# Returns a hash ref like {"A"=>"90",...}
sub standards {
    my $self = shift;
    my %h = comma_delimited_to_hash($self->class_data("standards"));
    return \%h;
}

sub percentage_to_letter_grade {
    my $self = shift;
    my $pct = shift;
    my $h = $self->standards();
    # Find the highest percentage that this grade meets or exceeds.
    my $highest = -999;
    my $highest_letter = "";
    foreach my $letter(keys(%$h)) {
        my $min = $h->{$letter};
        if ($pct>=$min && $min>$highest) {
            $highest = $min;
            $highest_letter = $letter;
        }
    }
    return $highest_letter;
}

# Returns a comma-delimited list, without the category name on the front of the
# assignment names.
sub assignments_in_category {
    my $self = shift;
    my $cat = shift;
    my $a = $self->assignment_list();
    $a = ",".$a; # so first one gets parsed correctly
    $a =~ s/\,$cat\.([^\,]+)/,$1/g; # strip cat names; comma ensures hw doesn't match w
    $a =~ s/[^\.\,]+\.[^\.\,]+\,?//g; # get rid of assmts from other cats
    $a =~ s/,$//;
    $a =~ s/^,//;
    return $a;
}



=head3 category_contains_assignments()

Takes a category key as its argument.

=cut

sub category_contains_assignments {
  my $self = shift;
  my $cat = shift;
  my $assignments = $self->array_of_assignments_in_category($cat);
  if (@$assignments) {
    return 1;
  }
  else {
    return 0;
  }
}

=head3 number_of_assignments_in_category()

Takes a category key as its argument.

=cut

sub number_of_assignments_in_category {
  my $self = shift;
  my $cat = shift;
  my $assignments = $self->array_of_assignments_in_category($cat);
  return scalar @$assignments;
}


=head3 array_of_assignments_in_category()

Takes a category key as its argument.
Returns a reference to an array, like ["1","2",...].

=cut

sub array_of_assignments_in_category {
    my $self = shift;
    my $cat = shift;
    my @a = split(",",$self->assignments_in_category($cat));
    return \@a;
}

=head3 assignment_array()

Get or set the list of assignments. Uses arrays, not array references. Public method.

=cut

sub assignment_array {
  my $self = shift;
  if (@_) {$self->assignment_list(@_)}
  return split_comma_delimited_values($self->assignment_list());
}

=head3 assignment_list()

We keep this comma-delimited list so that we know the correct ordering of the
assignments. Can set it using a comma-delimited string or an array. This
is a private method.

=cut

sub assignment_list {
    my $self = shift;
    if (@_) {
        if ($#_ == 0) {
          $self->{ASSIGNMENT_LIST} = shift;
        }
        else {
            $self->{ASSIGNMENT_LIST} = "";
            foreach my $a(@_) {$self->{ASSIGNMENT_LIST}=$self->{ASSIGNMENT_LIST}.",".$a}
        }
        $self->{ASSIGNMENT_LIST} =~ s/^,//
    }
    return $self->{ASSIGNMENT_LIST};
}

sub set_all_category_properties {
    my $self = shift;
    my $key = shift;
    my $stuff = shift;
    $key = lc($key);
    if (! $self->category_exists($key)) {return 0}
    my $c = $self->categories_private_method();
    $c->{$key} = $stuff;
    $self->categories_private_method($c);
    return 1;
}

=head3 add_category()

Arguments: key, comma-delimited properties string, weight or null

Returns 1 on success, 0 on error.

=cut

sub add_category {
    my $self = shift;
    my $key = shift;
    my $stuff = shift;
    my $w = shift; # null string if not using weights
    $key = lc($key);
    if ($self->category_exists($key)) {return 0}
    $self->category_list($self->category_list().",".$key); # null case checked for inside
    my $c = $self->categories_private_method();
    $c->{$key} = $stuff;
    $self->categories_private_method($c);
    return 1;
}

=head3 category_list()

We keep a comma-delimited list so that we know the correct ordering of the
categories. This routine gets or sets the list. (When setting the list, it
automatically cleans up any leading comma.) This is a private method. To
add a category, use the public method add_category(). To get the list of
categories, use category_array().

=cut
sub  category_list {
    my $self = shift;
    if (@_) {$self->{CATEGORY_LIST} = shift; $self->{CATEGORY_LIST} =~ s/^,//}
    return $self->{CATEGORY_LIST};
}

=head3 category_exists()

Only argument is key to check for.

=cut

sub category_exists {
    my $self = shift;
    my $key = shift;
    my $cats = $self->category_array();
    foreach my $cat(@$cats) {
      if ($cat eq $key) {return 1}
    }
    return 0;
}

=head3 category_array()

Returns a reference to an array. Can't be used to set the category list.

=cut

sub category_array {
    my $self = shift;
    my $l = $self->category_list();
    $l =~ s/,$//;
    my @a = split(",",$l);
    return \@a;
}

=head3 has_categories()

Tells whether any categories have been created.

=cut

sub has_categories {
    my $self = shift;
    my $c = $self->category_array();
    my @c = @$c;
    return ($#c >= 0);
}

=head3 set_class_data()

Get or set the class's properties using a hash reference. Keys can be
title, staff, days, term, dir, and standards. Any old properties are lost, even
those not specied in the new list.

=cut

sub set_class_data {
    my $self = shift;
    my $stuff = shift;
    $self->class_private_method($stuff);
}

=head3 class_private_method()

Get or set the class's properties using a hash reference. 

=cut

sub class_private_method {
    my $self = shift;
    if (@_) {$self->{CLASS} = shift}
    return $self->{CLASS};
}

=head3 categories_private_method()

Currently implemented as a reference to a hash whose values are the
unparsed tails of the category-setup lines.

=cut

sub categories_private_method {
    my $self = shift;
    if (@_) {$self->{CATEGORIES} = shift}
    return $self->{CATEGORIES};
}

=head3 roster_private_method()

A reference to a hash like {".newton_isaac"=>"id:1"',...}

=cut

sub roster_private_method {
    my $self = shift;
    if (@_) {$self->{ROSTER} = shift}
    return $self->{ROSTER};
}

=head3 assignments_private_method()

A reference to a hash like {".hw.1"=>"due:2/28",...}

=cut

sub assignments_private_method {
    my $self = shift;
    if (@_) {$self->{ASSIGNMENTS} = shift}
    return $self->{ASSIGNMENTS};
}

=head3 grades_private_method()

A reference to a hash like {".Newton_Isaac.hw"=>"1:12",...}

=cut

sub grades_private_method {
    my $self = shift;
    if (@_) {$self->{GRADES} = shift}
    return $self->{GRADES};
}

=head3 preferences_private_method()

A reference to a hash like {"backups_on_server"=>"true",...}

=cut

sub preferences_private_method {
    my $self = shift;
    if (@_) {$self->{PREFERENCES} = shift}
    return $self->{PREFERENCES};
}


sub iterations_for_line_by_line {
  return 25;
}

1;
