#----------------------------------------------------------------
# Copyright (c) 2002 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------



=head2 Input.pm

The Input class is a way of checking a string input by the user
to see if it's valid. It also contains information about the
default value and the prompt, which are only for convenience ---
Input's methods don't use them. Synopsis:

  $i = Input->new(
         KEY=>"file_name",
         PROMPT=>"Filename",
         DEFAULT=>"foo.bar",
         TYPE=>"numeric",MIN=>"...",MAX=>"..."
  );
  $i = Input->new(
         ...
         TYPE=>"string",BLANK_ALLOWED=>1
  );
  if ($i->check($x)) { $error_code = ($i->check($x))[0]; ...}
          # ...may have other error info in [1], ...

=cut

use strict;

package Input;


sub new {
    my $words_prefix = "b.input";
    my $class = shift;
    my @save_args = @_; # otherwise might get clobbered by w('yes') ...?
    my %args = (
      KEY=>"",
      PROMPT=>"",
      DEFAULT=>"",
      TYPE=>"numeric",
      MIN=>"",
      MAX=>"",
      BLANK_ALLOWED=>1,
      WIDGET_TYPE=>'entry', # can also be 'text' for multiline text, or 'radio_buttons'
      RADIO_BUTTONS=>{1=>Browser::get_w($words_prefix,'yes'),0=>Browser::get_w($words_prefix,'no')}, # a ref to a hash, giving the text for each value
      RADIO_BUTTON_ORDER=>[1,0],
            @save_args,
    );
    my $self = {};
    bless($self,$class);
    $self->{KEY} = $args{KEY};
    $self->{PROMPT} = $args{PROMPT};
    $self->{DEFAULT} = $args{DEFAULT};
    $self->{TYPE} = $args{TYPE};
    $self->{MAX} = $args{MAX};
    $self->{MIN} = $args{MIN};
    $self->{BLANK_ALLOWED} = $args{BLANK_ALLOWED};
    $self->{WIDGET_TYPE} = $args{WIDGET_TYPE};
    $self->{RADIO_BUTTONS} = $args{RADIO_BUTTONS};
    $self->{RADIO_BUTTON_ORDER} = $args{RADIO_BUTTON_ORDER};
    return $self;
}

sub check {
  my $self = shift;
  my $x = shift;

  #print "dumping...\n";
  #dump_str($x);

  my @stuff = ();
  if ($self->{TYPE} eq "numeric") {
    if (!($x eq "" && $self->{BLANK_ALLOWED})) {
      if ($self->{MIN} ne "" && $x<$self->{MIN}) {@stuff = ("below_min",$self->{MIN})}
      if ($self->{MAX} ne "" && $x>$self->{MAX}) {@stuff = ("above_max",$self->{MAX})}
    }
  }
  if ($self->{TYPE} eq "string") {
    if ($x eq "" && !($self->{BLANK_ALLOWED})) {push @stuff,"blank_not_allowed"} # sometimes causes mysterious complaints when not blank!?
  }
  if (!@stuff) {return @stuff}
  @stuff[0] = $Browser::words->get("input.".$stuff[0]);
  return @stuff;
}

# for debugging
sub dump_str {
  my $x = shift;
  print "string=$x=\n";
  while ($x=~m/(.)/g) {
    my $c = $1;
    print "==$c==".ord($c)."==\n";
  }
}

1;









